#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>

#include "config.h"
#include "prefs.h"
#include "conf.h"
#include "conv.h"
#include "vxerror.h"
#include "tools.h"
#include "rfc822.h"
#include "sysauth.h"
#include "version.h"
#include "vxopt.h"

static int check_all (DBIF_FILE udb, char *dom);
static int check_folder (char *user, dbif_rec *urec);
static int cleanup_folder (char *em_addr, char *folder, int nonstd);
static int is_user (char *lk, dbif_rec *lrec, char *rk, dbif_rec *rrec);
static int re_check (DBIF_FILE udb, char *un);

void usage (void)
{
    fprintf (stderr, "\nUsage: %s [-f <configfile>] [-a | <email addr> ...]\n"
		     "       %s -h\n"
		     "       %s -V\n\n",
		     progname,
		     progname,
		     progname);
    exit (EX_USAGE);
}

int main (int argc, char *argv[])
{
    int opt, all_folders = 0, errs = 0, rt;
    DBIF_FILE udb;
    dbif_rec *urec;
    char *un;
    vxopt (&argc, argv, usage);
    opterr = 0;
    while ((opt = getopt (argc, argv, "a")) != -1) {
	switch (opt) {
	    case 'a':
		all_folders = 1;
		break;
	    default:
		fprintf (stderr,
			 "Illegal option `%s'! Please call `%s -h' for help!\n",
			 argv[optind],
			 progname);
		return 1;
	}
    }
    /* check if other arguments are specified and if these are correct */
    if (all_folders) {
	if (optind < argc) {
	    fprintf (stderr, "%s: The `-a'-option and additional e-mail "
			     "address are mutually exclusive!\n"
			     "Please call `%s -h' for help!\n",
			     progname,
			     progname);
	    return 1;
	}
    } else if (optind >= argc) {
	fprintf (stderr, "%s: You must specify either `-a' or at least one "
			 "e-mail address!\n"
			 "Please call `%s -h' for help!\n",
			 progname,
			 progname);
	return 1;
    }
    getconf ();
    conv_init ();
    dbif_reset (&udb, prefs.userdb_name);
    if (!udb) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
		    "opening user database failed!\n");
    }
    if (all_folders) {
	errs = check_all (udb, NULL);
	dbif_close (udb);
	if (errs) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
			"checking of %d folders failed!\n",
			errs);
	}
	return 0;
    }
    while (optind < argc) {
	un = gdupl (__FILE__, __LINE__, argv[optind++]);
	if (!(urec = dbif_resolve (udb, un, NULL))) {
	    /* Untersuche, ob Benutzer existieren, die im ersten Schritt nicht
	     * gefunden wurden.
	     */
	    errs += re_check (udb, un);
	    gfree (un);
	    continue;
	}
	rt = dbif_rectype (urec);
	if (rt == RT_DOMAIN) {
	    /* Untersuche alle Benutzer in einer Domain */
	    errs += check_all (udb, un);
	    gfree (un); dbif_free (urec);
	    continue;
	}
	if (rt == RT_NSDOMAIN) {
	    /* Die Untersuchung aller Systembenutzer ist noch nicht
	     * implementiert.
	     */
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"sorry, but scanning of non-standard domains isn't "
			"implemented yet!\n");
	    gfree (un); dbif_free (urec);
	    continue;
	}
	/* Alles, was jetzt noch untersucht werden darf, sind gewhnliche
	 * Benutzer.
	 */
	if (rt != RT_USER && rt != RT_ADMIN && rt != RT_MAILSYS) {
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"`%s' isn't a valid user entry!\n", un);
	    gfree (un); dbif_free (urec);
	    continue;
	}
	errs += check_folder (un, urec);
	gfree (un); dbif_free (urec);
    }
    dbif_close (udb);
    if (errs) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "checking of %d folders failed!\n", errs);
    }
    return 0;
}

static int check_folder (char *user, dbif_rec *urec)
{
    return cleanup_folder (user, urec->folder, dbif_rectype(urec) == RT_NSUSER);
}

/* re_check (userdb, user)
 *     untersucht alle Benutzernamen, zu denen im ersten Schritt kein Eintrag
 *     gefunden wurde.
 */
static int re_check (DBIF_FILE udb, char *un)
{
    int rt, errs = 0;
    dbif_rec *urec, *drec;
    char *dn;
    char *p = strchr (un, '@');
    if (!p) {
	vxerror_at (0, 0, __FILE__, __LINE__ - 3,
		    "no user `%s' found!\n", un);
	gfree (un);
	return ++errs;
    }
    /* Checking for a domain link or a non-standard domain */
    if (!(urec = dbif_resolve (udb, ++p, &dn))) {
	vxerror_at (0, 0, __FILE__, __LINE__ - 3,
		    "no domain `%s' found!\n");
	gfree (un);
	return ++errs;
    }
    rt = dbif_rectype(urec);
    if (rt == RT_DOMAIN) {
	/* The part following the `@' was the name of a domain link;
	 * construct a new key from the username (the part preceding
	 * the `@') and the name of the real domain.
	 */
	size_t unsz = p - un;
	size_t nusz = unsz + strlen (dn);
	char *nu = galloc (__FILE__, __LINE__, nusz);
	memcpy (nu, un, unsz);
	memcpy (nu + unsz, dn, strlen(dn));
	nu[nusz] = '\0';
	gfree (un);
	un = nu;
	free (dn); dbif_free (urec);
	if (!(urec = dbif_resolve (udb, un, NULL))) {
	    /* No matching user found; this is an error! */
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"no user `%s' found!\n");
	    gfree (un);
	    return ++errs;
	}
	/* Check the folder of the user */
	errs += check_folder (un, urec);
	gfree (un); dbif_free (urec);
	return errs;
    }
    free (dn);
    if (rt != RT_NSDOMAIN) {
	/* The entry which was found by resolving the domain part
	 * is neither a domain nor a non-standard domain; this is
	 * an error.
	 */
	vxerror_at (0, 0, __FILE__, __LINE__ - 1,
		    "user database is not consistent!\n");
	gfree (un); dbif_free (urec);
	return ++errs;
    }
    /* Non-standard user! Try to find him/her in the system's user
     * database.
     */
    drec = urec;
    if (!(urec = sysauth (drec, un))) {
	/* No matching user in the system's user database found;
	 * this is an error.
	 */
	vxerror_at (0, 0, __FILE__, __LINE__ - 1,
		    "no system user `%s' found!\n");
	gfree (un); dbif_free (drec);
	return ++errs;
    }
    dbif_free (drec);
    errs += check_folder (un, urec);
    gfree (un); dbif_free (urec);
    return errs;
}

/* Untersuche alle Benutzer in der Datenbank (dom == NULL) oder in der
** Domain dom (dom != NULL)
*/
static int check_all (DBIF_FILE udb, char *dom)
{
    DBIF_CURSOR sqc;
    char *user;
    dbif_rec *urec;
    int errs = 0;
    if (!(sqc = dbif_select (udb, dom, is_user))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "selecting users %s%s%s failed!",
		    (dom ? "from domain `" : ""),
		    (dom ? dom : ""),
		    (dom ? "'" : ""));
    }
    for (user = dbif_first(sqc, &urec); user; user = dbif_next(sqc, &urec)) {
	errs += check_folder (user, urec);
	free (user); dbif_free (urec);
    }
    return errs;
}

/* A day has 86400 seconds */
#define days(x) ((x) * 86400)

struct _se {
    ssize_t sz;
    time_t dt;
    int del;
};

/* forward decl for `update_folder' */
static int update_folder (xFile fx,
			  char *folder,
			  struct _se *msgtbl,
			  int nmsgs,
			  char *line,
			  size_t lsz);

static int cleanup_folder (char *em_addr, char *folder, int nonstd)
{
    struct _se *msgtbl = (struct _se *) NULL;
    char line[1024];
    xFile fx;
    int ix, nmsgs, errs;
    time_t now;
    if (!(fx = xOpen (folder, (nonstd ? "ix" : "rx"), prefs.fperm))) {
	vxerror_at (0, 0, __FILE__, __LINE__ - 1,
		 "Sorry, but the folder for user `%s' does not exist\n"
		 "    or is unreadable for me! I assume it is an empty "
		 "folder!\n",
		 em_addr);
	return 1;
    }
    /* Konvertiere einen NICHT-Standard-Folder in das Standardformat */
    if (nonstd) { conv (fx); xReSet (fx); }
    /* check the size of any message */
    for (ix = 0; xMsgLen (fx, ix) >= 0; ++ix);
    nmsgs = ix;
    if (!(msgtbl = (struct _se *) malloc (nmsgs * sizeof (struct _se)))) {
	xClose (fx);
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 2, "(malloc)");
    }
    for (ix = 0; ix < nmsgs; ++ix) {
	msgtbl[ix].sz = xMsgLen (fx, ix);
	xSeek (fx, ix);
	msgtbl[ix].dt = (time_t) 0;
	msgtbl[ix].del = 0;
	while (xGets (line, 1024, fx) > 0) {
	    if (!lccmp (line, "Date: ", 6)) {
		if (rfc822_time (line + 6, &msgtbl[ix].dt)) {
		    fprintf (stderr,
			     "%s: msg#%d's date is not rfc822 compliant; "
			     "ignoring!\n",
			     progname,
			     ix);
		}
	    }
	}
    }
    /* The important data is read; now i must check each message and
    ** set it's delete flag if it violates the limits
    */
    time (&now);
    for (ix = 0; ix < nmsgs; ++ix) {
	/* If a size check is enabled and the message size is greater
	** than the maximum size allowed, then mark the message as deleted!
	*/
	if (prefs.check_size && msgtbl[ix].sz > prefs.check_size) {
	    msgtbl[ix].del = 1;
	}
	/* If a date check is enabled and the message is older than the
	** number of days allowed, then mark the message as deleted!
	*/
	if (prefs.check_days && msgtbl[ix].dt < now - days(prefs.check_days)) {
	    msgtbl[ix].del = 1;
	}
    }
    errs = update_folder (fx, folder, msgtbl, nmsgs, line, 1024);
    xClose (fx);
    if (nonstd) { xNoIndex (folder); }
    return errs;
}

/* remove all deleted messages from the folder */

static int update_folder (xFile fx,
			  char *folder,
			  struct _se *msgtbl,
			  int nmsgs,
			  char *line,
			  size_t lsz)
{
    char *tmp;
    xFile dx, rx;
    int ix;
    tmp = galloc (__FILE__, __LINE__, strlen (progname) + 20);
    sprintf (tmp, "/tmp/%s.%d\n", progname, getpid ());
    if (!(dx = xOpen (tmp, "ux", prefs.fperm))) {
	vxerror_at (0, 0, __FILE__, __LINE__ - 1,
		    "creating temporary folder `%s' failed",
		    tmp);
	return 1;
    }
    for (ix = 0; ix < nmsgs; ++ix) {
	if (!msgtbl[ix].del) {
	    xSeek (fx, ix);
	    while (xGets (line, lsz, fx) > 0) {
		xPuts (line, dx);
	    }
	    xSet (dx);
	}
    }
    if (!(rx = xOpen (tmp, "rx"))) {
	xUnlink (tmp);
	vxerror_at (0, 0, __FILE__, __LINE__ - 2,
		    "re-reading temporary folder `%s' failed",
		    tmp);
	return 1;
    }
    xClose (dx);
    if (!(dx = xOpen (folder, "ux"))) {
	/* Sorry, but i cannot recover from this error - it is FATAL */
	/*xUnlink (tmp);*/
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 2,
		    "re-writing folder `%s' failed",
		    folder);
    }
    for (ix = 0; xMsgLen (rx, ix) >= 0; ++ix) {
	xSeek (rx, ix);
	while (xGets (line, lsz, rx) >= 0) {
	    xPuts (line, dx);
	}
	xSet (dx);
    }
    xClose (dx);
    xClose (rx);
    xUnlink (tmp);
    return 0;
}

static int is_user (char *lk, dbif_rec *lrec, char *rk, dbif_rec *rrec)
{
    int rt = 0;
    if (lk) {
	rt = dbif_rectype (lrec);
	if ((1 << rt) & (CC_USER|CC_NSUSER|CC_ADMIN)) {
	    return (rt == dbif_rectype (rrec)) && !strcmp (lk, rk);
	}
	if ((1 << rt) & (CC_DOMAIN|CC_NSDOMAIN)) {
	    char *p;
	    rt = dbif_rectype (rrec);
	    if (!((1 << rt) & (CC_USER|CC_ADMIN|CC_NSUSER))) { return 0; }
	    p = strchr (rk, '@');
	    if (!p) { return 0; }
	    return !strcmp (++p, lk);
	}
	return 0;
    } else {
	return ((1 << rt) & (CC_USER|CC_NSUSER|CC_ADMIN|CC_MAILSYS));
    }
}
