#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <getopt.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <sys/param.h>

#ifdef HAVE_ZLIB
#include <zlib.h>
#endif

#include "prefs.h"
#include "userdb.h"
#include "config.h"
#include "conf.h"
#include "vxerror.h"
#include "tools.h"
#include "xfile.h"
#include "version.h"


/* Vorwrtsdeklarationen fr interne Funktionen */
static void send_failed_to_admin (char *em_addr,
				  char *tmpfile,
				  char *sender,
				  char *message);

void cleanup (void)
{
    char tmpfile[PATH_MAX + 1];
    sprintf (tmpfile, "/tmp/vxdeliver.%d", getpid ());
    unlink (tmpfile);
}

void sighandle (int *signum)
{
    cleanup ();
    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		"%s: caught signal number %d\n",
		progname, signum);
}

void usage (void)
{
    fprintf (stderr,
	     "\nUsage: %s [-f <configfile>] sender-addr email-addr ..."
	     "\n       %s -h"
	     "\n       %s -V\n"
	     "\n   -f <configfile> (alt: --config=<configfile>)"
	     "\n       read configuration from file <configfile>"
	     "\n       (default: %s)"
	     "\n   -h (alt: --help)"
	     "\n       show this help"
	     "\n   -V (alt: --version)"
	     "\n       show the program's version number\n\n",
	     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}

static int errflag = 0;

int main (int argc, char *argv[])
{
    int fd;
    FILE *fp;
#ifdef HAVE_ZLIB
    gzFile zp;
#endif
    char tmpfile[PATH_MAX + 1];
    char line[1024];
    int ix, c;
    char *sender = "";
    char filler[PATH_MAX + 1];
    int ll = strlen (argv[0]), lc;
    long sz;

    /* reset error counter */
    errflag = 0;

    /* Settings: */
    progname = strrchr (argv[0], '/');
    if (progname) ++progname; else progname = argv[0];
    sprintf (tmpfile, "/tmp/vxdeliver.%d", getpid ());
    memset (filler, '\0', PATH_MAX + 1);
    memset (filler, ' ', (ll > PATH_MAX ? PATH_MAX : ll));

    /* Option checking: */
    for (;;) {
	int option_index = 0;
	static struct option long_options[] = {
	    { "config", 1, NULL, 'f' },
	    { "help", 0, NULL, 'h' },
	    { "version", 0, NULL, 'V' },
	    { NULL, 0, NULL, 0 }
	};
	c = getopt_long (argc, argv, "f:hV", long_options, &option_index);
	if (c == -1) break;
	switch (c) {
	    case 'f':
		prefs.config_name = gdupl (__FILE__, __LINE__, optarg);
		break;
	    case 'h':
		usage ();
	    case 'V':
		fprintf (stderr, "%s\n", vxVersion ());
		exit (0);
	    default:
		vxerror_at (EX_USAGE, 0,
			    __FILE__, __LINE__ - 1,
			    "unknown option `%s'!\n"
			    "Use `%s --help' for help, please!\n",
			    argv[optind], progname);
	}
    }

    getconf ();

    /* User database initialisieren */
    init_userdb ();

    /* Wenn ich erst hier userid und group id setze, dann kann die
    ** Konfiguration auch in einer Datei stehen, die nicht `vxuid'
    ** gehrt und fr `vxuid' gesperrt ist.
    */
    setgid (prefs.vxgid);
    setuid (prefs.vxuid);

    /* Argument checking: */
    if (argc - optind < 2) {
	vxerror_at (EX_USAGE, 0,
		    __FILE__, __LINE__ - 2,
		    "you hav to specify at least one e-mail address!\n"
		    "Use `%s --help' for help, please!\n",
		    progname);
    }
    sender = argv[optind++];

    /* install a `cleanup' function in the atexit protocol */
    if (atexit (cleanup)) {
	vxerror_at (EX_OSERR, errno,
		    __FILE__, __LINE__ - 2,
		    "atexit (cleanup) failed\n");
    }

    /* write the `stdin' contents to a temporary file */
    if ((fd = open (tmpfile, O_WRONLY|O_EXCL|O_CREAT|O_TRUNC, 0600)) < 0) {
	vxerror_at (EX_CANTCREAT, errno,
		    __FILE__, __LINE__ - 2,
		    "writing tempfile `%s' failed", tmpfile);
    }
#ifdef HAVE_ZLIB
    if (prefs.compression) {
	if (!(zp = gzdopen (fd, "wb9"))) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
			"gzdopen (\"%s\", ...) failed", tmpfile);
	}
    } else {
	if (!(fp = fdopen (fd, "w"))) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
			"fdopen (\"%s\", ...) failed", tmpfile);
	}
    }
#else
    if (!(fp = fdopen (fd, "w"))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "fdopen (\"%s\", ...) failed", tmpfile);
    }
#endif
    sz = 0l;
    while ((fgets (line, 1024, stdin))) {
#ifdef HAVE_ZLIB
	if (prefs.compression) {
	    gzputs (line, zp);
	} else {
	    fputs (line, fp);
	}
#else
	fputs (line, fp);
#endif
	sz += strlen (line);
    }
#ifdef HAVE_ZLIB
    if (prefs.compression) { gzclose (zp); } else { fclose (fp); }
#else
    fclose (fp);
#endif
    /* Check the size of the incoming mail
    **   - size checking is generally enabled and
    **   - if delivery size checking is enabled
    */
    if (prefs.check_size && prefs.dsizecheck && (sz > prefs.check_size)) {
	/* The Message is too long, so exit with an error message */
	/* I will try EX_DATAERR as exit code and an error message
	** containing the exact reason why `vxdeliver' terminates here!
	*/
	if (prefs.errors == ERRORS_ADMIN) {
	    /* NEVER include the body of a message which is too long */
	    prefs.biirm = 0;
	    /* send the message to the system mail administrator */
	    send_failed_to_admin ("::MAILADMIN::",
				  tmpfile,
				  sender,
				  "maximum message size exceeded!\n");
	    exit (0);
	} else {
	    /* Return with a non-zero exit status */
	    vxerror_at (EX_DATAERR, 0, __FILE__, __LINE__,
			"Sorry but the message you've sent is (%ld bytes) "
			"too long!\n",
			(sz - prefs.check_size));
	}
    }
    for (ix = optind; ix < argc; ++ix) {
	char *folder;
	/* Hiermit versuche ich einigen Schwierigkeiten, die mir der
	** `sendmail' bereitet, aus dem Weg zu gehen.
	*/
/*	if (prefs.sendmail_vdwa) {
**	    char *p = argv[ix], *q;
**	    while (!isalpha (*p)) { ++p; }
**	    strcpy (argv[ix], p);
**	    p = strchr (argv[ix], '@');
**	    if (!p) { p = argv[ix] + strlen (argv[ix]); }
**	    for (q = p - 1; q != argv[ix] && !isalpha (*q); --q);
**	    if (isalpha (*q)) ++q; strcpy (q, p);
**	    p = argv[ix]; p += strlen (p);
**	    if (*--p == '.') { *p= '\0'; }
**	}
*/
	/* So, jetzt kann der Name des Folders bestimmt werden. */
	folder = user_folder (argv[ix]);
	/* Wenn der User im ersten Schritt nicht bestimmt werden konnte,
	** dann mu im zweiten Schritt nachgeschaut werden, ob der
	** Domain-Anteil vielleicht einen Domain-Alias bezeichnet; wenn ja,
	** dann mu die ganze Prozedur nach Ersetzung des Domainanteils in
	** der E-Mail-Adresse noch einmal wiederholt werden. Erst wenn das
	** keinen Erfolg bringt, dann ist die Suche endgltig fehlgeschlagen.
	** Fr alle diese Schritte habe ich eine weitere Routine geschrieben:
	** `user_rewrite'. Falls diese erfolgreich zurckkehrt, dann hat der
	** Domain-Anteil in der E-Mail-Adresse auf eine andere Domain
	** verwiesen.
	*/
	if (!folder) { folder = user_rewrite (argv[ix]); }
	if (!folder) {
	    if (prefs.errors == ERRORS_ADMIN) {
		send_failed_to_admin (argv[ix], tmpfile, sender, (char *) NULL);
	    } else if (prefs.errors == ERRORS_SYSTEM) {
		vxerror (0, 0, "user/domain `%s' not found!\n", argv[ix]);
		++errflag;
	    }
	} else {
	    FILE *src; xFile tgt;
	    if (!(tgt = xOpen (folder, "wx", prefs.fperm))) {
		vxerror_at (EX_CANTCREAT, errno,
			    __FILE__, __LINE__ - 2,
			    "writing folder `%s' for user `%s' failed",
			    folder, argv[ix]);
	    }
	    if (!(src = fopen (tmpfile, "r"))) {
		vxerror_at (EX_NOINPUT, errno,
			    __FILE__, __LINE__ - 2,
			    "reading tempfile `%s' failed",
			    tmpfile);
	    }
	    while ((fgets (line, 1024, src))) {
		xPuts (line, tgt);
	    }
	    fclose (src);
	    xClose (tgt);
	    gfree (folder);
	}
    }
    if (errflag) {
	vxerror (EX_NOUSER, 0, "%d e-mail user%s do%s not exist!\n",
		 errflag,
		 (errflag == 1 ? "" : "s"),
		 (errflag == 1 ? "es" : "n't"));
    }
    return 0;
}

/* Diese Funktion liefert den Namen des Rechners, auf dem VxDeliver aufgreufen
** wurde. Wenn das zu diesem Zweck aufgerufene `gethostname' fehlschlgt, dann
** wird stattdessen einfach `localhost' zurckgegeben.
*/

static void send_failed_to_admin (char *em_addr,
				  char *tmpfile,
				  char *sender,
				  char *message)
{
    /* Der Identity-Eintrag `::VXMAIL::' enthlt zugleich den Namen des
     * Administrators einer Domain.
     */
    char *admname = user_folder ("::VXMAIL::");
    char *domain = strchr (em_addr, '@');
    char *admin, *admfolder, line[1024];
    FILE *tmp; xFile adm;
    int ssz;
    time_t tim;
    if (!domain) {
	vxerror_at (EX_SOFTWARE, 0,
		    __FILE__, __LINE__ - 2,
		    "e-mail address is invalid (no domain)!\n");
    }
    ssz = strlen (admname) + strlen (domain) + 1;
    admin = galloc (__FILE__, __LINE__, ssz);
    strcpy (admin, admname);
    strcat (admin, domain);
    if (!domain_check (admin)) {
	/* Wenn die Domain-Adresse nicht existiert, dann mu eine
	 * entsprechende Mail anstatt an den Domainadministrator an den
	 * Mail-Administrator geschickt werden.
	 */
	if (!(admfolder = user_folder ("::MAILADMIN::"))) {
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"MAIL SYSTEM administrator does not exist! "
			"This is a FATAL ERROR!\n",
			strrchr (admin, '@') + 1);
	    ++errflag;
	    gfree (admin);
	    return;
	}
    } else {
	/* Der ADMIN-Eintrag zu einer Domain MU (!!!) immer vorhanden
	** sein. Ansonsten haben wir eine Datenbankinkonsistenz, gegen die
	** an dieser Stelle sowieso nichts getan werden kann!
	*/
	admfolder = user_folder (admin);
	if (!admfolder) { admfolder = user_rewrite (admin); }
	if (!(admfolder = user_folder (admin))) {
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"inconsistent database! This is a FATAL ERROR\n");
	    ++errflag;
	    gfree (admin);
	    return;
	}
    }
    if (!(adm = xOpen (admfolder, "wx", prefs.fperm))) {
	vxerror_at (EX_CANTCREAT, errno,
		    __FILE__, __LINE__ - 2,
		    "failed to append to admin folder `%s'",
		    admin);
    }
    if (!(tmp = fopen (tmpfile, "r"))) {
	vxerror_at (EX_NOINPUT, errno,
		    __FILE__, __LINE__ - 2,
		    "failed to open tmpfile `%s'",
		    tmpfile);
    }
    (void) time (&tim);
    /* From VxDeliver `datum' */
    xPuts ("From VxDeliver ", adm);
    strcpy (line, asctime (localtime (&tim)));
    if (line[strlen (line) - 1] == '\n') { line[strlen (line) - 1] = '\0'; }
    xPuts (line, adm);
    xPuts ("\r\n", adm);
    /* Return-Path: <`sender'> */
    xPuts ("Return-Path: <", adm); xPuts (sender, adm); xPuts (">\r\n", adm);
    /* From: VxDeliver@`current host' */
    xPuts ("From: <VxDeliver@", adm); xPuts (mailhost (), adm);
    xPuts (">\r\n",  adm);
    /* To: `admin-addr' */
    xPuts ("To: ", adm); xPuts (admin, adm); xPuts ("\r\n", adm);
    /* Date: ddd, dd mmm yyyy HH:MM:SS +xxxx */
    (void) strftime (line, 1024, "Date: %a, %d %b %Y %H:%M:%S\r\n",
		     localtime (&tim));
    xPuts (line, adm);
    /* Subject: delivering mail to `<user>' failed! */
    xPuts ("Subject: delivering mail to `", adm); xPuts (em_addr, adm);
    xPuts ("' failed!\r\n\r\n", adm);
    /* |------------------------- Failed addresses follow: ----------------------| */
    xPuts ("|------------------------- Failed addresses follow: ", adm);
    xPuts ("----------------------|\r\n", adm);
    /*   `email-address'   */
    xPuts ("  ", adm); xPuts (em_addr, adm); xPuts ("  \r\n", adm);
    if (message) {
	xPuts ("Delivery message: ", adm); xPuts (message, adm);
	xPuts ("\r\n", adm);
    }
    /* |------------------------- Message text follows: -------------------------| */
    xPuts ("|------------------------- Message text follows: ", adm);
    xPuts ("-------------------------|\r\n", adm);
    /* Message header */
    while (fgets (line, 1024, tmp)) {
	xPuts (line, adm);
	if (*line == '\n' && !line[1]) break;
	if (*line == '\r' && line[1] == '\n' && !line[2]) break;
    }
    /* Message body */
    if (prefs.biirm) {
	while (fgets (line, 1024, tmp)) {
	    xPuts (line, adm);
	}
    } else {
	xPuts ("[no body included]\r\n\r\n", adm);
    }
    fclose (tmp);
    xClose (adm);
    gfree (admin);
    gfree (admfolder);
}
