/* deliver.c
**
** send a mail to all recipients in a list file
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/wait.h>

#include "prefs.h"
#include "tools.h"
#include "vxerror.h"
#include "xfile.h"
#include "dbif.h"

#include "userdb.h"
#include "fcheck.h"
#include "hlsock.h"
#include "smtp.h"

extern char **environ;

/* Diese Routine soll eine Optimierung enthaltenen:
**   Empfnger, die als Domainteil eine der Domains des VxMail-Systems
**   enthalten, sollen direkt zugestellt werden.
*/

static int delivery_level = 0;
static int errflag = 0;

/* return the number of delivery errors */
int delivery_errs(void) { return errflag; }

/* Sende eine von `vxdeliver' abgewiesene E-Mail an den Default-Administrator
** der betreffenden Domain bzw. `::MAILSYS::' - falls die erste Mglichkeit
** nicht gegeben ist.
*/
void deliver_error (char *rorig, char *sender, char *em_addr,
		    char *tmpfile, char *message);

/* Verschicke eine E-Mail an alle in einer Mailingliste stehenden Empfnger. */
static void deliver_list (char *rorig, char *sender, char *addr,
			  char *tmpfile);

/* Schreibe eine E-Mail in die Standardeingabe eines zu diesem Zweck von
** `vxdeliver' gestarteten Programmes.
*/
static void deliver_prog (char *rorig, char *sender, char *cmdline,
			  char *tmpfile);

/* Hnge eine E-Mail an eine Datei (Textdatei oder xFile) an. */
static void deliver_file (char *rorig, char *sender, char *outfile,
			  char *tmpfile);

/* send a message to a (potentially non-local) user */
static void deliver_nonlocal (char *rorig, char *sender, char *addr,
			      char *tmpfile);

static void deliver_smtp (char *rorig, char *sender, char *recip_host,
			  char *recipient, char *tmpfile);

static void deliver_bsmtp (char *sender, char *_batchfile,
			   char *recipient, char *tmpfile);

/* berlies eine eventuell in der die E-Mail enthaltenden temporren Datei
** vorkommende `From ...'-Zeile.
*/
static off_t skip_From (char *tmpfile);

/* Erzeuge eine neue Zeichenkette durch Ersetzung aller vorkommenden
** Symbole `old' durch den Text `new'.
*/
static char *subst (char *text, char *old, char *new);

/* Liefere eine Meldung an einen lokalen Empfnger aus */
void deliver_local (char *rorig, char *sender, char *em_addr, char *tmpfile)
{
    int rtype;
    char *folder, *p, line[1024];
    FILE *src; xFile tgt;
    if (rorig && !strcmp (rorig, em_addr)) {
	deliver_error (NULL, sender, rorig, tmpfile,
		       "recursive mail target found!\n");
	return;
    }
    if (delivery_level > prefs.maxlevel) {
	deliver_error (rorig, sender, em_addr, tmpfile,
		       "maximum delivery level exceeded; you should check "
		       "if one of your\n"
		       "`list link' entries is a recursive reference "
		       "(directly or indirectly)\n"
		       "to itself.\n");
	return;
    }
    ++delivery_level;
    /* convert the e-mail to lower case if it is not the name of the
    ** mail system administrator and the configuration requires it
    */
    if (!strcmp (em_addr, "::MAILSYS::")) {
	p = strchr (em_addr, '@');
	if (!p) { p = em_addr + strlen (em_addr); }
	if (prefs.lcuser) { lcconv (em_addr, p - em_addr); }
	if (prefs.lcdomain) { lcconv (p, -1); }
    }
    /* check if the corresponding mail filter rejects the e-mail */
    if (!fcheck (em_addr, tmpfile)) {
	if (prefs.rejverb) {
	    deliver_error (NULL, sender, rorig, tmpfile,
			   "mail rejected by filter");
	}
	--delivery_level;
	return;
    }
    /* try to find the recipient */
    if (!(folder = user_folder (em_addr, &rtype))) {
	vxlog (1, "`%s' not found - retrying ... ", em_addr);
	folder = user_rewrite (em_addr, &rtype);
	vxlog (1, "done.\n");
    }
    /* if the recipient does not exist, send an error message to the
    ** corresponding default domain administrator (or to ::MAILSYS::)
    */
    if (!folder) {
	vxlog (0, "delivering error message!\n");
	deliver_error (rorig, sender, em_addr, tmpfile, NULL);
	--delivery_level;
	return;
    }
    if (rtype == RT_BSMTP) {
	/* (Syntax: `/var/spool/mail/vx/bsmtp/relay' (Beispiel))
	** Die E-Mail soll in einem VxMail-Folder als halbfertige SMTP-
	** Session abgespeichert werden und per SMTP spter an eine
	** dynamische Zieladresse gesendet werden - getriggert durch
	** eine POP3-Anfrage auf diesen Benutzereintrag; dazu mu ein
	** Pawort beim Anlegen dieses Benutzereintrages in der Datenbank
	** mit angegeben werden. Auch dieser Eintrag ergibt nur dann einen
	** wirklichen Sinn, wenn mehrere Empfnger (oder eine ganze Domain)
	** auf ihn verweisen.
	*/
	deliver_bsmtp (sender, folder, em_addr, tmpfile);
	gfree (folder);
	--delivery_level;
	return;
    } else if (rtype == RT_LLINK) {
	char *recipient = (rorig ? rorig : em_addr);
	if (*folder == '|') {
	    /* (Syntax: `|/usr/local/lib/mlsender' (Beispiel))
	     * Die bermittlung der E-Mail soll durch ein Programm geschehen.
	    ** blicherweise ist dies ein Programm zur Verwaltung von Mailing-
	    ** listen.
	    */
	    deliver_prog (recipient, sender, folder + 1, tmpfile);
	} else if (*folder == '>') {
	    /* (Syntax: `>/var/lists/Data/liste1.txt' oder
	     *          `>>/var/lists/Data/liste1.txt' oder
	     *          `>:/var/lists/Data/liste1.txt' (Beispiel))
	     * Die E-Mail soll an eine Textdatei oder einen VxMail-Folder
	    ** angehngt werden - interessant im Zusammenhang mit der
	    ** Archivierung von E-Mails an Mailinglisten.
	    */
	    deliver_file (recipient, sender, folder + 1, tmpfile);
	} else if (*folder == '/') {
	    /* (Syntax: `/var/vxmail/liste1' (Beispiel))
	    ** Die E-Mail soll an alle in der angegebenen Datei stehenden
	    ** Empfnger weitergereicht werden; es handelt sich hierbei um
	    ** einen einfachen Mechanismus fr Mailinglisten und Aliases mit
	    ** mehr als einem Adressaten.
	    ** Die Mailingliste mu in diesem Fall als absoluter Pfadname
	    ** angegeben werden.
	    */
	    deliver_list (recipient, sender, folder, tmpfile);
	} else if (strchr (folder, '@')) {
	    /* (Syntax: `fbj@blinx.de' (Beispiel))
	    ** Ein (mglicherweise nicht-lokaler) Empfnger ist das Ziel
	    ** der E-Mail. Dies eignet sich fr die bertragung von E-Mails
	    ** an einzelne Empfnger auf einem anderen Rechner.
	    */
	    deliver_nonlocal (recipient, sender, folder, tmpfile);
	} else if (!strncmp (folder, "smtp:", 5)) {
	    /* (Syntax: `smtp:mail.zweitehand.de' oder `smtp:[194.115.191.214]'
	    **  (Beispiel))
	    ** Eine SMTP-bertragung an den durch <hostname> oder <ip-addr>
	    ** angegebenen Host soll stattfinden. Dies eignet sich fr eine
	    ** Direkbertragung an eine nicht auf dem gleichen Rechner liegende
	    ** Domain - bzw. jeden in dieser Domain stehenden Benutzer.
	    ** Das ganze ergibt natrlich nur dann einen Sinn, wenn mehrere
	    ** E-Mail-Empfnger auf diesen Eintrag verweisen (USER "<OTHER>"),
	    ** da ansonsten der Empfnger u.U. nicht korrekt ermittelt werden
	    ** kann; auerdem mu der empfangende SMTP-Server die Domain als
	    ** seine eigene (lokale Domain) werten (oder evtl. als `virtual'?:).
	    ** (Z. Zt. noch nicht implementiert.)
	    */
	    deliver_smtp (rorig, sender, folder + 5, em_addr, tmpfile);
	} else {
	    /* (Syntax: `VxMail-Liste' (Beispiel))
	    ** Die E-Mail soll an alle in der angegebenen Datei stehenden
	    ** Empfnger weitergereicht werden; es handelt sich hierbei um
	    ** einen einfachen Mechanismus fr Mailinglisten und Aliases mit
	    ** mehr als einem Adressaten.
	    ** Die Mailingliste mu in diesem Fall in dem durch die VxMail-
	    ** Konfiguration (`listdir /var/VxMail/lists') spezifizierten
	    ** Verzeichnis stehen.
	    */
	    char *newpath;
	    newpath=gconc(__FILE__, __LINE__, prefs.listdir, "/", folder, NULL);
	    deliver_list (recipient, sender, newpath, tmpfile);
	    gfree (newpath);
	}
	gfree (folder);
	--delivery_level;
	return;
    }
    /* Liefere die E-Mail ganz normal aus */
    if (!(tgt = xOpen (folder, "wx", prefs.fperm))) {
	vxerror_at (EX_CANTCREAT, errno, __FILE__, __LINE__ - 1,
		    "writing folder for user `%s' failed", em_addr);
    }
    if (!(src = fopen (tmpfile, "r"))) {
	vxerror_at (EX_NOINPUT, errno, __FILE__, __LINE__ - 1,
		    "reading temporary file `%s' failed", tmpfile);
    }
    vxlog (5, "delivering message to user `%s' ", em_addr);
    while (fgets (line, 1024, src)) { xPuts (line, tgt); vxlog (5, "."); }
    xClose (tgt);
    vxlog (5, " done.\n");
    if (!prefs.syswithidx && rtype == RT_NSUSER) { xNoIndex (folder); }
    fclose (src);
    gfree (folder);
    --delivery_level;
}

/* Liefere eine von `vxdeliver' abgewiesene E-Mail an den jeweiligen
** Domain-Administrator bzw. `::MAILSYS::' aus.
*/
void deliver_error (char *rorig, char *sender, char *em_addr,
		    char *tmpfile, char *message)
{
    int rtype;
    char *admname, *domain, *admin, *admfolder, line[1024], *p;
    FILE *tmp; xFile adm;
    time_t tim;
    off_t fsoff;
    ++errflag;
    vxlog (1, "error delivery mode ... ");
    if (prefs.errors != ERRORS_ADMIN) { vxlog (1, "system.\n"); return; }
    vxlog (1, "admin.\n");
    admname = defadmin ();
    if (strcmp (em_addr, "::MAILSYS::")) {
	if (!(domain = strchr (em_addr, '@'))) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
			"e-mail address is invalid (no domain)!\n");
	}
	admin = gconc (__FILE__, __LINE__, admname, domain, NULL);
    } else {
	admin = gdupl (__FILE__, __LINE__, "::MAILSYS::");
    }
    if (!domain_check (admin)) {
	/* Wenn die Domain-Adresse nicht existiert, dann mu eine
	** entsprechende Mail anstatt an den Domainadministrator an den
	** Mail-Administrator geschickt werden.
	*/
	if (!(admfolder = user_folder ("::MAILSYS::", &rtype))) {
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"MAIL SYSTEM administrator does not exist! "
			"This is a FATAL ERROR!\n");
	    gfree (admin);
	    gfree (admname);
	    return;
	}
    } else {
	/* Der ADMIN-Eintrag zu einer Domain MU (!!!) immer vorhanden
	** sein. Ansonsten haben wir eine Datenbankinkonsistenz, gegen die
	** an dieser Stelle sowieso nichts getan werden kann!
	*/
	if (!(admfolder = user_folder (admin, &rtype))) {
	    admfolder = user_rewrite (admin, &rtype);
	}
	if (!admfolder) {
	    vxerror_at (0, 0, __FILE__, __LINE__ - 1,
			"inconsistent database! This is a FATAL ERROR!\n");
	    gfree (admin);
	    gfree (admname);
	    return;
	}
    }
    if (!(adm = xOpen (admfolder, "wx", prefs.fperm))) {
	vxerror_at (EX_CANTCREAT, errno, __FILE__, __LINE__ - 1,
		    "writing folder for admin `%s' failed", admin);
    }
    fsoff = skip_From (tmpfile);
    if (!(tmp = fopen (tmpfile, "r"))) {
	vxlog (1, "tmpfile `%s' not found!\n", tmpfile);
	vxerror_at (EX_NOINPUT, errno, __FILE__, __LINE__ - 1,
		    "reading temporary file `%s' failed", tmpfile);
    }
    (void) fseek (tmp, fsoff, SEEK_SET);
    (void) time (&tim);
    /* From VxDeliver `datum' */
    xPuts ("From VxDeliver ", adm);
    strcpy (line, asctime (localtime (&tim)));
    p = line + strlen (line) - 1;
    if (*p == '\n') { *p = '\0'; }
    xPuts (line, adm); xPuts ("\r\n", adm);
    /* Return-Path: <`sender'> */
    xPuts ("Return-Path: <", adm); xPuts (sender, adm); xPuts (">\r\n", adm);
    /* From: VxDeliver@`current host' */
    xPuts ("From: <VxDeliver@", adm); xPuts (mailhost (), adm);
    xPuts (">\r\n",  adm);
    /* To: `admin-addr' */
    xPuts ("To: ", adm); xPuts (admin, adm); xPuts ("\r\n", adm);
    if (rorig) {
	/* Orig-To: `original address' */
	xPuts ("To: ", adm); xPuts (rorig, adm); xPuts ("\r\n", adm);
    }
    /* Date: ddd, dd mmm yyyy HH:MM:SS +xxxx */
    (void) strftime (line, 1024, "Date: %a, %d %b %Y %H:%M:%S\r\n",
		     localtime (&tim));
    xPuts (line, adm);
    /* Subject: delivering mail to `<user>' failed! */
    xPuts ("Subject: delivering mail to `", adm); xPuts (em_addr, adm);
    xPuts ("' failed!\n\n", adm);
    /* |------------------------- Failed addresses follow: ----------------------| */ 
    xPuts ("|------------------------- Failed addresses follow: ", adm);
    xPuts ("----------------------|\r\n", adm);
    /*   `email-address'   */
    xPuts ("  ", adm); xPuts (em_addr, adm); xPuts ("  \r\n", adm);
    if (message) {
	char *p, *q;
	xPuts ("Delivery error text: ", adm);
	q = message;
	p = strchr (q, '\n');
	while (p) {
	    strncpy (line, q, (p - q)); line[p - q] = '\0';
	    xPuts (line, adm);
	    q = p + 1;
	    p = strchr (q, '\n');
	    if (*q) { xPuts ("\r\n\t", adm); } else { xPuts ("\r\n", adm); }
	}
	if (*q) { xPuts (q, adm); xPuts ("\r\n", adm); }
    }
    /* |------------------------- Message text follows: -------------------------| */
    xPuts ("|------------------------- Message text follows: ", adm);
    xPuts ("-------------------------|\r\n", adm);
    /* Message header */
    while (fgets (line, 1024, tmp)) {
	xPuts (line, adm);
	if (*line == '\n' && !line[1]) break;
	if (*line == '\r' && line[1] == '\n' && !line[2]) break;
    }
    /* Message body */
    if (prefs.biirm) {
	while (fgets (line, 1024, tmp)) {
	    xPuts (line, adm);
	}
    } else {
	xPuts ("[no body included]\r\n\r\n", adm);
    }
    fclose (tmp);
    xClose (adm);
    gfree (admin);
    gfree (admfolder);
}

/* Entferne das Zeilenende (`\r', `\n' oder `\r\n') von einer Zeile und gib
** ein Flag zurck, das darber Auskunft gibt, ob die Zeile ein Zeilenende
** enthielt (also abgeschlossen war) oder nicht.
*/
static __inline__ int cuteol (char *line)
{
    char *p = line + strlen (line);
    if (p == line) { return 0; }
    if (*--p == '\r') {
	*p = '\0'; return 1;
    }
    if (*p == '\n') {
	*p = '\0';
	if (p != line && *--p == '\r') { *p = '\0'; }
	return 1;
    }
    return 0;
}

/* Liefere E-Mails an Empfnger, die in einer Listendatei stehen, oder aber
** an ein Programm oder eine Datei oder einen lokalen oder nicht-lokalen
** Benutzer aus.
*/
static void deliver_list (char *rorig, char *sender, char *addr,
			  char *tmpfile)
{
    char line[1024], rl[1024];
    FILE *lf;
    int skipeol;
    if (!(lf = fopen (addr, "r"))) {
	deliver_error (rorig, sender, rorig, tmpfile, strerror (errno));
	return;
    }
    while (fgets (line, 1024, lf)) {
	skipeol = !cuteol (line);
	while (skipeol) {
	    if (!fgets (rl, 1024, lf)) { break; }
	    skipeol = !cuteol (rl);
	}
	if (!*line || *line == '#') continue;
	if (*line == '|') {
	    deliver_prog (rorig, sender, line + 1, tmpfile);
	} else if (*line == '>') {
	    deliver_file (rorig, sender, line + 1, tmpfile);
	} else if (*line != '/' && strchr (line, '@')) {
	    deliver_nonlocal (rorig, sender, line, tmpfile);
	} else {
	    deliver_error (rorig, sender, rorig, tmpfile,
			   "delivery to a list file from within a list file "
			   "is illegal!\n");
	}
    }
    fclose (lf);
    return;
}

static char *shellpath = "/bin/sh";
static char *shell[] = { "sh", "-c", NULL, NULL };

/* Liefere eine E-Mail an ein Programm aus. */
static void deliver_prog (char *rorig, char *sender, char *cmdline,
			  char *tmpfile)
{
    int fd, pst;
    char *newcmd, *newcmd1;
    pid_t shpid;
    newcmd1 = subst (cmdline, "<RECIPIENT>", rorig);
    newcmd = subst (newcmd1, "<SENDER>", sender);
    gfree (newcmd1);
    if ((fd = open (tmpfile, O_RDONLY)) < 0) {
	vxlog (1, "tmpfile `%s' not found!\n", tmpfile);
	vxerror_at (EX_NOINPUT, errno, __FILE__, __LINE__ - 1,
		    "reading temporary file `%s' failed", tmpfile);
    }
    switch ((shpid = fork ())) {
	case -1:
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 2, "(fork)");
	case 0:
	    dup2 (fd, 0);
	    close (fd);
	    shell[2] = newcmd;
	    seteuid (getuid ());
	    setgid (prefs.vxgid);
	    setuid (prefs.vxuid);
	    execve (shellpath, shell, environ);
	    vxerror (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(execve)");
	default:
	    close (fd);
	    if (waitpid (shpid, &pst, 0) < 0) { return; }
	    if ((WIFEXITED(pst) && WEXITSTATUS(pst)) || WIFSIGNALED(pst)) {
		return;
	    }
    }
    gfree (newcmd);
    return;
}

/* Hnge eine E-Mail an eine Datei an:
** Wurde `>:dateiname' angegeben, dann wird die E-Mail an eine Datei vom Typ
** `xFile' (Folder mit Index) angehngt; wurde '>dateiname' oder `>>dateiname'
** angegeben, dann wird die E-Mail an eine normale Textdatei angehngt.
*/
static void deliver_file (char *rorig, char *sender, char *outfile,
			  char *tmpfile)
{
    int is_xFile = (*outfile == ':');
    xFile out; FILE *in;
    char *msg, *msg1, line[1024];
    if (*outfile == ':' || *outfile == '>') { ++outfile; }
    if (!(out = xOpen (outfile, "wx", prefs.fperm))) {
	msg = subst ("unable to append message <MF> to file <OF>: <EM>!\n",
		     "<EM>", strerror (errno));
	msg1 = subst (msg, "<OF>", outfile);
	gfree (msg);
	msg = subst (msg1, "<MF>", tmpfile);
	gfree (msg1);
	deliver_error (rorig, sender, rorig, tmpfile, msg);
	gfree (msg);
	return;
    }
    if (!(in = fopen (tmpfile, "r"))) {
	msg1 = subst ("reading message file <MF> failed: <EM>!\n",
		      "<EM>", strerror (errno));
	msg = subst (msg1, "<MF>", tmpfile);
	gfree (msg1);
	deliver_error (rorig, sender, rorig, tmpfile, msg);
	gfree (msg);
	xClose (out); if (!is_xFile) { xNoIndex (outfile); }
	return;
    }
    while (fgets (line, 1024, in)) { xPuts (line, out); }
    fclose (in);
    xClose (out); if (!is_xFile) { xNoIndex (outfile); }
    return;
}

/* Liefere eine E-Mail an einen (potentiell nicht-lokalen) Empfnger aus;
** `vxdeliver' erkennt hierbei, ob ein Empfnger lokal ist daran, da die
** entsprechende Domain in der VxMail-Datenbank eingetragen ist; steht sie
** dort nicht, dann versucht `vxdeliver', die E-Mail an das in der
** VxMail-Konfiguration unter `mailcmd' angegebene Programm zu bergeben.
** In der hinter `mailcmd' angegebenen Kommandozeile knnen die Parameter
** <SENDER>, <RECIPIENT> und <SUBJECT> nebst den fr das jeweilige Programm
** sonst noch bentigten Argumenten bergeben werden; erstere werden durch
** die Absenderadresse, die Empfngeradresse und das in der E-Mail-Datei
** enthaltene `subject' ersetzt; von der E-Mail wird nur der Rumpf an das
** gestartete Programm bergeben. Da das Programm als `/bin/sh'-Kommando
** gestartet wird, mssen Absenderadresse, Empfngeradresse und `subject'
** qequotet werden; dazu werden alle im `subject' enthaltenen `"' in `\"'
** verwandelt; da weder Absenderadresse noch Empfngeradresse dieses Zeichen
** enthalten drfen, geschieht dir Umwandlung dort nicht.
** Anschlieend werden alle drei Angaben in Paare von `"' eingeschlossen und
** an der durch das betreffende Symbol festgelegten Stelle eingesetzt.
*/

static char *msgs[] = {
    "internal/operating system error",
    "smtp connection failed",
    "HELO command failed",
    "MAIL command failed",
    "RCPT command failed",
    "DATA command failed",
    "transmitting mail body failed",
    "QUIT command failed"
};

/* deliver_smtp (rorig, sender, recip_host, recipient, tmpfile)
**    sendet die E-Mail direkt an den Empfnger `recipient' auf dem Server
**    `recip_host' per SMTP weiter. Diese Variante der Auslieferung ist
**    fr Standverbindungen zu anderen Hosts gedacht, die die entsprechende
**    E-Mail-Domain lokal (oder virtuell) behandeln knnen.
*/
static void deliver_smtp (char *rorig,
			  char *sender,
			  char *recip_host,
			  char *recipient,
			  char *tmpfile)
{
    /* Eine Verbindung mit dem Zielserver soll sowohl per Namen als auch per
    ** Adresse erfolgen knnen, je nachdem, ob `recip_host' ohne oder mit `['
    ** beginnt, oder nicht.
    */
    char line[1024];
    int ev;
    HLSock hlc;
    off_t fp = skip_From (tmpfile);
    /* Verbindung zum `relay host' herstellen */
    if (!(hlc = HLConnect ("tcp", recip_host, "smtp"))) {
	sprintf (line, "mailcmd internal: connecting server `%s' failed!\n",
		 recip_host);
	deliver_error (rorig, sender, rorig, tmpfile, line);
	return;
    }
    /* E-Mail per SMTP bertragen */
    ev = smtp_send (hlc, sender, recipient, tmpfile, fp);
    /* Verbindung beenden */
    HLClose (hlc);
    /* Eventuell aufgetretene Fehler ausgeben */
    if (ev && ev < 8) {
	sprintf (line, "mailcmd internal: %s!\n", msgs[ev - 1]);
	deliver_error (rorig, sender, rorig, tmpfile, line);
	return;
    }
    /* Evtl. eine Warnung in die LOG-Datei schreiben */
    if (ev == 8) {
	vxlog (0, "message to `%s' may have been delivered incompletely!\n",
		  recipient);
    }
}

static void deliver_nonlocal (char *rorig, char *sender, char *addr,
			      char *tmpfile)
{
    if (domain_check (addr)) {
	/* Auslieferung an einen lokalen Empfnger */
	deliver_local (rorig, sender, addr, tmpfile);
    } else {
	/* Auslieferung an einen nicht-lokalen Empfnger */
	char *relayhost;
	/* Erst mal gucken, ob das berhaupt mglich ist. */
	if (!(relayhost = prefs.nonlocalto)) {
	    deliver_error (rorig, sender, addr, tmpfile,
			   "mail recipient not found\n");
	    return;
	}
	deliver_smtp (rorig, sender, relayhost, addr, tmpfile);
    }
}

/* deliver_bsmtp (sender, bsmtp_file, recipient, inputfile)
**     hngt die in `inputfile' stehende E-Mail an den Folder `bsmtp_file'
**     (VxMail-Folder) als Folge von SMTP-Kommandos (+ Daten) an.
**     Diese `bsmtp'-Variante soll eine vom SMTP-Server unabhngige Variante
**     des Mail-Relaying ermglichen; das sptere abliefern an den eigentlichen
**     Zielserver erfolgt dann durch Authentifikation mittels POP3, wobei dann
**     aber der Inhalt von `bsmtp_file' nicht per POP3, sondern per SMTP an die
**     Adresse, die aus der POP3-Verbindung hervorgeht. Der POP3-Server liefert
**     an dieser Stelle dann nur eine leere Liste von E-Mails, da er nur die
**     Authentifikation bernimmt.
*/
static void deliver_bsmtp (char *sender,
			   char *_batchfile,
			   char *recipient,
			   char *tmpfile)
{
    xFile bFx; FILE *ifp;
    int lasteol;
    char *batchfile, line[1024], *recip;
    off_t sofs = skip_From (tmpfile);
    vxlog (5, "Delivering via BSMTP ...\n");
    if (*_batchfile == '/') {
	batchfile = gdupl (__FILE__, __LINE__, _batchfile);
    } else {
	batchfile = gconc(__FILE__,__LINE__,prefs.listdir,"/",_batchfile,NULL);
    }
    if (!(ifp = fopen (tmpfile, "r"))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "failed to read file `%s'!", tmpfile);
    }
    fseek (ifp, sofs, SEEK_SET);
    if (!(bFx = xOpen (batchfile, "wx", prefs.fperm))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "failed to write to file `%s'!", batchfile);
    }
    xPuts ("MAIL FROM:<", bFx); xPuts (sender, bFx); xPuts (">\r\n", bFx);
    vxlog (5, "MAIL FROM:<%s>\\r\\n\n", sender);
    recip = gdupl (__FILE__, __LINE__, recipient);
    if (prefs.bsmtp_nodomain) {
	char *p = strchr (recip, '@');
	if (p) *p = '\0';
    }
    xPuts ("RCPT TO:<", bFx); xPuts (recip, bFx); xPuts (">\r\n", bFx);
    vxlog (5, "RCPT TO:<%s>\\r\\n\n", recip);
    gfree (recip);
    xPuts ("DATA\r\n", bFx);
    vxlog (5, "DATA\\r\\n\n");
    lasteol = 1;
    while (fgets (line, 1024, ifp)) {
	int eolf = cuteol (line);
	if (lasteol && *line == '.') {
	    xPuts (".", bFx);
	    vxlog (5, ".");
	}
	xPuts (line, bFx);
	vxlog (5, "%s", line);
	if (eolf) {
	    xPuts ("\r\n", bFx);
	    vxlog (5, "\\r\\n\n");
	}
	lasteol = eolf;
    }
    if (!lasteol) { xPuts ("\r\n", bFx); vxlog (5, "\\r\\n\n"); }
    xPuts (".\r\n", bFx); vxlog (5, ".\\r\\n\n");
    xClose (bFx);
    fclose (ifp);
    vxlog (5, "... succeeded.\n");
    gfree (batchfile);
}

static off_t skip_From (char *tmpfile)
{
    FILE *in;
    off_t fp = 0;
    char line[1024];
    if (!(in = fopen (tmpfile, "r"))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "reading message file `%s' failed", tmpfile);
    }
    if (fgets (line, 1024, in) && !strncmp (line, "From ", 5)) {
	fp += strlen (line);
	while (!cuteol (line)) {
	    if (!fgets (line, 1024, in)) break;
	    fp += strlen (line);
	}
    }
    fclose (in);
    return fp;
}

/* Erzeuge eine neue Zeichenkette durch Ersetzung aller vorkommenden
** Symbole `old' durch den Text `new'.
*/
static char *subst (char *text, char *old, char *new)
{
    char *res = (char *) 0, *p, *q;
    size_t ressz = 0, olen = strlen(old), nlen = strlen(new), rlen;
    q = text;
    if (!(p = strstr (q, old))) {
	return gdupl (__FILE__, __LINE__, text);
    }
    ressz = p - q + nlen + 1;
    res = galloc (__FILE__, __LINE__, ressz);
    memcpy (res, q, (p - q));
    memcpy (res + (p - q), new, nlen);
    res[ressz - 1] = '\0';
    q = p + olen;
    while ((p = strstr (q, old))) {
	res = gresize (__FILE__, __LINE__, res, ressz + (p - q) + nlen);
	memcpy (res+ressz-1, q, (p - q));
	memcpy (res+ressz-1+(p-q), new, nlen);
	ressz = ressz + (p-q) + nlen;
	res[ressz - 1] = '\0';
	q = p + olen;
    }
    rlen = strlen (q);
    res = gresize (__FILE__, __LINE__, res, ressz + rlen);
    memcpy (res + ressz - 1, q, rlen);
    ressz += rlen;
    res[ressz - 1] = '\0';
    return res;
}
