/* fcheck.c
**
** check if a mail should be delivered to a specified user.
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <sys/types.h>

#include "regex.h"
#include "tools.h"
#include "vxerror.h"

#include "userdb.h"

/* insert a filter for a specified domain */
static void ins_rex (regex_t **_relist, size_t *_relist_sz,
		     char *domain, char *strg)
{
#define relist (*_relist)
#define relist_sz (*_relist_sz)
    regex_t rx;
    int rec;
    if ((rec = regcomp (&rx, strg, REG_EXTENDED|REG_NOSUB|REG_ICASE))) {
	char errmsg[256];
	(void) regerror (rec, &rx, errmsg, 1024);
	errmsg[255] = '\0';
	vxlog (1, "WARNING! Unable to insert '%s'-filter `%s': %s\n",
		  domain, strg, errmsg);
	return;
    }
    if (!relist) {
	relist_sz = 1;
	relist = galloc (__FILE__, __LINE__, sizeof(regex_t));
    } else {
	++relist_sz;
	relist = gresize(__FILE__, __LINE__, relist, relist_sz*sizeof(regex_t));
    }
    memcpy (&relist[relist_sz - 1], &rx, sizeof (rx));
#undef relist_sz
#undef relist
}

typedef struct _nd_list *nd_list;

struct _nd_list {
    nd_list next;
    char *domain;
    int do_not_deliver;
};

static nd_list reject_list = (nd_list) 0;

/* Fge das Ergebnis des Filters in eine Liste ein, damit der Filter
** nicht jedesmal erneut ablaufen mu; Filter sind grundstzlich
** Domain-bezogen. Wenn fr eine Domain bereits festgestellt wurde, da an
** ihre Benutzer die betreffende E-Mail nicht bekommen sollen, dann mu letzt-
** endlich nur noch nachgeschaut werden, ob die E-Mail zugestellt werden soll
** oder nicht.
*/
static void nd_add (char *domain, int do_not_deliver)
{
    nd_list cc;
    char *p;
    for (cc = reject_list; cc; cc = cc->next) {
	if (!strcmp (cc->domain, domain)) { return; }
    }
    cc = galloc (__FILE__, __LINE__, sizeof(struct _nd_list)+strlen(domain)+1);
    p = (char *) cc + sizeof(struct _nd_list);
    cc->domain = p;
    strcpy (p, domain);
    cc->do_not_deliver = do_not_deliver;
}

/* Suche den Eintrag in der Liste fr die angegebene Domain. Das zugehrige
** Flag gibt darber Auskunft, ob die Mail zugestellt werden soll oder nicht.
*/
static nd_list nd_find (char *domain)
{
    nd_list cc;
    for (cc = reject_list; cc; cc = cc->next) {
	if (!strcmp (cc->domain, domain)) { return cc; }
    }
    return (nd_list) 0;
}

/* Lies die Filterdatei ein, bersetze enthaltenen angegebenen regulren
** Ausdrcke und speichere die bersetzten Ausdrcke in dem als Parameter
** bergebenen Variablenpaar (rx, rxsz).
*/
static int get_rex (char *domain, char *rxfile, regex_t **rx, size_t *rxsz)
{
    FILE *rxf;
    char line[1024];
    int skipeol = 0;
    size_t ll = 0;
    if (!(rxf = fopen (rxfile, "r"))) { return 1; }
    while (fgets (line, 1024, rxf)) {
	ll = strlen (line);
	skipeol = 1;
	if (!ll) continue;
	if (line[--ll] == '\r') {
	    line[ll] = '\0';
	    skipeol = 0;
	} else if (line[ll] == '\n') {
	    line[ll] = '\0';
	    if (ll > 0 && line[ll - 1] == '\r') { line[--ll] = '\0'; }
	    skipeol = 0;
	}
	if (ll && *line != '#') { ins_rex (rx, rxsz, domain, line); }
	if (skipeol) {
	    while (fgets (line, 1024, rxf)) {
		ll = strlen (line) - 1;
		if (line[ll] == '\r' || line[ll] == '\n') { break; }
	    }
	}
    }
    fclose (rxf);
    return 0;
}

/* fcheck (<e-mail addr>, <message file>)
**     stellt fest, ob die in <message file> enthaltene E-Mail an den durch
**     <e-mail addr> spezifizierten Adressaten zugestellt werden darf oder
**     nicht. Zugleich wird das dabei gefundene Ergebnis gespeichert, so da
**     beim nchsten Aufruf das Ergebnis von `fcheck' schneller vorgelegt
**     werden kann.
*/
int fcheck (char *em_addr, char *tmpfile)
{
    regex_t *rxlist = (regex_t *) 0;
    size_t rxlsz = 0;
    char *ffile = get_ffile (em_addr);
    FILE *tf;
    char *line, *p, *dp;
    size_t lsz, ll, nmatch = 1;
    regmatch_t pm[1];
    nd_list dc;
    int ix;

    if (!(dp = strchr (em_addr, '@'))) { return 1; }
    ++dp;
    if ((dc = nd_find (dp))) { return !dc->do_not_deliver; }
    if (!ffile) { nd_add (dp, 0); return 1; }
    if (get_rex (dp, ffile, &rxlist, &rxlsz)) {
	gfree (ffile); nd_add (dp, 0); return 1;
    }
    gfree (ffile);
    if (!(tf = fopen (tmpfile, "r"))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "unable to open temporary file `%s'", tmpfile);
    }
    lsz = 1024;
    line = galloc (__FILE__, __LINE__, lsz);
    while (fgets (line, lsz, tf)) {
	ll = strlen (line);
	p = line + ll - 1;
	while (*p != '\r' && *p != '\n') {
	    line = gresize (__FILE__, __LINE__, line, lsz + 1024);
	    p = line + ll;
	    if (!fgets (p, 1024, tf)) { break; }
	    lsz += 1024;
	    p += strlen (p) + 1;
	}
	for (ix = 0; ix < rxlsz; ++ix) {
	    if (regexec (&rxlist[ix], line, nmatch, pm, 0)) { goto NODELIV; }
	}
    }
    fclose (tf);
    nd_add (dp, 0);
    for (ix = 0; ix < rxlsz; ++ix) { regfree (&rxlist[ix]); }
    gfree (rxlist);
    return 1;
NODELIV:
    fclose (tf);
    nd_add (dp, 1);
    for (ix = 0; ix < rxlsz; ++ix) { regfree (&rxlist[ix]); }
    gfree (rxlist);
    return 0;
}
