/* userdb.c
 *
 * Datebase functions for vxmail
 */
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>

#include "prefs.h"
#include "dbif.h"
#include "vxerror.h"
#include "tools.h"
#include "sysauth.h"
#include "userdb.h"

static DBIF_FILE userdb = (DBIF_FILE) 0;

/* local function for initializing the userdb */
void init_userdb (void)
{
    if (userdb) { return; }
    dbif_reset (&userdb, prefs.userdb_name);
    if (!userdb) {
	vxerror_at (EX_NOINPUT, errno, __FILE__, __LINE__ - 7,
		    "user database `%s'", prefs.userdb_name);
    }
}

/* check for the existance of an e-mail domain */
int domain_check (char *em_addr)
{
    dbif_rec *rec;
    char *domain = strchr (em_addr, '@');
    if (domain) ++domain; else domain = em_addr;
    init_userdb ();
    rec = dbif_get (userdb, domain);
    if (rec) {
	int rt = dbif_rectype(rec);
	dbif_free (rec);
	return (rt == RT_DOMAIN);
    }
    return 0;
}

int xdomain_check (char *em_addr)
{
    dbif_rec *rec;
    char *domain = strchr (em_addr, '@');
    if (domain) ++domain; else domain = em_addr;
    init_userdb ();
    rec = dbif_get (userdb, domain);
    if (rec) {
	int rt = dbif_rectype(rec);
	dbif_free (rec);
	if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	    return 0;
	}
	return (rt == RT_DOMAIN ? 1 : -1);
    }
    return 0;
}

/* get_ffile (addr)
**     returns the name of a filter file for the domain, the user `addr' is
**     a member of - or NULL (if the domain doesn't have a filter).
*/
char *get_ffile (char *em_addr)
{
    dbif_rec *rec;
    char *p, *faddr, *ffile;
    if (!(p = strchr (em_addr, '@'))) { return NULL; }
    faddr = gconc (__FILE__, __LINE__, "::FILTER::", p, NULL);
    if (!(rec = dbif_get (userdb, faddr))) { gfree (faddr); return NULL; }
    if (dbif_rectype(rec) != RT_FILTER) {
	dbif_free (rec); gfree (faddr);
	return NULL;
    }
    ffile = gdupl (__FILE__, __LINE__, rec->folder);
    dbif_free (rec); gfree (faddr);
    return ffile;
}

/* get a user's folder;
** if the user entry is a link, then resolve this link first
*/

char *user_folder (char *em_addr, int *rtype)
{
    dbif_rec *rec;
    char *wfolder;
    size_t fll;
    int rt;
    char *p;
    init_userdb ();
    if (strcmp (em_addr, "::MAILSYS::")) {
	p = strchr (em_addr, '@');
	if (!p || !*++p) { return (char *) 0; }
    }
    rec = dbif_resolve (userdb, em_addr, NULL);
    if (!rec) {
	char *alladdr;
	if (!strcmp (em_addr, "::MAILSYS::")) { return NULL; }
	p = strchr (em_addr, '@');
	alladdr = gconc (__FILE__, __LINE__, "<OTHER>", p, NULL);
	p = strchr (alladdr, '@');
	rec = dbif_resolve (userdb, alladdr, NULL);
	gfree (alladdr);
	if (!rec) { return NULL; }
    }
    rt = dbif_rectype(rec);
    if (!((1 << rt) & (CC_USER|CC_ADMIN|CC_MAILSYS|CC_LLINK|CC_BSMTP))) {
	vxlog (1, "  wrong user type!\n");
	dbif_free (rec);
	return (char *) 0;
    }
    fll = strlen (rec->folder);
    wfolder = gdupl (__FILE__, __LINE__, rec->folder);
    *rtype = dbif_rectype (rec);
    dbif_free (rec);
    return wfolder;
}

char *user_rewrite (char *em_addr, int *rtype)
{
    dbif_rec *rec;
    char *new_user, *dompart, *wfolder, *domname;
    size_t nulen, uplen, fll;
    int rt;
    init_userdb ();
    dompart = strchr (em_addr, '@');
    /* `sendmail' doesn't set the `$h' macro in the case of the local host. */ 
    if (!dompart || !dompart[1]) {
	/* configurable domainname for the local host! */
	uplen = (dompart ? (dompart - em_addr) : strlen (em_addr));
	dompart = mailhost ();
	vxlog (5, "no domain found - using `%s' instead!\n", dompart);
    } else {
	uplen = dompart - em_addr;
	++dompart;
    }
    rec = dbif_resolve (userdb, dompart, &domname);
    if (!rec) { return NULL; }
    rt = dbif_rectype(rec);
    /* The result of the database access MUST be a domain entry! */
    if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) { dbif_free (rec); return NULL; }
    fll = strlen (domname);
    nulen = uplen + fll + 1;
    /* allocate enough memory for the new e-mail address */
    new_user = galloc (__FILE__, __LINE__, nulen + 1);
    /* concatenate the user part from the old e-mail address and the
    ** domain part (from the `folder' element of the domain link record
    ** to a new e-mail address
    */
    memcpy (new_user, em_addr, uplen);
    new_user[uplen] = '@';
    memcpy (new_user + uplen + 1, domname, fll);
    new_user[nulen] = '\0';
    /* get the folder path for the rewritten e-mail address */
    wfolder = user_folder (new_user, rtype);
    /* If the result is empty, check if the domain is a non-standard domain.
    ** If so, try again the `sysauth'-mechanism.
    */
    if (!wfolder && dbif_rectype(rec) == RT_NSDOMAIN && prefs.sysauth) {
	dbif_rec *rec1 = sysauth (rec, em_addr);
	wfolder = NULL;
	if (rec1) { wfolder = gdupl (__FILE__, __LINE__, rec1->folder); }
	dbif_free (rec1);
	*rtype = RT_NSUSER;
    }
    /* release the memory which was allocated for the domain link record */
    dbif_free (rec);
    free (domname);
    /* release the memory intermediately allocated for the rewritten
    ** e-mail address
    */
    gfree (new_user);
    return wfolder;
}

static dbif_rec *idrec = (dbif_rec *) 0;

char *mailhost (void)
{
    static char *mh = NULL;
    dbif_rec *rec;
    if (mh) return mh;
    init_userdb ();
    if (!idrec) { idrec = dbif_get (userdb, "::VXMAIL::"); }
    rec = idrec;
    if (!rec) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "database inconsistency! "
		    "This is a FATAL ERROR! Exiting!\n");
    }
    mh = gdupl (__FILE__, __LINE__, rec->passwd);
    return mh;
}

char *defadmin (void)
{
    static char *da = NULL;
    dbif_rec *rec;
    if (da) return da;
    init_userdb ();
    if (!idrec) { idrec = dbif_get (userdb, "::VXMAIL::"); }
    rec = idrec;
    if (!rec) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "database inconsistency! "
		    "This is a FATAL ERROR! Exiting!\n");
    }
    da = gdupl (__FILE__, __LINE__, rec->folder);
    return da;
}
