#include <stdio.h>
#include <stdlib.h>
#include <strings.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <errno.h>
#include <time.h>
#include <ctype.h>
#include <sys/param.h>
#include <sys/file.h>

#ifdef HAVE_ZLIB
#include <zlib.h>
#endif

#include "prefs.h"
#include "config.h"
#include "conf.h"
#include "vxerror.h"
#include "tools.h"
#include "xfile.h"
#include "version.h"
#include "vxopt.h"

#include "userdb.h"
#include "deliver.h"

void cleanup (void)
{
    char tmpfile[PATH_MAX + 1];
    sprintf (tmpfile, "/tmp/vxdeliver.%d", getpid ());
    unlink (tmpfile);
}

void sighandle (int *signum)
{
    cleanup ();
    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		"%s: caught signal number %d\n",
		progname, signum);
}

void usage (void)
{
    fprintf (stderr,
	     "\nUsage: %s [-f <configfile>] [-l <logfile>] sender-addr "
	     "email-addr ..."
	     "\n       %s -h"
	     "\n       %s -V\n"
	     "\n   -f <configfile>"
	     "\n       read configuration from file <configfile>"
	     "\n       (default: %s)"
	     "\n   -h"
	     "\n       show this help"
	     "\n   -V"
	     "\n       show the program's version number\n\n",
	     progname, progname, progname, CONFIGFILE);
    exit (EX_USAGE);
}


int main (int argc, char *argv[])
{
    int fd;
    FILE *fp = (FILE *) 0;
    char tmpfile[PATH_MAX + 1];
    char line[1024];
    int ix, c;
    char *sender = "", *logname = (char *) 0;
    char filler[PATH_MAX + 1];
    int ll = strlen (argv[0]);
    long sz;

    /* Settings: */
    vxopt (&argc, argv, usage);

    if (getuid () != geteuid () && geteuid () == 0) {
	vxerror (EX_USAGE, 0,
		 "Sorry, but this program cannot be run `setuid root'!\n"
		 "If you want to execute `%s' with root permissions, you "
		 "should start it\n"
		 "in a root environment!\n", progname);
    }
    sprintf (tmpfile, "/tmp/vxdeliver.%d", getpid ());
    memset (filler, '\0', PATH_MAX + 1);
    memset (filler, ' ', (ll > PATH_MAX ? PATH_MAX : ll));

    /* Option checking: */
    for (;;) {
	c = getopt (argc, argv, "l:v");
	if (c == -1) break;
	switch (c) {
	    case 'l':
		if (logname) {
		    fprintf (stderr, "ambigeous `-l'-option!\n");
		    break;
		}
		logname = gdupl (__FILE__, __LINE__, optarg);
		break;
	    case 'v':
		++loglevel;
		break;
	    default:
		vxerror_at (EX_USAGE, 0,
			    __FILE__, __LINE__ - 1,
			    "unknown option `%s'!\n"
			    "Use `%s -h' for help, please!\n",
			    argv[optind], progname);
	}
    }

    getconf ();

    /* User database initialisieren */
    init_userdb ();

    if (argc - optind < 2) {
	vxerror_at (EX_USAGE, 0,
		    __FILE__, __LINE__ - 2,
		    "you hav to specify at least one e-mail address!\n"
		    "Use `%s -h' for help, please!\n",
		    progname);
    }
    if (logname) {
	int fd = open (logname, O_CREAT|O_APPEND|O_WRONLY, 0600);
	if (fd < 0) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 2,
			"opening logfile `%s' failed", logname);
	}
	dup2 (fd, 2);
	close (fd);
    }

    /* Wenn ich erst hier userid und group id setze, dann kann die
    ** Konfiguration auch in einer Datei stehen, die nicht `vxuid'
    ** gehrt und fr `vxuid' gesperrt ist.
    */
    setuid (geteuid ());
    setgid (prefs.vxgid);
    seteuid (prefs.vxuid);

    /* Argument checking: */
    sender = argv[optind++];
    if (!strcmp (sender, "ENVIRON")) {
	if (!(sender = getenv ("SENDER"))) {
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
			"no sender address found!\n");
	}
    }

    /* install a `cleanup' function in the atexit protocol */
    if (atexit (cleanup)) {
	vxerror_at (EX_OSERR, errno,
		    __FILE__, __LINE__ - 2,
		    "atexit (cleanup) failed\n");
    }

    vxlog (1, "creating temporary file from message\n");
    /* write the `stdin' contents to a temporary file */
    if ((fd = open (tmpfile, O_WRONLY|O_EXCL|O_CREAT|O_TRUNC, 0640)) < 0) {
	vxerror_at (EX_CANTCREAT, errno,
		    __FILE__, __LINE__ - 2,
		    "writing tempfile `%s' failed", tmpfile);
    }
    if (!(fp = fdopen (fd, "w"))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "fdopen (\"%s\", ...) failed", tmpfile);
    }
    sz = 0l;
    while ((fgets (line, 1024, stdin))) {
	fputs (line, fp);
	sz += strlen (line);
    }
    fclose (fp);
    /* Check the size of the incoming mail
    **   - size checking is generally enabled and
    **   - if delivery size checking is enabled
    */
    if (prefs.check_size && prefs.dsizecheck && (sz > prefs.check_size)) {
	/* The Message is too long, so exit with an error message */
	/* I will try EX_DATAERR as exit code and an error message
	** containing the exact reason why `vxdeliver' terminates here!
	*/
	/* NEVER include the body of a message which is too long */
	prefs.biirm = 0;
	/* send the message to the system mail administrator */
	deliver_error (NULL, sender, "::MAILSYS::", tmpfile,
		       "maximum message size exceeded!\n");
	if (prefs.errors == ERRORS_ADMIN) {
	    exit (0);
	} else {
	    /* Return with a non-zero exit status */
	    vxerror_at (EX_DATAERR, 0, __FILE__, __LINE__,
			"Sorry but the message you've sent is (%ld bytes) "
			"too long!\n",
			(sz - prefs.check_size));
	}
    }
    for (ix = optind; ix < argc; ++ix) {
	char *ra;
	/* Hiermit versuche ich einigen Schwierigkeiten, die mir der
	** `sendmail' bereitet, aus dem Weg zu gehen.
	*/
	/* So, jetzt kann der Name des Folders bestimmt werden. */
	ra = argv[ix];
	/* Remove an enclosing angle-bracket pair */
	if (*ra == '<' && ra[strlen(ra) - 1] == '>') {
	    ++ra; ra[strlen(ra) - 1] = '\0';
	}
	deliver_local (NULL, sender, ra, tmpfile);
    }
    seteuid (getuid ());
    unlink (tmpfile);
    if (prefs.errors != ERRORS_ADMIN && delivery_errs()) {
	vxerror (EX_NOUSER, 0,
		 "message delivering failed for %d recipient%s!\n",
		 delivery_errs(),
		 (delivery_errs() == 1 ? "" : "s"));
    }
    return 0;
}
