#include <stdlib.h>
#include <strings.h>
#include <errno.h>
#include <stdarg.h>
#include <sys/utsname.h>
#include <sys/stat.h>

#define _PROCS_C
#include "procs.h"

#include "prefs.h"
#include "vxerror.h"
#include "dbif.h"
#include "tools.h"

/* ccwarn: write a non-critical error message to stderr and increase the
** error counter
*/
void ccwarn (ParserData *sd, char *fmt, ...)
{
    va_list ap;
    va_start (ap, fmt);
    fprintf (stderr,
	     "In file %s(%ld:%ld): WARNING! ",
	     sd->filename,
	     sd->lc,
	     sd->cl);
    vfprintf (stderr, fmt, ap);
    fprintf (stderr, "\n");
}

/* ccerror: write a critical error message to stdout and increase the
** error counter
*/
void ccerror (ParserData *sd, char *fmt, ...)
{
    va_list ap;
    va_start (ap, fmt);
    fprintf (stderr, "In file %s(%ld:%ld): ", sd->filename, sd->lc, sd->cl+1);
    vfprintf (stderr, fmt, ap);
    fprintf (stderr, "\n");
    ++sd->numerrs;
}

void yyerror (char *message)
{
    ++sd.lc;
    ccerror (&sd, "%s", message);
    --sd.lc;
}

void yyinit (char *filename, int tabsize)
{
    sd.adminname = (char *) 0;
    sd.domainname = (char *) 0;
    sd.folderdir = (char *) 0;
    sd.mailhost = (char *) 0;
    sd.dbf = (DBIF_FILE) 0;
    sd.tabsize = tabsize;
    sd.filename = filename;
    sd.state = 0;
    sd.newdb = 0;
    sd.nonstd = 0;
    sd.maxusers = 0l;
    sd.users = 0l;
}

void yyclose (ParserData *sd)
{
    dbif_close (sd->dbf);
}

static char *rtstr[] = {
    "user", "link", "nonstd. user", "admin user", "domain", "domain link",
    "nonstd. domain", "mailsys admin", "list link", "IPv4 translation",
    "IPv6 translation", "domain redirection", "database identity",
    "VXTS secret", "invalid user", "mail filter", "BSMTP-user"
};

static void ins_userrec (char *fn, int lc, ParserData *sd, char *ema,
			 char *pwd, char *fld, int rt, ...)
{
    va_list ap;
    int nusers, maxusers;
    dbif_rec *rec;
    if (rt == RT_DOMAIN || rt == RT_NSDOMAIN) {
	va_start (ap, rt);
	nusers = va_arg (ap, int);
	maxusers = va_arg (ap, int);
	va_end (ap);
	rec = dbif_mkrec (pwd, fld, rt, nusers, maxusers);
    } else {
	rec = dbif_mkrec (pwd, fld, rt);
    }
    if (!rec) { vxerror_at (EX_OSERR, errno, fn, lc, "(ins_userrec)"); }
    if (dbif_set (sd->dbf, ema, rec)) {
	vxerror_at (EX_SOFTWARE, 0, fn, lc,
		    "inserting %s `%s' into the database failed: %s!\n",
		    rtstr[rt], ema, dbif_errormsg ());
    }
    dbif_free (rec);
}

typedef void (*opf_t) (DBIF_FILE *dbf, char *dbname, ...);

static void opendb (DBIF_FILE *dbf, char *dbname, int mode, opf_t opf)
{
    unsigned int xdelay, seed;
    opf (dbf, dbname, mode);
    if (*dbf) return;
    if (max_sleep_delay == 0) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 2,
		    "opening user database `%s' failed: %s!\n",
		    dbname, dbif_errormsg ());
    }
    while (!*dbf && dbif_errno() == DBIF_NOWR) {
	seed = (unsigned int) time (NULL)
	     + (unsigned int) (getpid() << 5 || getppid());
	srand (seed);
	xdelay = (unsigned int) rand () % (max_sleep_delay + 1);
	sleep (xdelay);
	opf (dbf, dbname, mode);
    }
    if (!*dbf) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "opening user database `%s' failed: %s!\n",
		    dbname, dbif_errormsg ());
    }
}

/* Open/Create the database */
void InitDB (ParserData *sd)
{
    DBIF_FILE dbf;
    dbif_rec *rec;
    if (sd->dbf) return;
    if (sd->adminname) {
	opendb (&dbf, prefs.userdb_name, 0600, (opf_t) dbif_create);
	sd->dbf = dbf;
	ins_userrec (__FILE__, __LINE__, sd, "::VXMAIL::", sd->mailhost,
		     sd->adminname, RT_IDENTITY);
    } else {
	opendb (&dbf, prefs.userdb_name, 0600, (opf_t) dbif_rw);
	sd->dbf = dbf;
	rec = dbif_get (dbf, "::VXMAIL::");
	if (!rec) {
	    dbif_close (dbf);
	    vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 3,
			"invalid VxMail database `%s'!\n",
			prefs.userdb_name);
	}
	sd->adminname = gdupl (__FILE__, __LINE__, rec->folder);
	/* Setze den MAILHOST - jedoch nur dann, wenn berhaupt einer
	** spezifiziert wurde.
	*/
	if (sd->mailhost) {
	    ins_userrec (__FILE__, __LINE__, sd, "::VXMAIL::", sd->mailhost,
			 sd->adminname, RT_IDENTITY);
	}
	dbif_free (rec);
    }
}

/* Check for a new domain
**   This function returns `1' if the domain name of the domain to be inserted
**   is invalid or if an entry with the same name as the domain already exists
**   and `0' otherwise.
*/
int new_domain (ParserData *sd)
{
    dbif_rec *rec;
    InitDB (sd);
    /* A domainname NEVER can contain a `@' or a `:' */
    if (strchr (sd->domainname, '@') || strchr (sd->domainname, ':')) {
	ccerror (sd, "invalid name `%s' for a domain!", sd->domainname);
	return 1;
    }
    rec = dbif_get (sd->dbf, sd->domainname);
    if (rec) {
	ccerror (sd, "an entry with the same name as the domain `%s' already "
		     "exists!",
		     sd->domainname);
	dbif_free (rec);
	return 1;
    }
    return 0;
}

/* Read a domain entry from the database
**     If the database entry exists and is a domain, then insert it's data
**     and return `0'; write an error message and return `1' otherwise.
*/
int get_domain (ParserData *sd)
{
    dbif_rec *rec;
    int rt;
    InitDB (sd);
    rec = dbif_get (sd->dbf, sd->domainname);
    if (!rec) {
	ccerror (sd, "there is no domain `%s' in the database!",
		     sd->domainname);
	return 1;
    }
    rt = dbif_rectype(rec);
    if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	ccerror (sd, "`%s' is no domain (may be a link to a domain)!",
		     sd->domainname);
	dbif_free (rec);
	return 1;
    }
    if (sd->folderdir) { gfree (sd->folderdir); }
    /* The data we get are
    ** - the folderdir of the domain
    ** - the nonstd flag
    ** - the number of users and the maximum number of users allowed
    */
    sd->folderdir = gdupl (__FILE__, __LINE__, rec->folder);
    sd->nonstd = (rt == RT_NSDOMAIN);
    (void) dbif_getlim (rec, &sd->users, &sd->maxusers);
    dbif_free (rec);
    return 0;
}

/* insert a new user into the database or update an existing one */
int ins_user (ParserData *sd, char *user, char *pass, int urt)
{
    dbif_rec *rec;
    int newuser = 1, rt;
    char *em_addr, *folder;
    if (!sd->folderdir) {
	ccerror (sd, "no domain for USER `%s' specified!", user);
	return 0;
    }
    em_addr = gconc (__FILE__, __LINE__, user, "@", sd->domainname, NULL);
    folder = gconc (__FILE__, __LINE__, sd->folderdir, "/", user, NULL);
    rec = dbif_get (sd->dbf, em_addr);
    if (rec) {
	rt = dbif_rectype(rec);
	if (!((1 << rt) & (CC_USER|CC_ADMIN|CC_INVALID|CC_NSUSER))) {
	    ccerror (sd, "unable to overwrite %s `%s' with a user entry!\n",
			 rtstr[rt], em_addr);
	    dbif_free (rec); gfree (folder); gfree (em_addr);
	    return 0;
	}
	ccwarn (sd, "Overwriting %s `%s'!", rtstr[rt], em_addr);
	dbif_free (rec);
	newuser = 0;
    } else if (sd->maxusers && sd->users >= sd->maxusers) {
	ccerror (sd, "user limit for the domain `%s' exceeded!\n",
		     sd->domainname);
	return 0;
    }
    ins_userrec (__FILE__, __LINE__, sd, em_addr, pass, folder, urt);
    gfree (em_addr);
    return newuser;
}

void ins_iptrans (ParserData *sd, char *iptag, char *domain, int isIPv6)
{
    dbif_rec *rec;
    int rt = (isIPv6 ? RT_IP6TRANS : RT_IP4TRANS);
    InitDB (sd);
    rec = dbif_get (sd->dbf, domain);
    if (!rec) {
	if (strcmp (domain, sd->domainname)) {
	    ccerror (sd, "domain entry `%s' not found!", domain);
	    return;
	}
    } else {
	int rt = dbif_rectype (rec);
	dbif_free (rec);
	if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	    ccerror (sd, "`%s' is no valid domain (maybe a domain link)!",
			 domain);
	    return;
	}
    }
    rec = dbif_get (sd->dbf, iptag);
    if (rec) {
	ccwarn (sd, "Overwriting existing IP->domain translation `%s'!", iptag);
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__, sd, iptag, "", domain, rt);
}

void ins_domlink (ParserData *sd, char *link, char *domain)
{
    dbif_rec *rec;
    int rt;
    InitDB (sd);
    rec = dbif_get (sd->dbf, domain);
    if (!rec) { ccerror (sd, "domain `%s' not found!", domain); return; }
    rt = dbif_rectype(rec);
    if (rt != RT_DOMAIN && rt != RT_NSDOMAIN) {
	ccerror (sd, "target `%s' is no domain!", domain);
	dbif_free (rec);
	return;
    }
    dbif_free (rec);
    rec = dbif_get (sd->dbf, link);
    if (rec) {
	rt = dbif_rectype(rec);
	if (rt == RT_DOMLINK) {
	    ccwarn (sd, "Overwriting an existing domain link `%s'!", link);
	} else {
	    ccerror (sd, "`%s' already exists and is no domain link!", link);
	    dbif_free (rec);
	    return;
	}
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__, sd, link, "", domain, RT_DOMLINK);
}

/* insert a link to an existing(!) user into the database */
void ins_link (ParserData *sd, char *user, char *linkto)
{
    dbif_rec *rec;
    int rt;
    char *em_addr, *link_addr;
    if (!sd->folderdir) {
	ccerror (sd, "no domain for LINK `%s' specified!", user);
	return;
    }
    em_addr = gconc (__FILE__, __LINE__, user, "@", sd->domainname, NULL);
    if (strchr (linkto, '@')) {
	link_addr = gdupl (__FILE__, __LINE__, linkto);
    } else {
	link_addr = gconc(__FILE__,__LINE__, linkto, "@", sd->domainname, NULL);
    }
    rec = dbif_get (sd->dbf, link_addr);
    if (!rec) {
	ccerror (sd, "target `%s' for link `%s' not found!",
		     link_addr, em_addr);
	gfree (link_addr); gfree (em_addr);
	return;
    }
    rt = dbif_rectype(rec);
    if (rt != RT_USER && rt != RT_ADMIN && rt != RT_MAILSYS && rt != RT_BSMTP) {
	ccerror (sd, "invalid link target type!", link_addr);
	dbif_free (rec); gfree (link_addr); gfree (em_addr);
	return;
    }
    dbif_free (rec);
    rec = dbif_get (sd->dbf, em_addr);
    if (rec) {
	rt = dbif_rectype(rec);
	if (rt != RT_LINK) {
	    ccerror (sd, "unable to overwrite %s `%s' with a user link!\n",
			 rtstr[rt], em_addr);
	    dbif_free (rec);
	    gfree (link_addr);
	    gfree (em_addr);
	    return;
	}
	ccwarn (sd, "Overwriting an existing link `%s'!", em_addr);
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__, sd, em_addr, "", link_addr, RT_LINK);
    gfree (link_addr);
    gfree (em_addr);
}

void ins_bsmtp (ParserData *sd, char *user, char *pass, char *bsmtp)
{
    dbif_rec *rec;
    int rt;
    char *em_addr, *folder;
    if (!sd->folderdir) {
	ccerror (sd, "no domain for BSMTP-user `%s' specified!", user);
	return;
    }
    em_addr = gconc (__FILE__, __LINE__, user, "@", sd->domainname, NULL);
    rec = dbif_get (sd->dbf, em_addr);
    if (rec) {
	rt = dbif_rectype(rec);
	if (rt != RT_BSMTP) {
	    ccerror (sd, "unable to overwrite %s `%s' with a BSMTP-user!\n",
			 rtstr[rt], em_addr);
	    dbif_free (rec);
	    gfree (em_addr);
	    return;
	}
	ccwarn (sd, "Overwriting an existing BSMTP-user `%s'!", em_addr);
	dbif_free (rec);
    }
    if (bsmtp) {
	folder = gdupl (__FILE__, __LINE__, bsmtp);
    } else {
	folder = gconc (__FILE__, __LINE__, sd->folderdir, "/", user, NULL);
    }
    ins_userrec (__FILE__, __LINE__, sd, em_addr, pass, folder, RT_BSMTP);
    gfree (folder); gfree (em_addr);
}

/* insert an alias|a mailing list file|a mailing list command */
void ins_alp (ParserData *sd, char *user, char *alp)
{
    dbif_rec *rec;
    int rt;
    char *em_addr;
    if (!sd->folderdir) {
	ccerror (sd, "no domain for LLINK `%s' specified!", user);
	return;
    }
    em_addr = gconc (__FILE__, __LINE__, user, "@", sd->domainname, NULL);
    rec = dbif_get (sd->dbf, em_addr);
    if (rec) {
	rt = dbif_rectype(rec);
	if (rt != RT_LLINK) {
	    ccerror (sd, "unable to overwrite %s `%s' with a list link!\n",
			 rtstr[rt], em_addr);
	    dbif_free (rec);
	    gfree (em_addr);
	    return;
	}
	ccwarn (sd, "Overwriting an existing list link `%s'!", em_addr);
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__, sd, em_addr, "", alp, RT_LLINK);
    gfree (em_addr);
}

/* insert a mail delivery filter */
void ins_filter (ParserData *sd, char *filter)
{
    dbif_rec *rec;
    char *faddr;
    if (!sd->folderdir) {
	ccerror (sd, "no domain for filter specified!\n");
	return;
    }
    faddr = gconc (__FILE__, __LINE__, "::FILTER::", "@", sd->domainname, NULL);
    if ((rec = dbif_get (sd->dbf, faddr))) {
	if (dbif_rectype(rec) != RT_FILTER) {
	    ccerror (sd,
		    "database inconsistency: `::FILTER::' has the wrong "
		    "type!\n");
	    dbif_free (rec);
	    return;
	}
	ccwarn (sd, "Overwriting an existing filter entry!");
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__, sd, faddr, "", filter, RT_FILTER);
    gfree (faddr);
}

/* insert a domain redirection */
void ins_redirect (ParserData *sd,
			  char *domain,
			  char *password,
			  char *targethost)
{
    dbif_rec *rec;
    int rt;
    InitDB (sd);
    rec = dbif_get (sd->dbf, domain);
    if (rec) {
	rt = dbif_rectype(rec);
	if (rt != RT_REDIRECT) {
	    ccerror (sd, "unable to overwrite %s `%s' with a domain "
			 "redirection!\n",
			 rtstr[rt], domain);
	    dbif_free (rec);
	    return;
	}
	ccwarn (sd, "Overwriting an existing domain redirection `%s'!", domain);
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__,
		 sd, domain, password, targethost, RT_REDIRECT);
}

/* set/change the VXTS password */
void ins_vxts (ParserData *sd, char *vxtssecret)
{
    dbif_rec *rec;
    int rt;
    InitDB (sd);
    rec = dbif_get (sd->dbf, "::VXTS::");
    if (rec) {
	rt = dbif_rectype(rec);
	if (rt != RT_VXTS) {
	    ccerror (sd,
		    "database inconsistency: `::VXTS::' has the wrong type!\n");
	    dbif_free (rec);
	    return;
	}
	ccwarn (sd, "Changing the VXTS secret!");
	dbif_free (rec);
    }
    ins_userrec (__FILE__, __LINE__, sd, "::VXTS::", vxtssecret, "", RT_VXTS);
}

/* insert the mail administrator into the database */
void ins_mailadm (ParserData *sd, char *passwd, char *folder)
{
    InitDB (sd);
    ins_userrec (__FILE__, __LINE__, sd, "::MAILSYS::", passwd, folder,
		 RT_MAILSYS);
    return;
}

void insert_domain (ParserData *sd)
{
    int rt = (sd->nonstd ? RT_NSDOMAIN : RT_DOMAIN),
	nu = sd->users,
	mu = (sd->nonstd ? 0 : sd->maxusers);
    if (!sd->folderdir) { ccerror (sd, "NO DOMAIN specified!"); return; }
    ins_userrec (__FILE__, __LINE__, sd, sd->domainname, "", sd->folderdir, rt,
		 nu, mu);
    dbif_sync (sd->dbf);
}

/* data checking functions */

static char *unpfx = "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz._"
		     "0123456789";
static char *unchr = "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz._"
		     "-+0123456789";

int uncheck (ParserData *sd, char *uname, int adm)
{
    char *p;
    if (!*uname) { ccerror (sd, "invalid empty username!"); return 1; }
    if (sd->adminname && !adm && !strcmp (sd->adminname, uname)) {
	++sd->lc;
	ccerror (sd, "Sorry, but this user is already specified as "
		     "domain admin!");
	--sd->lc;
	return 1;
    }
    if (!strcmp (uname, "<OTHER>")) { return 0; }
    if (!strchr (unpfx, *uname)) {
	++sd->lc;
	ccerror (sd, "username must not start with `%c'!", *uname);
	--sd->lc;
	return 1;
    }
    p = uname;
    while (*++p) {
	if (!strchr (unchr, *p)) {
	    ++sd->lc;
	    ccerror (sd, "username must not contain a `%c'!", *p);
	    --sd->lc;
	    return 1;
	}
    }
    return 0;
}

static char *dnpfx = "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz"
		     "0123456789";
static char *dnchr = "AaBbCcDdEeFfGgHhIiJjKkLlMmNnOoPpQqRrSsTtUuVvWwXxYyZz._"
		     "-0123456789";

int dncheck (ParserData *sd, char *dname)
{
    char *p;
    if (!*dname) {
	++sd->lc;
	ccerror (sd, "invalid empty domainname!");
	--sd->lc;
	return 1;
    }
    if (!strchr (dnpfx, *dname)) {
	++sd->lc;
	ccerror (sd, "domainname must not start with `%c'!", *dname);
	--sd->lc;
	return 1;
    }
    p = dname;
    while (*++p) {
	if (!strchr (dnchr, *p)) {
	    ++sd->lc;
	    ccerror (sd, "domainname must not contain a `%c'!", *p);
	    --sd->lc;
	    return 1;
	}
	if (*p == '.' && p[-1] == '.') {
	    ++sd->lc;
	    ccerror (sd, "domainname must not contain two contigeous `.'!");
	    --sd->lc;
	    return 1;
	}
    }
    return 0;
}

int fdcheck (ParserData *sd, char *folderdir)
{
/*    char *p;*/
    if (!*folderdir) {
	++sd->lc;
	ccerror (sd, "empty path!");
	--sd->lc;
	return 1;
    }
    if (*folderdir != '/') {
	++sd->lc;
	ccerror (sd, "you must specify an absolute path!");
	--sd->lc;
	return 1;
    }
    return 0;
}

int fpcheck (ParserData *sd, char *folderpath)
{
    char *p;
    struct stat st;
    if (fdcheck (sd, folderpath)) { return 1; }
    p = folderpath + strlen (folderpath);
    if (*--p == '/') {
	++sd->lc;
	ccerror (sd, "folder path must not end in `/'!");
	--sd->lc;
	return 1;
    }
    if (stat (folderpath, &st)) { return 0; }
    if (!S_ISREG(st.st_mode)) {
	++sd->lc; ccwarn (sd, "`%s' is no regular file!"); --sd->lc;
    }
    return 0;
}

int hncheck (ParserData *sd, char *hostname)
{
    return dncheck (sd, hostname);
}

int macheck (ParserData *sd, char *maddr)
{
    char *p = strchr (maddr, '@');
    if (!p) { return uncheck (sd, maddr, 1); }
    *p = '\0';
    if (uncheck (sd, maddr, 1)) { *p = '@'; return 1; }
    *p = '@';
    return dncheck (sd, p + 1);
}
