/* bsmtp.c
**
** send each message in a BSMTP-folder via SMTP to a specified host.
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>

#include "prefs.h"
#include "hlsock.h"
#include "vxerror.h"
#include "tools.h"
#include "xfile.h"
#include "exithl.h"
#include "vxhost.h"

/* Entferne das Zeilenende (`\r', `\n' oder `\n') von einer Zeile und gib
** ein Flag zurck, das darber Auskunft gibt, ob die Zeile ein Zeilenende
** enthielt (also abgeschlossen war) oder nicht.
*/
static __inline__ int cuteol (char *line)
{
    char *p = line + strlen (line);
    if (p == line) { return 0; }
    if (*--p == '\r') {
        *p = '\0'; return 1;
    }
    if (*p == '\n') {
        *p = '\0';
        if (p != line && *--p == '\r') { *p = '\0'; }
        return 1;
    }
    return 0;
}

/* Lies smtliche Zeilen einer SMTP-Statusmeldung ein; gib aber nur den
** Status-Code ser ersten Zeile zurck.
*/
static int smtp_result (HLSock hlc)
{
    char line[1024], *p;
    int ecode = 0, stv;
    while (HLGets (line, 1024, hlc)) {
        cuteol (line);
        vxlog (4, "%s\n", line);
        stv = (int) strtol (line, &p, 10);
        if (!ecode) { ecode = stv; }
        if (*p == '\r' || *p == '\n' || *p == ' ') { break; }
    }
    return ecode;
}

/* RestoreFailed (from, to)
**     bertrage eine E-Mail aus einem VxMail-Folder in einen anderen;
**     diese Funktion heit wegen ihrer Verwendung so.
*/
static void RestoreFailed (xFile sav, xFile fld)
{
    char line[1024];
    while (xGets (line, 1024, sav) > 0) { xPuts (line, fld); }
    xSet (fld);
}

/* SaveFailed (from, &to)
**     ffnet - falls notwendig - einen temporren VxMail-Folder und schreibt
**     die gerade ausgewhlte E-Mail aus dem offenen VxMail-Folder `from'
**     hinein; falls der temporre VxMail-Folder neu geffnet wurde, befindet
**     sich das zugehrige `xFile'-handle hinter in dem Ein-/Ausgabeparameter
**     `to'; sonst ist dies das `xFile'-handle, an das die E-Mails gesendet
**     werden.
*/
static void SaveFailed(xFile fld, xFile *sav)
{
    xFile sfh;
    char tmpfile[60];
    sprintf (tmpfile, "/tmp/vxpop-bsmtp.%d\n", getpid ());
    if (!(sfh = *sav)) {
	if (!(sfh = xOpen (tmpfile, "ux", 0600))) {
	    vxlog (0, "opening save file `%s' failed: %s\n",
		      tmpfile, strerror (errno));
	    return;
	}
	*sav = sfh;
    }
    xReSet (fld);
    RestoreFailed (fld, sfh);
}

/* UnlinkSaved ()
**     lscht den temporren VxMail-Folder nach erfolgter Benutzung wieder.
*/
static void UnlinkSaved (void)
{
    char tmpfile[60];
    sprintf (tmpfile, "/tmp/vxpop-bsmtp.%d\n", getpid ());
    xUnlink (tmpfile);
}

/* bsmtp (target_host, bsmtp_folder, thishost)
**     ffnet eine SMTP-Verbindung zum Server `target_host' und bertrgt
**     smtliche in dem VxMail-Folder `bsmtp_folder' stehenden (schon im
**     SMTP-Protokollformat vorliegenden) E-Mails diesen.
**     Der Parameter `thishost' dient der HELLO-Authentifizierung.
**     E-Mails, die nicht erfolgreich zugestellt werden konnten, verbleiben
**     in `bsmtp_folder' (bzw. werden nach beendigung der bertragung dorthin
**     zurckgeschrieben); alle erfolgreich bertragenen E-Mails werden hinge-
**     gen aus `bsmtp_folder' entfernt.
*/
int bsmtp (const char *host, char *_folder, const char *mailhost)
{
    xFile bFx, mfx, sav = (xFile) 0;
    char line[1024], *folder;
    int ev, ix;
    HLSock hlc;
    vxlog (1, "Entering `bsmtp'\n");
    /* release the terminal?? */
    setpgid (getpid (), getpid ());
    fclose (stdin); fclose (stdout);
    if (*_folder == '/') {
	folder = gdupl (__FILE__, __LINE__, _folder);
    } else {
	folder = gconc (__FILE__, __LINE__, prefs.listdir, "/", _folder, NULL);
    }
    vxlog (3, "Folder = `%s'\n", folder);
    if (!(bFx = xOpen (folder, "rx"))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "opening folder `%s' failed");
    }
    vxlog (3, "Folder successfully opened!\n");
    vxlog (3, "Remote HOST = %s\n", host);
    if (!(hlc = HLConnect ("tcp", host, "smtp"))) {
	vxlog(0, "SMTP connection to `%s' failed: %s\n", host, strerror(errno));
	xClose (bFx);
	gfree (folder);
	return 1;
    }
    vxlog (3, "SMTP(tcp)-connection to host `%s' established.\n", host);
    if (HLhfInit (hlc)) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(HLhfInit)");
    }
    ev = smtp_result (hlc);
    if (ev / 100 != 2) {
	vxlog (0, "SMTP-server `%s' didn't send welcome message!\n", host);
	HLClose (hlc);
	xClose (bFx);
	gfree (folder);
	return 2;
    }
    vxlog (3, "HLSock initializes for high-level access.\n");
    HLPuts ("HELO ", hlc); HLPuts (mailhost, hlc); HLPuts ("\r\n", hlc);
    vxlog (5, ">> HELO %s\\r\\n\n", mailhost);
    ev = smtp_result (hlc);
    if (ev / 100 != 2) {
	vxlog (0, "SMTP HELO command failed!\n");
	HLClose (hlc);
	xClose (bFx);
	gfree (folder);
	return 3;
    }
    for (ix = 0; xSeek (bFx, (off_t) ix) >= 0; ++ix) {
	vxlog (5, ">> ");
	while (xGets (line, 1024, bFx) > 0) {
	    HLPuts (line, hlc);
	    vxlog (5, "%s\n", line);
	    if (cuteol (line)) {
		vxlog (5, ">> ");
		ev = smtp_result (hlc);
		if (ev / 100 != 3) { break; }
		if (ev / 100 != 2) {
		    SaveFailed (bFx, &sav);
		    vxlog (0, "SMTP transmission of message %d failed!\n", ix);
		    vxlog (0, "SMTP-Status is: %d.\n", ev);
		    goto MSG_DONE;
		}
	    }
	}
	while (xGets (line, 1024, bFx) > 0) {
	    HLPuts (line, hlc);
	    vxlog (5, "%s", line);
	    if (cuteol (line)) {
		vxlog (5, ">> ");
		if (!strcmp (line, ".")) { break; }
	    }
	}
	if (strcmp (line, ".")) {
	    if (!cuteol (line)) { HLPuts ("\r\n", hlc); vxlog (5, "\n"); }
	    HLPuts (".\r\n", hlc);
	    vxlog (5, ">> .\\r\\n\n");
	}
	ev = smtp_result (hlc);
	if (ev / 100 != 2) {
	    SaveFailed (bFx, &sav);
	    vxlog (0, "SMTP transmission of message %d failed!\n", ix);
	}
    MSG_DONE:
	;
    }
    HLPuts ("QUIT\r\n", hlc);
    vxlog (5, ">> QUIT\\r\\n\n");
    ev = smtp_result (hlc);
    if (ev / 100 != 2) {
	vxlog (0, "SMTP QUIT command failed "
		  "- nevertheless closing the connection!\n");
    }
    HLClose (hlc);
    vxlog (3, "SMTP-transfer completed -- cleaning up folder.\n");
    mfx = xOpen (folder, "u", prefs.fperm);
    if (sav) {
	xReSet (sav);
	for (ix = 0; xSeek (sav, (off_t) ix) >= 0; ++ix) {
	    RestoreFailed (sav, mfx);
	}
	xClose (sav);
	UnlinkSaved ();
	vxlog (1, "Copied %d failed messages back to folder.\n", ix);
    }
    xClose (mfx);
    xClose (bFx);
    gfree (folder);
    vxlog (1, "Leaving `bsmtp'!\n");
    return 0;
}
