/* vxlist.c
**
** a list datatype for the VxMail-system
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>

#include "vxerror.h"
#include "vxlist.h"
#include "tools.h"

/* I intend to implement this data type as list with pairs of references;
** this makes it possible to implement
**   - doubly linked lists
**   - FIFOs
*/

#define _vxl_head _vxl_first
#define _vxl_last _vxl_second
#define _vxl_next _vxl_first
#define _vxl_prev _vxl_second

/* these structures are invisible outside of this implementation
** of `vxlist'; so `vxlist' is an abstract data type.
*/
/* The `vxlist' type structure */
struct _vxlist {
    vxlist_type list_type;
    vxlist _vxl_first, _vxl_second;
    void *info;
};

/* The type structure for all iterators on `vxlist' */
struct _vxlist_iterator {
    vxlist_direction dir;
    vxlist base;
    vxlist current;
};

/* return the string representation of an element `lt' of the
** enumeration type `vxlist_type'.
*/
static char *repr (vxlist_type lt)
{
    switch (lt) {
	case LT_FIFO:	return "LT_FIFO";
	case LT_DOUBLE:	return "LT_DOUBLE";
	case LT_RING:	return "LT_RING";
    }
    return "";
}

/* insert a new element into the list `ent'; the new list is the
** result of this operation;
*/
vxlist vxlist_insert (vxlist ent, void *info, vxlist_type lt, int append)
{
    vxlist new;
    if (!ent) {
	if (!(ent = (vxlist) malloc (sizeof (struct _vxlist)))) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
			"(malloc)");
	}
	ent->list_type = lt;
	ent->info = info;
	switch (lt) {
	    case LT_FIFO:
		ent->_vxl_head = (vxlist) 0;
		ent->_vxl_last = ent;
		return ent;
	    case LT_DOUBLE:
		ent->_vxl_prev = ent->_vxl_next = (vxlist) 0;
		return ent;
	    case LT_RING:
		ent->_vxl_first = ent->_vxl_second = ent;
		return ent;
	}
    }
    if (ent->list_type != lt) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__ - 1,
		    "Sorry! Wrong list type `%s' (must be `%s')!\n",
		    repr(lt), repr (ent->list_type));
    }
    if (!(new = (vxlist) malloc (sizeof (struct _vxlist)))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "(malloc)");
    }
    new->list_type = ent->list_type;
    new->info = info;
    new->_vxl_first = new->_vxl_second = (vxlist) 0;
    switch (ent->list_type) {
	case LT_FIFO:
	    if (append) {
		ent->_vxl_last->_vxl_next = new;
		ent->_vxl_last = new;
	    } else {
		new->_vxl_last = ent->_vxl_last;
		new->_vxl_next = ent;
		ent->_vxl_last = (vxlist) 0;
		ent = new;
	    }
	    break;
	case LT_RING:
	    if (ent->_vxl_next == ent) {
		ent->_vxl_next = new;
		ent->_vxl_prev = new;
		new->_vxl_next = ent;
		new->_vxl_prev = ent;
		ent = new;
		break;
	    }
	case LT_DOUBLE:
	    if (append) {
		new->_vxl_prev = ent;
		new->_vxl_next = ent->_vxl_next;
		if (ent->_vxl_next) {
		    ent->_vxl_next->_vxl_prev = new;
		}
		ent->_vxl_next = new;
		ent = new;
	    } else {
		new->_vxl_next = ent;
		new->_vxl_prev = ent->_vxl_prev;
		if (ent->_vxl_prev) {
		    ent->_vxl_prev->_vxl_next = new;
		}
		ent->_vxl_prev = new;
		ent = new;
	    }
	    break;
    }
    return ent;
}

/* create a new list iterator for the list `ent'; the iterator
** has to advance in the specified direction (`dir') throug the list.
*/
vxlist_iterator vxlist_newiterator (vxlist ent, vxlist_direction dir)
{
    vxlist_iterator itr;
    if (!ent) { return (vxlist_iterator) 0; }
    if (ent->list_type == LT_FIFO && dir == VXLD_BACK) {
	vxerror_at (EX_SOFTWARE, 0, __FILE__, __LINE__,
		    "Sorry! Wrong direction for Iterator on a FIFO!\n");
    }
    if (!(itr = (vxlist_iterator) malloc(sizeof(struct _vxlist_iterator)))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 2,
		    "(malloc)");
    }
    itr->dir = dir;
    itr->base = ent;
    itr->current = (vxlist) 0;
    return itr;
}

/* (Re-)Set the a list iterator to it's initial state */
void *vxlist_first (vxlist_iterator *itr)
{
    (*itr)->current = (*itr)->base;
    return (*itr)->current->info;
}

/* advance through an iterator */
void *vxlist_advance (vxlist_iterator *itr)
{
    /* Advance at first */
    if ((*itr)->dir == VXLD_FORW) {
	(*itr)->current = (*itr)->current->_vxl_next;
    } else if ((*itr)->dir == VXLD_BACK) {
	(*itr)->current = (*itr)->current->_vxl_prev;
    }
    /* A RING list never contains a NULL pointer, so i must
    ** check for another termination condition:
    **    the identity of the `current' and the `base' element.
    */
    if ((*itr)->current->list_type == LT_RING) {
	if ((*itr)->current == (*itr)->base) {
	    return NULL;
	}
	return (*itr)->current->info;
    }
    /* DOUBLE and FIFO typed list can have NULL pointers - indicating
    ** the end (or beginning - in the case of DOUBLE) of the list
    */
    if ((*itr)->current) {
	return (*itr)->current->info;
    }
    return NULL;
}

/* Release the memory allocated for a list iterator and reset the
** iterator to an uninitialized state
*/
void vxlist_freeiterator (vxlist_iterator *itr)
{
    if (*itr) {
	free (*itr);
	*itr = (vxlist_iterator) 0;
    }
}
