/* conf.c
**
** Parsing/Loading configuration from a file
**
*/
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <errno.h>
#include <pwd.h>
#include <grp.h>
#include <sys/types.h>
#include <limits.h>

#include "config.h"
#include "prefs.h"
#include "vxerror.h"
#include "tools.h"

#define LINEMAX 1024

static int num_errs, linenum;

typedef struct {
    char *flagname;
    int flagval;
} flags_t;

static flags_t flags[] = {
    { "admin", ERRORS_ADMIN },
    { "system", ERRORS_SYSTEM },
    { "system+admin", ERRORS_ALL },
    { "admin+system", ERRORS_ALL },
    { NULL, 0 }
};

static flags_t boolflags[] = {
    { "no", 0 },
    { "yes", 1 },
    { NULL, 0 }
};

static flags_t scantype[] = {
    { "regex", 0 },
    { "prefix+empty", 1 },
    { "prefix+prefix", 2 },
    { NULL, 0 }
};

static flags_t dbtypes[] = {
    { "hash", 0 },
    { "btree", 1 },
    { "recno", 2 },
    { NULL, 0 }
};

static void set_string (char **svar, char *val);
static void set_flag (char *ivar, char *val, flags_t *flags);
static void set_fperm (int *fperm, char *val);
static void set_int (int *ivar, char *val);
static void set_size (long *lvar, char *val);
static void set_user (uid_t *ivar, char *val);
static void set_group (gid_t *ivar, char *val);
static void set_path (char **svar, char *val);
static void do_nothing (char **svar, char *val);
typedef void (*action_t) (void *tgt, char *val, void *acc);
static struct {
    char *token;
    action_t action;
    void *tgt;
    void *xtra;
} tokenlist[] = {
    { "conffile", (action_t) do_nothing, &prefs.config_name, NULL },
    { "userdb", (action_t) set_path, &prefs.userdb_name, NULL },
    { "listdir", (action_t) set_path, &prefs.listdir, NULL },
    { "nonlocalto", (action_t) set_string, &prefs.nonlocalto, NULL },
    { "errors", (action_t) set_flag, &prefs.errors, flags },
    { "fperm", (action_t) set_fperm, &prefs.fperm, NULL },
    { "biirm", (action_t) set_flag, &prefs.biirm, boolflags },
    { "maxclients", (action_t) set_int, &prefs.maxclients, NULL },
    { "service", (action_t) set_string, &prefs.service, NULL },
    { "vxuser", (action_t) set_user, &prefs.vxuid, NULL },
    { "vxgroup", (action_t) set_group, &prefs.vxgid, NULL },
    { "msg_firstline", (action_t) set_string, &prefs.msg_firstline, NULL },
    { "msg_lastline", (action_t) set_string, &prefs.msg_lastline, NULL },
    { "sysauth", (action_t) set_flag, &prefs.sysauth, boolflags },
    { "check_size", (action_t) set_size, &prefs.check_size, NULL },
    { "check_days", (action_t) set_int, &prefs.check_days, NULL },
    { "qmbox", (action_t) set_string, &prefs.qmbox, NULL },
    { "redir_info", (action_t) set_string, &prefs.redir_info, NULL },
    { "dsizecheck", (action_t) set_flag, &prefs.dsizecheck, boolflags },
    { "compression", (action_t) set_flag, &prefs.compression, boolflags },
    { "dbtype", (action_t) set_flag, &prefs.dbtype, dbtypes },
    { "msg_scantype", (action_t) set_flag, &prefs.msg_scantype, scantype },
    { "lcuser", (action_t) set_flag, &prefs.lcuser, boolflags },
    { "lcdomain", (action_t) set_flag, &prefs.lcdomain, boolflags },
    { "syswithidx", (action_t) set_flag, &prefs.syswithidx, boolflags },
    { "maxlevel", (action_t) set_int, &prefs.maxlevel, NULL },
    { "rejverb", (action_t) set_flag, &prefs.rejverb, boolflags },
    { "dracauth", (action_t) set_flag, &prefs.dracauth, boolflags },
    { "bsmtp_nodomain", (action_t) set_flag, &prefs.bsmtp_nodomain, boolflags },
    { NULL, (action_t) 0, NULL }
};

static void cferror (char *message)
{
    vxerror (0, 0, "%s(%d): %s\n", prefs.config_name, linenum, message);
    ++num_errs;
}

static void tk_setval (char *tok, char *val)
{
    int ix;
    for (ix = 0; tokenlist[ix].token; ++ix) {
	if (!strcmp (tok, tokenlist[ix].token)) {
	    tokenlist[ix].action (tokenlist[ix].tgt, val, tokenlist[ix].xtra);
	    return;
	}
    }
    cferror ("unknown token!\n");
}

char *strskp (char *inp, char *accept)
{
    while (*inp && strchr (accept, *inp)) {
	++inp;
    }
    return inp;
}

void getconf (void)
{
    char line[LINEMAX + 1];
    FILE *conf = fopen (prefs.config_name, "r");
    num_errs = 0;
    linenum = 0;
    if (!conf) {
	vxerror_at (EX_NOINPUT, errno,
		    __FILE__, __LINE__ - 3,
		    "failed to open configuration file `%s'",
		    prefs.config_name);
    }
    while (memset (line, '\0', LINEMAX + 1), fgets (line, LINEMAX + 1, conf)) {
	char *p, *q, *v;
	++linenum;
	if (line[strlen (line) - 1] != '\n') {
	    cferror ("line too long!\n");
	}
	p = line + strlen (line);
	if (p != line) {
	    if (*--p == '\r') {
		*p = '\0';		/* MacOS `end of line' */
	    } else if (*p == '\n') {
		*p = '\0';		/* Unix or MSDOS `end of line' */
		if (p != line && *--p == '\r') {
		    *p = '\0';		/* MSDOS `end of line' */
		}
	    }
	}
	p = strskp (line, " \t");
	if (!*p || *p == '#') {
	    continue;
	}
	q = p;
	p = strpbrk (q, " \t");
	if (!p) { p = q + strlen (q); } else { *p++ = '\0'; }
	p = strskp (p, " \t");
	v = p;
	tk_setval (q, v);
    }
    fclose (conf);
    if (!strncmp (prefs.userdb_name, "<LIST>", 6)) {
	/* concatenate the `vxworkdir' with the `userdb' in order to get a
	** complete pathname to the userdb
	** i assume that
	** - either `listdir' ends in a `/' or
	** - `userdb_name' is of the form `<LIST>/xyz'.
	*/
	char *newudb, *p = prefs.userdb_name + 6;
	if (!(newudb = concat (prefs.listdir, p, NULL))) {
	    vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1, "(concat)");
	}
	gfree (prefs.userdb_name);
	prefs.userdb_name = newudb;
    }
    if (num_errs) {
	vxerror_at (EX_SOFTWARE, 0,
		    __FILE__, __LINE__ - 2,
		    "Sorry, but there are syntax errors in the "
		    "configuration file!\n");
    }
}

static void set_string (char **svar, char *val)
{
    *svar = gdupl (__FILE__, __LINE__, val);
}

static void set_flag (char *ivar, char *val, flags_t *flags)
{
    int ix;
    for (ix = 0; flags[ix].flagname; ++ix) {
	if (!strcmp (flags[ix].flagname, val)) {
	    *ivar = flags[ix].flagval;
	    return;
	}
    }
    cferror ("invalid flag value!");
}

static void set_fperm (int *fperm, char *val)
{
    char *ep;
    int perm = (int) strtol (val, &ep, 8);
    if (ep && (ep == val || *ep)) { cferror ("invalid permission mask!"); }
    *fperm = perm;
}

static void set_int (int *ivar, char *val)
{
    char *ep;
    int res = strtol (val, &ep, 0);
    if (ep && (ep == val || *ep)) { cferror ("invalid integer value!"); }
    *ivar = res;
}

static void set_size (long *lvar, char *val)
{
    char *ep;
    long res = strtol (val, &ep, 0);
    if (ep && (ep == val || *ep)) {
	if (strchr ("Kk", *ep) && !ep[1]) {
	    res <<= 10;
	} else if (strchr ("Mm", *ep) && !ep[1]) {
	    res <<= 20;
	} else {
	    cferror ("invalid size value!");
	}
    }
    *lvar = res;
}

static void set_user (uid_t *ivar, char *val)
{
    char *p;
    long lval = strtol (val, &p, 0);
    static struct passwd *pw;
    if (*p || p == val) {
	/* try to get the uid corresponding to the user */
	if (!(pw = getpwnam (val))) { cferror ("unknown user!"); }
	*ivar = pw->pw_uid;
	return;
    }
    if (!(pw = getpwuid ((uid_t) lval))) { cferror ("unknown user!"); }
    *ivar = pw->pw_uid;
}

static void set_group (gid_t *ivar, char *val)
{
    char *p;
    long lval = strtol (val, &p, 0);
    static struct group *gr;
    if (*p || p == val) {
	/* try to get the uid corresponding to the user */
	if (!(gr = getgrnam (val))) { cferror ("unknown group!"); }
	*ivar = gr->gr_gid;
	return;
    }
    if (!(gr = getgrgid ((gid_t) lval))) { cferror ("unknown group!"); }
    *ivar = gr->gr_gid;
}

static void set_path (char **svar, char *val)
{
    char *res, *p, *pval, *q, *r;
    struct passwd *pw = getpwuid (geteuid ());
    char cwd[PATH_MAX];
    size_t reslen = strlen (val) + 1, ll;
    if (!getcwd (cwd, PATH_MAX)) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "i couldn't get the current working directory");
    }
    pval = val;
    do {
	p = strstr (pval, "<HOME>");
	q = strstr (pval, "<CWD>");
	if (p || (p && q && (p - q < 0))) {
	    reslen += strlen (pw->pw_dir);
	    reslen -= strlen ("<HOME>");
	    pval = p + strlen ("<HOME>");
	} else if (q) {
	    reslen += strlen (cwd);
	    reslen -= strlen ("<CWD>");
	    pval = q + strlen ("<CWD>");
	}
    } while (p || q);
    if (!(res = (char *) malloc (reslen))) {
	vxerror_at (EX_OSERR, errno, __FILE__, __LINE__ - 1,
		    "(malloc)");
    }
    r = res;
    pval = val;
    do {
	p = strstr (pval, "<HOME>");
	q = strstr (pval, "<CWD>");
	if (p || (p && q && (p - q < 0))) {
	    memcpy (r, pval, pval - p);
	    r += (pval - p);
	    ll = strlen (pw->pw_dir);
	    memcpy (r, pw->pw_dir, ll);
	    r += ll;
	    *r = '\0';
	    pval = p + strlen ("<HOME>");
	} else if (q) {
	    memcpy (r, pval, pval - q);
	    r += (pval - q);
	    ll = strlen (cwd);
	    memcpy (r, cwd, ll);
	    r += ll;
	    *r = '\0';
	    pval = q + strlen ("<CWD>");
	}
    } while (p || q);
    ll = strlen (pval);
    memcpy (r, pval, ll);
    r[ll] = '\0';
    *svar = res;
}

static void do_nothing (char **svar, char *val)
{
    ;
}

/* Conversion part; required for `vxgetconf' */

typedef char *(*cvproc_t) (void *cval, flags_t *flags);

static char *convstr (char **_cval);
static char *convoct (int *_cval);
static char *convflag (int *_cval, flags_t *flags);
static char *convint (int *_cval);
static char *convuid (uid_t *_cval);
static char *convgid (gid_t *_cval);
static char *convsize (size_t *_cval);

static struct {
    action_t action;
    char *type;
    cvproc_t conv;
} actypes[] = {
    { (action_t) do_nothing, "unsetable string", (cvproc_t) convstr },
    { (action_t) set_string, "string", (cvproc_t) convstr },
    { (action_t) set_path, "path", (cvproc_t) convstr },
    { (action_t) set_fperm, "permissions", (cvproc_t) convoct },
    { (action_t) set_flag, "flag", (cvproc_t) convflag },
    { (action_t) set_int, "integer", (cvproc_t) convint },
    { (action_t) set_user, "user", (cvproc_t) convuid },
    { (action_t) set_group, "group", (cvproc_t) convgid },
    { (action_t) set_size, "size", (cvproc_t) convsize },
    { (action_t) 0, NULL }
};

int confitem (int ix, char **name, char **type, char **value)
{
    int mx, nx;
    for (mx = 0; tokenlist[mx].token; ++mx);
    if (ix < mx) {
	for (nx = 0; actypes[nx].action; ++nx) {
	    if (tokenlist[ix].action == actypes[nx].action) {
		*name = gdupl (__FILE__, __LINE__, tokenlist[ix].token);
		if (type) {
		    *type = gdupl (__FILE__, __LINE__, actypes[nx].type);
		}
		*value = actypes[nx].conv (tokenlist[ix].tgt,
					   (flags_t *) tokenlist[ix].xtra);
		return 1;
	    }
	}
    }
    return 0;
}

static char *convstr (char **_cval)
{
    if (!*_cval) { return NULL; }
    return gdupl (__FILE__, __LINE__, *_cval);
}

static char *convoct (int *_cval)
{
    char buf[60], *bufp = buf;
    sprintf (buf, "%04o", *_cval);
    return convstr (&bufp);
}

static char *convflag (int *_cval, flags_t *flags)
{
    int ix, fv = *_cval;
    for (ix = 0; flags[ix].flagname; ++ix) {
	if (flags[ix].flagval == fv) {
	    return convstr (&flags[ix].flagname);
	}
    }
    return convstr (&flags[0].flagname);
}

static char *convint (int *_cval)
{
    char buf[60], *bufp = buf;
    sprintf (buf, "%d", *_cval);
    return convstr (&bufp);
}

static char *convuid (uid_t *_cval)
{
    struct passwd *pw = getpwuid (*_cval);
    if (!pw) { return NULL; }
    return convstr (&pw->pw_name);
}

static char *convgid (gid_t *_cval)
{
    struct group *gr = getgrgid (*_cval);
    if (!gr) { return NULL; }
    return convstr (&gr->gr_name);
}

static char *convsize (size_t *_cval)
{
    char buf[60], *bufp = buf;
    size_t sz = *_cval;
    if (!(sz & 0xfffff)) {
	sprintf (buf, "%ldM", (long) (sz >> 20));
    } else if (!(sz & 0x3ff)) {
	sprintf (buf, "%ldK", (long) (sz >> 10));
    } else {
	sprintf (buf, "%ld", (long) (sz));
    }
    return convstr (&bufp);
}
