/* dbif.c
**
** Database Interface for the user/admin databases
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>

#include "dbif_priv.h"
#include "config.h"
#include "vxerror.h"
#include "tools.h"
#include "ifconf.h"
#include "prefs.h"

/* In den Flags gibt es einen Eintrag `admin'.
** Dieser Eintrag erlaubt es dem Benutzer, der ber dieses Flag verfgt, neue
** Benutzer in seiner jeweiligen Domain zu anzulegen, bzw. Benutzer aus seiner
** Domain zu entfernen.
** Der Eintrag `system' soll es einem (NUR EINEM) Benutzer erlauben, Benutzer
** in jeder Domain anzulegen, bzw. aus ihr zu entfernen.
*/

/* Die User-Datenbank besteht aus den Schlsseln (den E-Mail-Adressen) und
** den diesen Schlsseln zugeordneten Daten, die wie folgt angeordnet sind:
**	Offset		Bedeutung
**	 0		Lnge des Passwortes (NPW)
**	 4		Lnge des Pfadnamens des Mail-Folders (NFP)
**	 8		Lnge des Flag-Feldes (NFL)
**	12		Passwort
**	12+NPW		Pfadname des Mail-Folders
**	12+NPW+NFP	Flag-Feld
**	12+NPW+NFP+NFL	Ende des Datensatzes
*/

static DBTYPE dt[] = { DB_HASH, DB_BTREE, DB_RECNO };

/* ffne eine Datenbank im Nur-Lese-Modus */
void dbif_reset (DBIF_FILE *db, char *dbname)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = O_RDONLY;
	res->current.data = NULL;
	res->current.size = 0;
	res->dbf = dbopen (dbname, res->rw, 0644, dt[prefs.dbtype], NULL);
	if (!res->dbf) { free (res); res = (DBIF_FILE) 0; }
    }
    *db = res;
}

/* Erzeuge eine Datenbank neu */
void dbif_create (DBIF_FILE *db, char *dbname, int mode)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = O_CREAT|O_TRUNC|O_RDWR;
	res->current.data = NULL;
	res->current.size = 0;
	res->dbf = dbopen (dbname, res->rw, mode, dt[prefs.dbtype], NULL);
	if (!res->dbf) { free (res); res = (DBIF_FILE) 0; }
    }
    *db = res;
}

/* ffne eine Datenbank sowohl fr Schreib- als auch fr Lesebetrieb */
void dbif_rw (DBIF_FILE *db, char *dbname)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = O_RDWR;
	res->current.data = NULL;
	res->current.size = 0;
	res->dbf = dbopen (dbname, res->rw, 0644, dt[prefs.dbtype], NULL);
	if (!res->dbf) { free (res); res = (DBIF_FILE) 0; }
    }
    *db = res;
}

/* berprfe, ob eine Datenbank geffnet ist! */
/*int dbif_isopen (DBIF_FILE db)
**{
**    if (db->dbf != (DB *) 0) {
**	if (db->rw) {
**	    return -1;
**	} else {
**	    return 1;
**	}
**    }
**    return 0;
**}
*/

/* Schliee eine geffnete Datenbank! */
void dbif_close (DBIF_FILE db)
{
    if (db) {
	if (db->rw) { db->dbf->sync (db->dbf, 0); }
	db->dbf->close (db->dbf);
	free (db);
    }
}

/* Synchronisiere eine zum Schreiben geffnete Datenbank */
void dbif_sync (DBIF_FILE db)
{
    if (db->rw) {
	db->dbf->sync (db->dbf, 0);
    }
}

/* Lesen eines Benutzereintrages; im Gegensatz zu `dbif_resolve' wird
** dieser Eintrag - auch in dem Fall, da er ein gesetztes `link'-Flag
** enthlt - nicht weiter aufgelst. Diese Routine ermglicht daher
** - zusammen mit `dbif_set' - eine Administration dieses Eintrages.
*/
dbif_rec *dbif_get (DBIF_FILE db, char *em_addr)
{
    DBT data, key;
    dbif_rec *rec = (dbif_rec *) 0;
    key.size = strlen(em_addr); key.data = em_addr;
    db->dbf->get(db->dbf, &key, &data, 0);
    if (!data.data) { return rec; }
    rec = dbif_mkreca (data.data);
    free (data.data);
    return rec;
}

/* Lies einen Benutzer aus der Datenbank; hat der Eintrag das Flag `link',
** dann enthlt das `Folder'-Element statt eines Dateinamens einen Verweis auf
** einen weiteren Benutzereintrag und der `Password'-Eintrag ist ohne Bedeutung;
** in diesem Fall wird der Verweis auf den anderen Benutzer aus dem
** `Folder'-Element gelesen und zu diesem verzweigt. Das ganze wird solange
** fortgesetzt, bis ein Eintrag ohne gesetztes `link'-Flag gefunden wird.
** Dieser wird dann zurckgegeben.
** Auf diese Art werden einfache `alias'-Mechanismen ermglicht.
*/
dbif_rec *dbif_resolve (DBIF_FILE db, char *em_addr, char **res_addr)
{
    dbif_rec *rec;
    int rtype;
    char *newaddr;
    if (!(newaddr = malloc (strlen (em_addr) + 1))) { return (dbif_rec *) 0; }
    strcpy (newaddr, em_addr);
    /* Hole einen Datensatz */
    rec = dbif_get (db, newaddr);
    /* Solange ein passender Datensatz vorhanden ist */
    while (rec) {
	/* Hole den Typ des Datensatzes */
	rtype = dbif_rectype(rec);
	/* Wenn dieser Typ weder ein `link' noch ein `domain link' bezeichnet */
	if (rtype != RT_LINK && rtype != RT_DOMLINK) {
	    /* Wenn der Datensatz einen ungltigen Benutzer bezeichnet,
	    ** dann MUSS das Ergebnis NULL sein; gib in diesem Fall auch
	    ** den (mglichen) Schlssel frei.
	    */
	    if (rtype == RT_INVALID) {
		free(newaddr); dbif_free(rec); return (dbif_rec *) 0;
	    }
	    /* Wenn eine Ergebnisvariable fr den Schlssel angegeben wurde,
	    ** dann schreibe diesen dort hinein; sonst gib den fr den
	    ** Schlssel reservierten Speicherplatz frei
	    */
	    if (res_addr) { *res_addr = newaddr; } else { free (newaddr); }
	    /* Der aktuelle Datensatz ist das Ergebnis */
	    return rec;
	}
	/* Der Datensatz war ein `link', oder ein `domain link'; versuche es
	** mit dem nchsten Datensatz in der Kette
	*/
	free (newaddr);
	if (!(newaddr = malloc (strlen (rec->folder) + 1))) {
	    free (rec); rec = (dbif_rec *) 0; return rec;
	}
	strcpy (newaddr, rec->folder);
	free (rec); rec = dbif_get (db, newaddr);
    }
    free (newaddr);
    return rec;
}

/* Fge einen neuen Benutzerdatensatz ein oder ersetze einen vorhandenen. */
static int dbif_insrpl (DBIF_FILE db, char *em_addr, dbif_rec *rec, int stflag)
{
    size_t pwl = strlen (rec->passwd),
	   fll = strlen (rec->folder),
	   dsize = pwl + fll + 1 + 2 * sizeof(int);
    char *p;
    DBT key, data;
    int rtype = dbif_rectype(rec), nusers, maxusers, res;
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	dsize += 2 * sizeof(int);
    }
    if (!(p = (char *) malloc (dsize))) { return 1; }
    data.size = dsize; data.data = p;
    set_int ((unsigned char *) p, pwl); p += sizeof (int);
    set_int ((unsigned char *) p, fll); p += sizeof (int);
    memcpy (p, rec->passwd, pwl); p += pwl;
    memcpy (p, rec->folder, fll); p += fll;
    *p++ = (char) dbif_rectype(rec);
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	dbif_getlim (rec, &nusers, &maxusers);
	set_int ((unsigned char *) p, nusers); p += sizeof(int);
	set_int ((unsigned char *) p, maxusers);
    }
    key.size = strlen (em_addr); key.data = em_addr;
    res = db->dbf->put (db->dbf, &key, &data, stflag);
    free (data.data);
    return res;
}

int dbif_insert (DBIF_FILE db, char *em_addr, dbif_rec *rec)
{
    return dbif_insrpl (db, em_addr, rec, R_NOOVERWRITE);
}

int dbif_set (DBIF_FILE db, char *em_addr, dbif_rec *rec)
{
    if (dbif_insrpl (db, em_addr, rec, R_NOOVERWRITE)) {
	return dbif_insrpl (db, em_addr, rec, R_CURSOR);
    }
    return 0;
}

const char *dbif_errormsg (void)
{
    return strerror (errno);
}

int dbif_del (DBIF_FILE db, char *em_addr)
{
    DBT key;
    key.size = strlen (em_addr);
    key.data = em_addr;
    return db->dbf->del (db->dbf, &key, 0);
}
