/* dbif.c
**
** Database Interface for the user/admin databases
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#define _DBIF_C
#include "dbif.h"
#include "dbif_priv.h"
#include "config.h"
#include "vxerror.h"
#include "tools.h"
#include "ifconf.h"

/* In den Flags gibt es einen Eintrag `admin'.
** Dieser Eintrag erlaubt es dem Benutzer, der ber dieses Flag verfgt, neue
** Benutzer in seiner jeweiligen Domain zu anzulegen, bzw. Benutzer aus seiner
** Domain zu entfernen.
** Der Eintrag `system' soll es einem (NUR EINEM) Benutzer erlauben, Benutzer
** in jeder Domain anzulegen, bzw. aus ihr zu entfernen.
*/

static struct {
    char *fl_name;
    int fl_offset;
} flags[] = {
    { "link", 0 },
    { "admin", 1 },
    { "system", 2 },
    { "mailadmin", 3 },
    { "domain", 4 },
    { "ip4trans", 5 },
    { "ip6trans", 6 },
    { "identity", 7 },
    { "nonstandard", 8 },
    { "invalid", 9 },
    { NULL, 0 }
};

/* Vorwrts-Deklarationen fr die Hilfsfunktionen */
int is_member (char *userflags, char *flagname);
static int get_length (unsigned char *dvec);
void set_member (char *userflags, char *flagname);
static void set_length (unsigned char *dvec, int p);

/* Die User-Datenbank besteht aus den Schlsseln (den E-Mail-Adressen) und
** den diesen Schlsseln zugeordneten Daten, die wie folgt angeordnet sind:
**	Offset		Bedeutung
**	 0		Lnge des Passwortes (NPW)
**	 4		Lnge des Pfadnamens des Mail-Folders (NFP)
**	 8		Lnge des Flag-Feldes (NFL)
**	12		Passwort
**	12+NPW		Pfadname des Mail-Folders
**	12+NPW+NFP	Flag-Feld
**	12+NPW+NFP+NFL	Ende des Datensatzes
*/

static void dberr (char *message)
{
    vxerror (EX_SOFTWARE, 0, "database internal error: %s!\n",
			     gdbm_strerror (gdbm_errno));
}

/* ffne eine Datenbank im Nur-Lese-Modus */
void dbif_reset (DBIF_FILE *db, char *dbname)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = GDBM_READER;
	res->current.dptr = NULL;
	res->current.dsize = 0;
	res->dbf = gdbm_open (dbname, CACHESIZE, res->rw, 0644, dberr);
	if (!res->dbf) { free (res); res = (DBIF_FILE) 0; }
    }
    *db = res;
}

/* Erzeuge eine Datenbank neu */
void dbif_create (DBIF_FILE *db, char *dbname, int mode)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = GDBM_NEWDB | (CACHED ? GDBM_FAST : 0);
	res->current.dptr = NULL;
	res->current.dsize = 0;
	res->dbf = gdbm_open (dbname, CACHESIZE, res->rw, mode, dberr);
	if (!res->dbf) { free (res); res = (DBIF_FILE) 0; }
    }
    *db = res;
}

/* ffne eine Datenbank sowohl fr Schreib- als auch fr Lesebetrieb */
void dbif_rw (DBIF_FILE *db, char *dbname)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = GDBM_WRITER | (CACHED ? GDBM_FAST : 0);
	res->current.dptr = NULL;
	res->current.dsize = 0;
	res->dbf = gdbm_open (dbname, CACHESIZE, res->rw, 0644, dberr);
	if (!res->dbf) { free (res); res = (DBIF_FILE) 0; }
    }
    *db = res;
}

/* berprfe, ob eine Datenbank geffnet ist! */
/*int dbif_isopen (DBIF_FILE db)
**{
**    if (db->dbf != (GDBM_FILE) 0) {
**	if (db->rw & ~GDBM_FAST) {
**	    return -1;
**	} else {
**	    return 1;
**	}
**    }
**    return 0;
**}
*/

/* Schliee eine geffnete Datenbank! */
void dbif_close (DBIF_FILE db)
{
    if (db) {
	if (db->rw & GDBM_FAST) { gdbm_sync (db->dbf); }
	gdbm_close (db->dbf);
    }
}

/* Synchronisiere eine mit `fast' geffnete Datenbank */
void dbif_sync (DBIF_FILE db)
{
    if (db->rw & GDBM_FAST) {
	gdbm_sync (db->dbf);
    }
}

/* Lesen eines Benutzereintrages; im Gegensatz zu `dbif_resolve' wird
** dieser Eintrag - auch in dem Fall, da er ein gesetztes `link'-Flag
** enthlt - nicht weiter aufgelst. Diese Routine ermglicht daher
** - zusammen mit `dbif_set' - eine Administration dieses Eintrages.
*/
dbif_rec *dbif_get (DBIF_FILE db, char *em_addr)
{
    datum data, key;
    dbif_rec *rec = (dbif_rec *) 0;
    char *p;
    size_t len;
    key.dsize = strlen (em_addr);
    key.dptr = malloc (key.dsize);
    memcpy (key.dptr, em_addr, key.dsize);
    data = gdbm_fetch (db->dbf, key);
    if (!data.dptr) {
	free (key.dptr);
	return rec;
    }
    if (!(rec = (dbif_rec *) malloc (sizeof (dbif_rec) + data.dsize - 8 + 3))) {
	free (key.dptr);
	return rec;
    }
    /* 1. Element: passwd */
    p = data.dptr + 8;
    rec->passwd = (char *) rec + sizeof (dbif_rec);
    len = (size_t) get_length ((unsigned char *) data.dptr);
    memcpy (rec->passwd, p, len);
    rec->passwd[len] = '\0';
    /* 2. Element: folder */
    p += len;
    rec->folder = rec->passwd + len + 1;
    len = (size_t) get_length ((unsigned char *) data.dptr + 4);
    memcpy (rec->folder, p, len);
    rec->folder[len] = '\0';
    /* 3. Element: Flags */
    p += len;
    rec->flags = rec->folder + len + 1;
    len = N_FLAGS;
    memcpy (rec->flags, p, N_FLAGS);
    free (data.dptr);
    free (key.dptr);
    return rec;
}

/* Lies einen Benutzer aus der Datenbank; hat der Eintrag das Flag `link',
** dann enthlt das `Folder'-Element statt eines Dateinamens einen Verweis auf
** einen weiteren Benutzereintrag und der `Password'-Eintrag ist ohne Bedeutung;
** in diesem Fall wird der Verweis auf den anderen Benutzer aus dem
** `Folder'-Element gelesen und zu diesem verzweigt. Das ganze wird solange
** fortgesetzt, bis ein Eintrag ohne gesetztes `link'-Flag gefunden wird.
** Dieser wird dann zurckgegeben.
** Auf diese Art werden einfache `alias'-Mechanismen ermglicht.
*/
dbif_rec *dbif_resolve (DBIF_FILE db, char *em_addr)
{
    dbif_rec *rec;
    rec = dbif_get (db, em_addr);
    while (rec && is_member (rec->flags, "link")) {
	size_t slen = strlen (rec->folder);
	char *newaddr = malloc (slen + 1);
	if (!newaddr) {
	    free (rec);
	    rec = (dbif_rec *) 0;
	    return rec;
	}
	memcpy (newaddr, rec->folder, slen);
	newaddr[slen] = '\0';
	free (rec);
	rec = dbif_get (db, newaddr);
	free (newaddr);
    }
    if (!rec) { return rec; }
    if (is_member (rec->flags, "invalid")) {
	free (rec);
	rec = (dbif_rec *) 0;
    }
    return rec;
}

/* Fge einen neuen Benutzerdatensatz ein oder ersetze einen vorhandenen. */
static int dbif_insrpl (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 char *flags[],
		 int storeflag)
{
    char dsize = strlen (passwd) + strlen (folder) + N_FLAGS + 8;
    char *rv = malloc (dsize);
    char *xp;
    int ix, res;
    datum data, key;
    if (!rv) {
	return -2;
    }
    set_length ((unsigned char *) rv, strlen (passwd));
    set_length ((unsigned char *) rv + 4, strlen (folder));
    xp = rv + 8;
    memcpy (xp, passwd, strlen (passwd));
    xp += strlen (passwd);
    memcpy (xp, folder, strlen (folder));
    xp += strlen (folder);
    memset (xp, '\0', N_FLAGS);
    for (ix = 0; flags[ix]; ++ix) {
	set_member (xp, flags[ix]);
    }
    data.dptr = rv;
    data.dsize = dsize;
    key.dsize = strlen (em_addr);
    key.dptr = em_addr;
    res = gdbm_store (db->dbf, key, data, storeflag);
    free (rv);
    return res;
}

/* Fge einen NEUEN Benutzer in die Datenbank ein */
int dbif_vinsert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 char *flags[])
{
    return dbif_insrpl (db, em_addr, passwd, folder, flags, GDBM_INSERT);
}

int dbif_insert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 ...)
{
    int res;
    va_list ap;
    va_start (ap, folder);
    res = dbif_vinsert (db, em_addr, passwd, folder, (char **)ap);
    va_end (ap);
    return res;
}

/* Ersetze einen vorhandenen Benutzerdatensatz durch einen neuen;
** diese Routine wird unter anderem bei Passwortnderungen bentigt.
*/
int dbif_vset (DBIF_FILE db,
		  char *em_addr,
		  char *passwd,
		  char *folder,
		  char *flags[])
{
    return dbif_insrpl (db, em_addr, passwd, folder, flags, GDBM_REPLACE);
}

int dbif_set (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 ...)
{
    int res;
    va_list ap;
    va_start (ap, folder);
    res = dbif_vset (db, em_addr, passwd, folder, (char **)ap);
    va_end (ap);
    return res;
}

/* The combination of `dbif_seqinit(db)' and `dbif_seq(db,&rec)' enables a
** sequential access of the database `db'. There is no other way for
** completely reading the database contents.
*/

/* dbif_seqinit (db)
**     resets the sequential reading system; in the GDBM variant of this
**     module, it simply clears the db->current entry.
*/
void dbif_seqinit (DBIF_FILE db)
{
    if (db->current.dptr) {
	free (db->current.dptr);
	db->current.dptr = (char *) 0;
	db->current.dsize = 0;
    }
}

/* em_addr = dbif_seq (db, &rec)
 *     reads the next key/data pair from the database `db'. The key is
 *     returned as function result, the data pointer in the variable `rec'.
 *     If any error occurs or there is no more data to be read, the function
 *     result is the NULL pointer.
 */
char *dbif_seq (DBIF_FILE db, dbif_rec **rec)
{
    datum key;
    dbif_rec *res;
    char *em_addr;
    /* Get the first key if there was no key saved during the last call of
    ** `dbif_seq' or if dbif_seqinit was called before this call.
    */
    if (db->current.dptr) {
	key = gdbm_nextkey (db->dbf, db->current);
    } else {
	key = gdbm_firstkey (db->dbf);
    }
    /* If there is no more key, return immediately */
    if (!key.dptr) { return (char *) 0; }
    /* Allocate memory for an e-mail address; if this fails, don't change the
    ** current key (!!) and return immediately
    */
    if (!(em_addr = (char *) malloc (key.dsize + 1))) {
	free (key.dptr);
	return (char *) 0;
    }
    /* insert the key data into the allocated memory */
    memcpy (em_addr, key.dptr, key.dsize);
    em_addr[key.dsize] = '\0';
    /* get the corresponding data */
    res = dbif_get (db, em_addr);
    /* if this operation failed, don't change the current key (!!); free the
    ** memory at `key.dptr' and `em_addr' instead and return immediately.
    */
    if (!res) {
	free (em_addr);
	free (key.dptr);
	return (char *) 0;
    }
    /* at this point, all operations succeeded; now we can replace the
    ** saved key with the new one and return a valid e-mail address and
    ** the corresponding user record
    */
    if (db->current.dptr) { free (db->current.dptr); }
    db->current = key;
    *rec = res;
    return em_addr;
}


/******************************************************************************/

/* berprfe, ob ein Flag gesetzt ist */
int is_member (char *userflags, char *flagname)
{
    int ix;
    for (ix = 0; flags[ix].fl_name; ++ix) {
	if (!strcmp (flags[ix].fl_name, flagname)) {
	    return (int) userflags[flags[ix].fl_offset];
	}
    }
    return 0;
}

/* Hilfsfunktionen; diese mssen nach auen nicht sichtbar sein und werden
 * daher als `static' deklariert.
 */

/* Hole ein Lngenwort (MSB byte order) aus dem Datenvektor */
static int get_length (unsigned char *dvec)
{
    return (((((*dvec << 8) | dvec[1]) << 8) | dvec[2]) << 8) | dvec[3];
/*
    int p = *dvec;
    p <<= 8; p |= dvec[1];
    p <<= 8; p |= dvec[2];
    p <<= 8; p |= dvec[3];
    return p;
*/
}

/* Schreibe ein Lngenwort (MSB byte order) in den Datenvektor */
static void set_length (unsigned char *dvec, int p)
{
    dvec[3] = (unsigned char) (p & 0xff); p >>= 8;
    dvec[2] = (unsigned char) (p & 0xff); p >>= 8;
    dvec[1] = (unsigned char) (p & 0xff); p >>= 8;
    *dvec   = (unsigned char) (p & 0xff);
}

/* insert a flag into the flag set */
void set_member (char *userflags, char *flagname)
{
    int ix;
    for (ix = 0; flags[ix].fl_name; ++ix) {
	if (!strcmp (flags[ix].fl_name, flagname)) {
	    userflags[flags[ix].fl_offset] = '\1';
	}
    }
}

const char *dbif_errormsg (void)
{
    return gdbm_strerror (gdbm_errno);
}


/* Erzeuge einen Benutzerdatensatz aus den Daten fr Pawort, Foldername und
** Flags; die Flags werden als NULL-terminierte Liste von Flagnamen (Zeichen-
** ketten hinter dem Foldernamen bergeben.
*/
/*
dbif_rec *dbif_mkrec (char *passwd, char *folder, ...)
{
    va_list ap;
    char *flag;
    dbif_rec *res;
    size_t pwl = strlen (passwd), fll = strlen (folder);
    res = (dbif_rec *) malloc (sizeof (dbif_rec) + pwl + fll + N_FLAGS + 3);
    if (!res) { return res; }
    res->passwd = (char *) res + sizeof (dbif_rec);
    memcpy (res->passwd, passwd, pwl); res->passwd[pwl] = '\0';
    res->folder = res->passwd + pwl + 1;
    memcpy (res->folder, folder, fll); res->folder[fll] = '\0';
    res->flags = res->folder + fll + 1;
    memset (res->flags, '\0', N_FLAGS);
    va_start (ap, folder);
    while ((flag = va_arg (ap, char *))) {
	set_member (res->flags, flag);
    }
    va_end (ap);
    return res;
}

void dbif_free (dbif_rec *p)
{
    free (p);
}
*/
