/* dbif.c
**
** Database Interface for the user/admin databases
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>
#include <fcntl.h>

#define _DBIF_C
#include "dbif.h"
#include "config.h"
#include "prefs.h"
#include "vxerror.h"
#include "tools.h"
#include "dbifconf.h"

/* In den Flags gibt es einen Eintrag `admin'.
** Dieser Eintrag erlaubt es dem Benutzer, der ber dieses Flag verfgt, neue
** Benutzer in seiner jeweiligen Domain zu anzulegen, bzw. Benutzer aus seiner
** Domain zu entfernen.
** Der Eintrag `system' soll es einem (NUR EINEM) Benutzer erlauben, Benutzer
** in jeder Domain anzulegen, bzw. aus ihr zu entfernen.
*/

static struct {
    char *fl_name;
    int fl_offset;
} flags[] = {
    { "link", 0 },
    { "admin", 1 },
    { "system", 2 },
    { "mailadmin", 3 },
    { "domain", 4 },
    { "ip4trans", 5 },
    { "ip6trans", 6 },
    { "identity", 7 },
    { "nonstandard", 8 },
    { "invalid", 9 },
    { NULL, 0 }
};

/* Vorwrts-Deklarationen fr die Hilfsfunktionen */
int is_member (char *userflags, char *flagname);
static int get_length (unsigned char *dvec);
void set_member (char *userflags, char *flagname);
static void set_length (unsigned char *dvec, int p);

/* Die User-Datenbank besteht aus den Schlsseln (den E-Mail-Adressen) und
** den diesen Schlsseln zugeordneten Daten, die wie folgt angeordnet sind:
**	Offset		Bedeutung
**	 0		Lnge des Passwortes (NPW)
**	 4		Lnge des Pfadnamens des Mail-Folders (NFP)
**	 8		Lnge des Flag-Feldes (NFL)
**	12		Passwort
**	12+NPW		Pfadname des Mail-Folders
**	12+NPW+NFP	Flag-Feld
**	12+NPW+NFP+NFL	Ende des Datensatzes
*/

static void *set_info (size_t csz)
{
    static HASHINFO hi;
    static BTREEINFO bi;
    static RECNOINFO ri;
    memset (&hi, '\0', sizeof (hi));
    memset (&bi, '\0', sizeof (bi));
    memset (&ri, '\0', sizeof (ri));
    switch (prefs.dbtype) {
	case DB_HASH:  hi.cachesize = csz; return (void *) &hi;
	case DB_BTREE: bi.cachesize = csz; return (void *) &bi;
	case DB_RECNO: ri.cachesize = csz; return (void *) &ri;
    }
    return NULL;
}

static int dbtypes[] = { DB_HASH, DB_BTREE, DB_RECNO };

/* ffne eine Datenbank im Nur-Lese-Modus */
void dbif_reset (DBIF_FILE db, char *dbname)
{
    db->rw = O_RDONLY;
    db->dbtype = dbtypes[prefs.dbtype];
    db->dbf = dbopen (dbname, db->rw, 0, db->dbtype, set_info(CACHESIZE));
}

/* Erzeuge eine Datenbank neu */
void dbif_create (DBIF_FILE db, char *dbname, int mode)
{
    db->rw = O_CREAT|O_TRUNC|O_RDWR;
    db->dbtype = dbtypes[prefs.dbtype];
    db->dbf = dbopen (dbname, db->rw, mode, db->dbtype, set_info(CACHESIZE));
}

/* ffne eine Datenbank sowohl fr Schreib- als auch fr Lesebetrieb */
void dbif_rw (DBIF_FILE db, char *dbname)
{
    db->rw = O_RDWR;
    db->dbtype = dbtypes[prefs.dbtype];
    db->dbf = dbopen (dbname, db->rw, 0, db->dbtype, set_info(CACHESIZE));
}

void dbif_init (DBIF_FILE db)
{
    db->dbf = (DB *) 0;
    db->rw = 0;
    db->seq_flag = R_FIRST;
}

void dbif_copy (DBIF_FILE src, DBIF_FILE dst)
{
    dst->dbf = src->dbf;
    dst->dbtype = src->dbtype;
    dst->rw = src->rw;
}

/* berprfe, ob eine Datenbank geffnet ist! */
int dbif_isopen (DBIF_FILE db)
{
    if (db->dbf != (DB *) 0) {
	if (db->rw & (O_RDWR | O_WRONLY)) {
	    return -1;		/* Ergebnis = -1, falls Schreibberechtigung! */
	} else {
	    return 1;		/* Ergebnis = -1, falls Leseberechtigung! */
	}
    }
    return 0;			/* Ergebnis = 0, falls nicht offen! */
}

/* Schliee eine geffnete Datenbank! */
void dbif_close (DBIF_FILE db)
{
    if (db->dbf) {
	db->dbf->close (db->dbf);
	db->dbf = (DB *) 0;
    }
}

/* Synchronisiere eine mit `fast' geffnete Datenbank */
void dbif_sync (DBIF_FILE db)
{
    if (db->rw & (O_RDWR | O_WRONLY)) {
	db->dbf->sync (db->dbf, 0);
    }
}

/* Lesen eines Benutzereintrages; im Gegensatz zu `dbif_resolve' wird
** dieser Eintrag - auch in dem Fall, da er ein gesetztes `link'-Flag
** enthlt - nicht weiter aufgelst. Diese Routine ermglicht daher
** - zusammen mit `dbif_set' - eine Administration dieses Eintrages.
*/
dbif_rec dbif_get (DBIF_FILE db, char *em_addr)
{
    DBT data, key;
    dbif_rec rec;
    char *p;
    size_t len;
    key.size = strlen (em_addr);
    key.data = malloc (key.size);
    memcpy (key.data, em_addr, key.size);
    if (db->dbf->get (db->dbf, &key, &data, 0)) {
	rec.passwd = rec.folder = rec.flags = rec.end = rec.base = NULL;
	free (key.data);
	return rec;
    }
    if (!(rec.base = malloc (data.size - 8 + 3))) {
	rec.passwd = rec.folder = rec.flags = rec.end = rec.base = NULL;
	free (key.data);
	return rec;
    }
    /* 1. Element: passwd */
    p = (char *) data.data + 8;
    rec.passwd = rec.base;
    len = (size_t) get_length (data.data);
    memcpy (rec.passwd, p, len);
    rec.passwd[len] = '\0';
    /* 2. Element: folder */
    p += len;
    rec.folder = rec.passwd + len + 1;
    len = (size_t) get_length ((unsigned char *)data.data + 4);
    memcpy (rec.folder, p, len);
    rec.folder[len] = '\0';
    /* 3. Element: Flags */
    p += len;
    rec.flags = rec.folder + len + 1;
    len = N_FLAGS;
    memcpy (rec.flags, p, N_FLAGS);
    rec.end = rec.flags + len;
    free (data.data);
    free (key.data);
    return rec;
}

/* Lies einen Benutzer aus der Datenbank; hat der Eintrag das Flag `link',
** dann enthlt das `Folder'-Element statt eines Dateinamens einen Verweis auf
** einen weiteren Benutzereintrag und der `Password'-Eintrag ist ohne Bedeutung;
** in diesem Fall wird der Verweis auf den anderen Benutzer aus dem
** `Folder'-Element gelesen und zu diesem verzweigt. Das ganze wird solange
** fortgesetzt, bis ein Eintrag ohne gesetztes `link'-Flag gefunden wird.
** Dieser wird dann zurckgegeben.
** Auf diese Art werden einfache `alias'-Mechanismen ermglicht.
*/
dbif_rec dbif_resolve (DBIF_FILE db, char *em_addr)
{
    dbif_rec rec;
    rec = dbif_get (db, em_addr);
    while (rec.base && is_member (rec.flags, "link")) {
	size_t slen = strlen (rec.folder);
	char *newaddr = malloc (slen + 1);
	if (!newaddr) {
	    rec.passwd = rec.folder = rec.flags = rec.end = rec.base = NULL;
	    return rec;
	}
	memcpy (newaddr, rec.folder, slen);
	newaddr[slen] = '\0';
	free (rec.base);
	rec = dbif_get (db, newaddr);
	free (newaddr);
    }
    if (!rec.base) { return rec; }
    if (is_member (rec.flags, "invalid")) {
	free (rec.base);
	rec.base = rec.passwd = rec.folder = rec.flags = rec.end = NULL;
    }
    return rec;
}

/* Fge einen neuen Benutzerdatensatz ein oder ersetze einen vorhandenen. */
static int dbif_insrpl (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 char *flags[],
		 int storeflag)
{
    char dsize = strlen (passwd) + strlen (folder) + N_FLAGS + 8;
    char *rv = malloc (dsize);
    char *xp;
    int ix, res;
    DBT data, key;
    if (!rv) {
	return -2;
    }
    set_length ((unsigned char *)rv, strlen (passwd));
    set_length ((unsigned char *)rv + 4, strlen (folder));
    xp = rv + 8;
    memcpy (xp, passwd, strlen (passwd));
    xp += strlen (passwd);
    memcpy (xp, folder, strlen (folder));
    xp += strlen (folder);
    memset (xp, '\0', N_FLAGS);
    for (ix = 0; flags[ix]; ++ix) {
	set_member (xp, flags[ix]);
    }
    data.data = rv;
    data.size = dsize;
    key.size = strlen (em_addr);
    key.data = em_addr;
    res = db->dbf->put (db->dbf, &key, &data, storeflag);
    free (rv);
    return res;
}

/* Fge einen NEUEN Benutzer in die Datenbank ein */
int dbif_vinsert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 char *flags[])
{
    return dbif_insrpl (db, em_addr, passwd, folder, flags, R_NOOVERWRITE);
}

int dbif_insert (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 ...)
{
    int res;
    va_list ap;
    va_start (ap, folder);
    res = dbif_vinsert (db, em_addr, passwd, folder, (char **)ap);
    va_end (ap);
    return res;
}

/* Ersetze einen vorhandenen Benutzerdatensatz durch einen neuen;
** diese Routine wird unter anderem bei Passwortnderungen bentigt.
*/
int dbif_vset (DBIF_FILE db,
		  char *em_addr,
		  char *passwd,
		  char *folder,
		  char *flags[])
{
    return dbif_insrpl (db, em_addr, passwd, folder, flags, R_CURSOR);
}

int dbif_set (DBIF_FILE db,
		 char *em_addr,
		 char *passwd,
		 char *folder,
		 ...)
{
    int res;
    va_list ap;
    va_start (ap, folder);
    res = dbif_vset (db, em_addr, passwd, folder, (char **)ap);
    va_end (ap);
    return res;
}

/* The combination of `dbif_seqinit(db)' and `dbif_seq(db,&rec)' enables a
 * ** sequential access of the database `db'. There is no other way for
 * ** completely reading the database contents.
 * */

/* dbif_seqinit (db)
 * **     resets the sequential reading system; in the GDBM variant of this
 * **     module, it simply clears the db->current entry.
 * */
void dbif_seqinit (DBIF_FILE db)
{
    db->seq_flag = R_FIRST;
}

/* em_addr = dbif_seq (db, &rec)
**     reads the next key/data pair from the database `db'. The key is
**     returned as function result, the data pointer in the variable `rec'.
**     If any error occurs or there is no more data to be read, the function
**     result is the NULL pointer.
**/
char *dbif_seq (DBIF_FILE db, dbif_rec *rec)
{
    DBT key, data;
    int st;
    char *em_addr, *p;
    dbif_rec res;
    size_t len;
    if ((st = db->dbf->seq (db->dbf, &key, &data, db->seq_flag))) {
	return NULL;
    }
    if (!(em_addr = (char *) malloc (key.size + 1))) {
	free (key.data);
	free (data.data);
	return NULL;
    }
    if (!(res.base = (char *) malloc (data.size - 8 + 3))) {
	free (em_addr);
	free (key.data);
	free (data.data);
	return NULL;
    }
    /* copy the e-mail address into the allocated memory area */
    memcpy (em_addr, key.data, key.size);
    em_addr[key.size] = '\0';
    /* insert the password data into the user record */
    res.passwd = res.base;
    p = (char *) data.data + 8;
    len = get_length (data.data);
    memcpy (res.passwd, p, len);
    res.passwd[len] = '\0';
    /* insert the folder data into the user record */
    res.folder = res.passwd + len + 1;
    p += len;
    len = get_length ((unsigned char *)data.data + 4);
    memcpy (res.folder, p, len);
    res.folder[len] = '\0';
    /* insert the flags into the user record */
    res.flags = res.folder + len + 1;
    p += len;
    memcpy (res.flags, p, N_FLAGS);
    res.end = res.flags + N_FLAGS;
    /* release the memory allocated for `key' and `data' */
    free (key.data);
    free (data.data);
    /* return the user record in the argument referenced by `rec' and the
    ** key as the function result
    */
    *rec = res;
    return em_addr;
}

/******************************************************************************/

/* berprfe, ob ein Flag gesetzt ist */
int is_member (char *userflags, char *flagname)
{
    int ix;
    for (ix = 0; flags[ix].fl_name; ++ix) {
	if (!strcmp (flags[ix].fl_name, flagname)) {
	    return (int) userflags[flags[ix].fl_offset];
	}
    }
    return 0;
}

/* Hilfsfunktionen; diese mssen nach auen nicht sichtbar sein und werden
 * daher als `static' deklariert.
 */

/* Hole ein Lngenwort (MSB byte order) aus dem Datenvektor */
static int get_length (unsigned char *dvec)
{
    return (((((*dvec << 8) | dvec[1]) << 8) | dvec[2]) << 8) | dvec[3];
/*
    int p = *dvec;
    p <<= 8; p |= dvec[1];
    p <<= 8; p |= dvec[2];
    p <<= 8; p |= dvec[3];
    return p;
*/
}

/* Schreibe ein Lngenwort (MSB byte order) in den Datenvektor */
static void set_length (unsigned char *dvec, int p)
{
    dvec[3] = (unsigned char) (p & 0xff); p >>= 8;
    dvec[2] = (unsigned char) (p & 0xff); p >>= 8;
    dvec[1] = (unsigned char) (p & 0xff); p >>= 8;
    *dvec   = (unsigned char) (p & 0xff);
}

/* insert a flag into the flag set */
void set_member (char *userflags, char *flagname)
{
    int ix;
    for (ix = 0; flags[ix].fl_name; ++ix) {
	if (!strcmp (flags[ix].fl_name, flagname)) {
	    userflags[flags[ix].fl_offset] = '\1';
	}
    }
}

const char *dbif_errormsg (void)
{
#ifdef __USE_BSD
    return sys_errlist[errno];
#else
#ifdef __USE_GNU
    return _sys_errlist[errno];
#else
    return "database internal error";
#endif
#endif
}


/* Erzeuge einen Benutzerdatensatz aus den Daten fr Pawort, Foldername und
** Flags; die Flags werden als NULL-terminierte Liste von Flagnamen (Zeichen-
** ketten hinter dem Foldernamen bergeben.
*/
dbif_rec dbif_mkrec (char *passwd, char *folder, ...)
{
    va_list ap;
    char *flag;
    dbif_rec res;
    size_t pwl = strlen (passwd), fll = strlen (folder);
    res.base = (char *) malloc (pwl + fll + N_FLAGS + 3);
    if (!res.base) { return res; }
    res.passwd = res.base;
    memcpy (res.passwd, passwd, pwl); res.passwd[pwl] = '\0';
    res.folder = res.passwd + pwl + 1;
    memcpy (res.folder, folder, fll); res.folder[fll] = '\0';
    res.flags = res.folder + fll + 1;
    memset (res.flags, '\0', N_FLAGS);
    res.end = res.folder + N_FLAGS;
    va_start (ap, folder);
    while ((flag = va_arg (ap, char *))) {
	set_member (res.flags, flag);
    }
    va_end (ap);
    return res;
}
