/* dbif.c
**
** Database Interface for the user/admin databases
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <stdarg.h>

#include "dbif_priv.h"
#include "config.h"
#include "vxerror.h"
#include "tools.h"
#include "ifconf.h"

/* In den Flags gibt es einen Eintrag `admin'.
** Dieser Eintrag erlaubt es dem Benutzer, der ber dieses Flag verfgt, neue
** Benutzer in seiner jeweiligen Domain zu anzulegen, bzw. Benutzer aus seiner
** Domain zu entfernen.
** Der Eintrag `system' soll es einem (NUR EINEM) Benutzer erlauben, Benutzer
** in jeder Domain anzulegen, bzw. aus ihr zu entfernen.
*/

/* Die User-Datenbank besteht aus den Schlsseln (den E-Mail-Adressen) und
** den diesen Schlsseln zugeordneten Daten, die wie folgt angeordnet sind:
**	Offset		Bedeutung
**	 0		Lnge des Passwortes (NPW)
**	 4		Lnge des Pfadnamens des Mail-Folders (NFP)
**	 8		Lnge des Flag-Feldes (NFL)
**	12		Passwort
**	12+NPW		Pfadname des Mail-Folders
**	12+NPW+NFP	Flag-Feld
**	12+NPW+NFP+NFL	Ende des Datensatzes
*/

dbif_error if_errno = DBIF_NOERR;

static void dberr (char *message)
{
    vxerror (EX_SOFTWARE, 0, "database internal error: %s!\n",
			     gdbm_strerror (gdbm_errno));
}

/* ffne eine Datenbank im Nur-Lese-Modus */
void dbif_reset (DBIF_FILE *db, char *dbname)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = GDBM_READER;
	res->current.dptr = NULL;
	res->current.dsize = 0;
	res->dbf = gdbm_open (dbname, CACHESIZE, res->rw, 0644, dberr);
	if (!res->dbf) {
	    dbif_traseterr ();
	    free (res); res = (DBIF_FILE) 0;
	} else {
	    if_errno = DBIF_NOERR;
	}
    } else {
	if_errno = DBIF_ENOMEM;
    }
    *db = res;
}

/* Erzeuge eine Datenbank neu */
void dbif_create (DBIF_FILE *db, char *dbname, int mode)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = GDBM_NEWDB | (CACHED ? GDBM_FAST : 0);
	res->current.dptr = NULL;
	res->current.dsize = 0;
	res->dbf = gdbm_open (dbname, CACHESIZE, res->rw, mode, dberr);
	if (!res->dbf) {
	    dbif_traseterr ();
	    free (res); res = (DBIF_FILE) 0;
	} else {
	    if_errno = DBIF_NOERR;
	}
    } else {
	if_errno = DBIF_ENOMEM;
    }
    *db = res;
}

/* ffne eine Datenbank sowohl fr Schreib- als auch fr Lesebetrieb */
void dbif_rw (DBIF_FILE *db, char *dbname)
{
    DBIF_FILE res;
    if ((res = (DBIF_FILE) malloc (sizeof (struct _DBIF_FILE_S)))) {
	res->rw = GDBM_WRITER | (CACHED ? GDBM_FAST : 0);
	res->current.dptr = NULL;
	res->current.dsize = 0;
	res->dbf = gdbm_open (dbname, CACHESIZE, res->rw, 0644, dberr);
	if (!res->dbf) {
	    dbif_traseterr ();
	    free (res); res = (DBIF_FILE) 0;
	} else {
	    if_errno = DBIF_NOERR;
	}
    } else {
	if_errno = DBIF_ENOMEM;
    }
    *db = res;
}

/* berprfe, ob eine Datenbank geffnet ist! */
/*int dbif_isopen (DBIF_FILE db)
**{
**    if (db->dbf != (GDBM_FILE) 0) {
**	if (db->rw & ~GDBM_FAST) {
**	    return -1;
**	} else {
**	    return 1;
**	}
**    }
**    return 0;
**}
*/

/* Schliee eine geffnete Datenbank! */
void dbif_close (DBIF_FILE db)
{
    if_errno = DBIF_NOERR;
    if (db) {
	if (db->rw & GDBM_FAST) { gdbm_sync (db->dbf); }
	gdbm_close (db->dbf);
	free (db);
	dbif_traseterr ();
    }
}

/* Synchronisiere eine mit `fast' geffnete Datenbank */
void dbif_sync (DBIF_FILE db)
{
    if_errno = DBIF_NOERR;
    if (db->rw & GDBM_FAST) {
	gdbm_sync (db->dbf);
	dbif_traseterr ();
    }
}

/* Lesen eines Benutzereintrages; im Gegensatz zu `dbif_resolve' wird
** dieser Eintrag - auch in dem Fall, da er ein gesetztes `link'-Flag
** enthlt - nicht weiter aufgelst. Diese Routine ermglicht daher
** - zusammen mit `dbif_set' - eine Administration dieses Eintrages.
*/
dbif_rec *dbif_get (DBIF_FILE db, char *em_addr)
{
    datum data, key;
    dbif_rec *rec = (dbif_rec *) 0;
    key.dsize = strlen(em_addr); key.dptr = em_addr;
    data = gdbm_fetch(db->dbf, key);
    if (!data.dptr) { dbif_traseterr (); return rec; }
    rec = dbif_mkreca (data.dptr);
    free (data.dptr);
    return rec;
}

/* Lies einen Benutzer aus der Datenbank; hat der Eintrag das Flag `link',
** dann enthlt das `Folder'-Element statt eines Dateinamens einen Verweis auf
** einen weiteren Benutzereintrag und der `Password'-Eintrag ist ohne Bedeutung;
** in diesem Fall wird der Verweis auf den anderen Benutzer aus dem
** `Folder'-Element gelesen und zu diesem verzweigt. Das ganze wird solange
** fortgesetzt, bis ein Eintrag ohne gesetztes `link'-Flag gefunden wird.
** Dieser wird dann zurckgegeben.
** Auf diese Art werden einfache `alias'-Mechanismen ermglicht.
*/
dbif_rec *dbif_resolve (DBIF_FILE db, char *em_addr, char **res_addr)
{
    dbif_rec *rec;
    int rtype;
    char *newaddr;
    rtype = strlen (em_addr) + 1;
    if (!(newaddr = malloc (strlen (em_addr) + 1))) { return (dbif_rec *) 0; }
    strcpy (newaddr, em_addr);
    /* Hole einen Datensatz */
    rec = dbif_get (db, newaddr);
    /* Solange ein passender Datensatz vorhanden ist */
    while (rec) {
	/* Hole den Typ des Datensatzes */
	rtype = dbif_rectype(rec);
	/* Wenn dieser Typ weder ein `link' noch ein `domain link' bezeichnet */
	if (rtype != RT_LINK && rtype != RT_DOMLINK) {
	    /* Wenn der Datensatz einen ungltigen Benutzer bezeichnet,
	    ** dann MUSS das Ergebnis NULL sein; gib in diesem Fall auch
	    ** den (mglichen) Schlssel frei.
	    */
	    if (rtype == RT_INVALID) {
		free(newaddr); dbif_free(rec);
		if_errno = DBIF_NOIT;
		return (dbif_rec *) 0;
	    }
	    /* Wenn eine Ergebnisvariable fr den Schlssel angegeben wurde,
	    ** dann schreibe diesen dort hinein; sonst gib den fr den
	    ** Schlssel reservierten Speicherplatz frei
	    */
	    if (res_addr) { *res_addr = newaddr; } else { free (newaddr); }
	    /* Der aktuelle Datensatz ist das Ergebnis */
	    if_errno = DBIF_NOERR;
	    return rec;
	}
	/* Der Datensatz war ein `link', oder ein `domain link'; versuche es
	** mit dem nchsten Datensatz in der Kette
	*/
	free (newaddr);
	if (!(newaddr = malloc (strlen (rec->folder) + 1))) {
	    free (rec); rec = (dbif_rec *) 0;
	    if_errno = DBIF_ENOMEM;
	    return rec;
	}
	strcpy (newaddr, rec->folder);
	free (rec); rec = dbif_get (db, newaddr);
    }
    free (newaddr);
    if_errno = DBIF_NOIT;
    return rec;
}

/* Fge einen neuen Benutzerdatensatz ein oder ersetze einen vorhandenen. */
static int dbif_insrpl (DBIF_FILE db, char *em_addr, dbif_rec *rec, int stflag)
{
    size_t pwl = strlen (rec->passwd),
	   fll = strlen (rec->folder),
	   dsize = pwl + fll + 1 + 2 * sizeof(int);
    char *p;
    datum key, data;
    int rtype = dbif_rectype(rec), nusers, maxusers, res;
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	dsize += 2 * sizeof(int);
    }
    if (!(p = (char *) malloc (dsize))) { if_errno = DBIF_ENOMEM; return 1; }
    data.dsize = dsize; data.dptr = p;
    set_int ((unsigned char *) p, pwl); p += sizeof (int);
    set_int ((unsigned char *) p, fll); p += sizeof (int);
    memcpy (p, rec->passwd, pwl); p += pwl;
    memcpy (p, rec->folder, fll); p += fll;
    *p++ = (char) dbif_rectype(rec);
    if (rtype == RT_DOMAIN || rtype == RT_NSDOMAIN) {
	dbif_getlim (rec, &nusers, &maxusers);
	set_int ((unsigned char *) p, nusers); p += sizeof(int);
	set_int ((unsigned char *) p, maxusers);
    }
    key.dsize = strlen (em_addr); key.dptr = em_addr;
    res = gdbm_store (db->dbf, key, data, stflag);
    free (data.dptr);
    dbif_traseterr ();
    return res;
}

int dbif_insert (DBIF_FILE db, char *em_addr, dbif_rec *rec)
{
    return dbif_insrpl (db, em_addr, rec, GDBM_INSERT);
}

int dbif_set (DBIF_FILE db, char *em_addr, dbif_rec *rec)
{
    return dbif_insrpl (db, em_addr, rec, GDBM_REPLACE);
}

const char *dbif_errormsg (void)
{
    return gdbm_strerror (gdbm_errno);
}

/* Fehlercodes: ich schrnke die vielen GDBM-Fehlercodes etwas ein, weil zum
**              einen nicht alle Fehlersituationen auftreten sollten, zum
**              anderen, weil ich nicht so spezifische Fehlercodes bentige.
*/
/*
typedef enum {
    DBIF_NOERR, DBIF_ENOMEM, DBIF_ACFAIL, DBIF_WRFAIL, DBIF_RDFAIL,
    DBIF_SLFAIL, DBIF_NOWR, DBIF_NORD, DBIF_NOIT, DBIF_NODB, DBIF_INT,
    DBIF_DATA, DBIF_CCFAIL
} dbif_error;
*/

static dbif_error errtra[] = {
    /* GDBM_NO_ERROR			=> */ DBIF_NOERR,
    /* GDBM_MALLOC_ERROR		=> */ DBIF_ENOMEM,
    /* GDBM_BLOCK_SIZE_ERROR		=> */ DBIF_INT,
    /* GDBM_FILE_OPEN_ERROR		=> */ DBIF_ACFAIL,
    /* GDBM_FILE_WRITE_ERROR		=> */ DBIF_WRFAIL,
    /* GDBM_FILE_SEEK_ERROR		=> */ DBIF_SLFAIL,
    /* GDBM_FILE_READ_ERROR		=> */ DBIF_RDFAIL,
    /* GDBM_BAD_MAGIC_NUMBER		=> */ DBIF_NODB,
    /* GDBM_EMPTY_DATABASE		=> */ DBIF_NODB,
    /* GDBM_CANT_BE_READER		=> */ DBIF_NORD,
    /* GDBM_CANT_BE_WRITER		=> */ DBIF_NOWR,
    /* GDBM_READER_CANT_DELETE		=> */ DBIF_NOWR,
    /* GDBM_READER_CANT_STORE		=> */ DBIF_NOWR,
    /* GDBM_READER_CANT_REORGANIZE	=> */ DBIF_NOWR,
    /* GDBM_UNKNOWN_UPDATE		=> */ DBIF_WRFAIL,
    /* GDBM_ITEM_NOT_FOUND		=> */ DBIF_NOIT,
    /* GDBM_REORGANIZE_FAILED		=> */ DBIF_WRFAIL,
    /* GDBM_CANNOT_REPLACE		=> */ DBIF_WRFAIL,
    /* GDBM_ILLEGAL_DATA		=> */ DBIF_DATA,
    /* GDBM_OPT_ALREADY_SET		=> */ DBIF_INT,
    /* GDBM_OPT_ILLEGAL			=> */ DBIF_INT
};

void dbif_traseterr (void)
{
    if_errno = errtra[gdbm_errno];
}

dbif_error dbif_errno (void)
{
    return if_errno;
}

int dbif_del (DBIF_FILE db, char *em_addr)
{
    datum key;
    int rstat;
    key.dsize = strlen (em_addr);
    key.dptr = em_addr;
    rstat = gdbm_delete (db->dbf, key);
    dbif_traseterr ();
    return rstat;
}
