/* exithl.c
**
** REMOVABLE exit functions (based on a handler which is automatically
** installed with `atexit' when the first exit handler is installed).
*/

#include <stdio.h>
#include <stdlib.h>
#include <strings.h>

#include "vxerror.h"
#include "tools.h"

/* With the following flag variable i decide, if the exit handler is
** already installed or not.
*/
static int exit_handler_installed = 0;

/* Exit functions should be called in the reverse order of their insertion.
** So a simple LIFO list should be enough for this.
*/
static struct exit_list {
    struct exit_list *next;
    char *handle_name;
    void (*exit_func) (void);
} *handler_list = (struct exit_list *) 0;

/* This is the exit handler which will be automatically installed during the
** first `add_exithl' (see below). When called (from `atexit') it steps
** through the `handler_list' and calls each function in this list excatly
** once.
*/
static void exit_handler (void)
{
    struct exit_list *curr;
    for (curr = handler_list; curr; curr = curr->next) {
	curr->exit_func ();
    }
}

/* add_exithl (name, exit_func)
**     inserts the function `exit_func' with the handle-name `name' at the
**     top of the `handler_list' and installs the function `exit_handler'
**     with `atexit' if it is not already installed.
*/
void add_exithl (char *handle_name, void (*exit_func) (void))
{
    struct exit_list *new;
    size_t newl = sizeof(struct exit_list) + strlen (handle_name) + 1;
    char *p;
    if (!exit_handler_installed) {
	atexit (exit_handler);
	exit_handler_installed = 1;
    }
    new = galloc (__FILE__, __LINE__, newl);
    new->next = handler_list;
    p = (char *) new + sizeof (struct exit_list);
    new->handle_name = p;
    strcpy (p, handle_name);
    new->exit_func = exit_func;
    handler_list = new;
}

/* del_exithl (name)
**     removes each occurrence of the exit handler with the handle-name `name'
**     from the `handler_list'.
*/
void del_exithl (char *handle_name)
{
    struct exit_list *rmv, *prv;
    if (!handler_list) { return; }
    while (!strcmp (handler_list->handle_name, handle_name)) {
	rmv = handler_list;
	handler_list = rmv->next;
	gfree (rmv);
    }
    if (!handler_list) { return; }
    for (prv = handler_list, rmv = prv->next; rmv; prv = rmv, rmv = prv->next) {
	if (!strcmp (rmv->handle_name, handle_name)) {
	    prv->next = rmv->next;
	    gfree (rmv);
	    rmv = prv;
	}
    }
}

/* clean_exithl ()
**     removes all exit handlers from the `handler_list'
*/
void clean_exithl (void)
{
    struct exit_list *curr;
    while (handler_list) {
	curr = handler_list;
	handler_list = curr->next;
	gfree (curr);
    }
}
