/* rfc822.c
**
** convert a date string as specified in RFC822 into a UTC `time_t' value
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <ctype.h>
#if !(defined(__USE_SVID) || defined(__USE_XOPEN))
#warning neither __USE_SVID nor __USE_XOPEN defined!
#define timezone __timezone
#endif
#include <time.h>

#include "tools.h"

/*## moved to `tools.c', because i need it in other modules, too
**static int lccmp (char *s1, char *s2, size_t n)
**{
**    while (n-- > 0) {
**	if (tolower (*s1++) != tolower (*s2++)) {
**	    return (int) tolower(*--s1) - (int) tolower(*--s2);
**	}
**    }
**    return 0;
**}
*/
static char *months[] = {
    "Jan", "Feb", "Mar", "Apr",
    "May", "Jun", "Jul", "Aug",
    "Sep", "Oct", "Nov", "Dec"
};

static int month (char *s)
{
    int ix;
    for (ix = 0; ix < 12; ++ix) {
	if (!lccmp(s, months[ix], strlen(months[ix]))) { return ix; }
    }
    return 0;
}

static char *wdays[] = { "Sun", "Mon", "Tue", "Wed", "Thu", "Fri", "Sat" };

static int wday (char *s)
{
    int ix;
    for (ix = 0; ix < 7; ++ix) {
	if (!lccmp(s, wdays[ix], strlen(wdays[ix]))) { return ix; }
    }
    return -1;
}

static struct {
    char *zname;
    long zdist;
} zones[] = {
    { "UT", 0l }, { "GMT", 0l }, { "Z", 0l },
    { "EST", -5l * 3600l }, { "EDT", -4l * 3600l },
    { "CST", -6l * 3600l }, { "CDT", -5l * 3600l },
    { "MST", -7l * 3600l }, { "MDT", -6l * 3600l },
    { "PST", -8l * 3600l }, { "PDT", -7l * 3600l },
    { "MESZ", +2l * 3600l }, { "MEZ", +1l * 3600l }, { "MEWZ", +1l * 3600l },
    { "MEST", +2l * 3600l }, { "MET", +1l * 3600l }, { "MEWT", +1l * 3600l },
    { "A", -1l * 3600l }, { "M", -12l * 3600l },
    { "N", +1l * 3600l }, { "Y", +12l * 3600l },
    { (char *) NULL, 0 }
};

#ifndef isblank
#define isblank(c) ((c) == ' ' || (c) == '\t')
#endif

#define isblz(c) (!(c) || isblank((c)))

static int get_tz (char *p, long *res, char **q)
{
    long hv, mv, tsgn;
    int ix;
    char *zone;
    while (isblank (*p)) ++p;
    if (!*p) { *q = p; *res = 0l; return 0; }
    for (ix = 0; (zone = zones[ix].zname); ++ix) {
	if (!lccmp(p, zone, strlen(zone)) && isblz(p[strlen(zone)])) {
	    *q = p + strlen(zone);
	    *res = zones[ix].zdist;
	    return 0;
	}
    }
    if (*p != '+' && *p != '-') { *q = p; return 1; }
    tsgn = (*p == '+' ? +1 : -1);
    ++p;
    hv = strtol (p, q, 10);
    if ((*q - 3) != p && (*q - 4) != p) { return 1; }
    if (**q && !isblank (**q)) { return 1; }
    mv = hv % 100l; hv /= 100l;
    *res = tsgn * (60l * (mv + 60l * hv));
    return 0;
}

/* rfc822_time (rfc822_timestring, &res)
**     calculates the UTC time (GMT time) from given string
**     `rfc822_timestring' and stores it in the resulting parameter `res'.
**     If the string is in rfc822 time format, the function result is 0;
**     if not, the function returns the one plus the position in the string
**     where the conversion failed; in the latter case the value of `res'
**     remains unchanged.
*/

int rfc822_time (char *rfc822, time_t *res)
{
    char *p = rfc822, *q;
    struct tm ts;
    long iv, tz;
    time_t rs;
    while (isblank (*p)) ++p;
    if (!*p) { return p - rfc822 + 1; }
    /* Check, if the day of the week is present */
    if ((q = strchr (p, ','))) {
	if ((ts.tm_wday = wday (p)) < 0) { return p - rfc822 + 1; }
	while (*p && !isblank (*p) && *p != ',') ++p;
	if (!*p) { return p - rfc822 + 1; }
	/* There must not be another character than `,' after the weekday
	** (with the exception of white spaces).
	*/
	while (isblank (*p)) ++p;
	if (*p != ',') { return p - rfc822 + 1; }
	p = q + 1;
    }
    while (isblank (*p)) ++p;
    if (!*p) { return p - rfc822 + 1; }
    iv = strtol (p, &q, 10);
    if (!isblank (*q)) { return q - rfc822 + 1; }
    ts.tm_mday = (int) iv;
    p = q;
    while (isblank (*p)) ++p;
    if ((ts.tm_mon = month (p)) < 0) { return p - rfc822 + 1; }
    while (*p && !isblank (*p)) ++p;
    if (!*p) { return p - rfc822 + 1; }
    iv = strtol (p, &q, 10);
    if (!isblank (*q)) { return q - rfc822 + 1; }
    ts.tm_year = (int) (iv - 1900l);
    p = q;
    while (isblank (*p)) ++p;
    if (!*p) { return p - rfc822 + 1; }
    iv = strtol (p, &q, 10);
    if (*q != ':') { return q - rfc822 + 1; }
    if (iv < 0l || iv > 23l) { return p - rfc822 + 1; }
    ts.tm_hour = (int) iv;
    p = q + 1;
    iv = strtol (p, &q, 10);
    if (*q != ':' && !isblz(*q)) { return q - rfc822 + 1; }
    if (iv < 0l || iv > 59l) { return p - rfc822 + 1; }
    ts.tm_min = (int) iv;
    ts.tm_sec = 0;
    if (*q == ':') {
	p = q + 1;
	iv = strtol (p, &q, 10);
	if (*q && !isblank (*q)) { return q - rfc822 + 1; }
	if (iv < 0l || iv > 59l) { return p - rfc822 + 1; }
	ts.tm_sec = (int) iv;
    }
    p = q;
    /* Calculate the timezone distance */
    if (get_tz (p, &tz, &q)) { return q - rfc822 + 1; }
    p = q;
    /* Daylight saving time - what is this? */
    ts.tm_isdst = 1;
    /* This value is not needed by `mktime' */
    ts.tm_yday = 0;
    if ((rs = mktime (&ts)) == (time_t)(-1)) { return p - rfc822 + 1; }
    rs = rs + timezone - tz;
    *res = rs;
    return 0;
}
