/* xfile.c
**
** Handle folder files with index files
**
*/

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <strings.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/file.h>
#include <fcntl.h>
#include <stdarg.h>
#include <errno.h>

#include "xfile.h"

#define IXSFX ".idx"

#ifndef __GNUC__
#define __inline__
#endif

/* Was ist zu tun?
** Ich bentige
** a) Schreiboperationen, die ausschlielich an eine vorhandene
**    Folderdatei anhngen oder eine neue (cum Indexdatei) erzeugen.
** b) Leseoperationen, die ber einen Index die Position bestimmen, an der
**    sie in der Folderdatei lesen und am Ende des Eintrages ein EOF liefern.
*/
struct __xFile {
    int dfd, xfd;
    int mset;
    char mode[2];
    off_t pos;
    /* `len' enthlt die Lnge der durch `xSeek' adressierten E-Mail */
    size_t len;
    /* `buf' zeigt IMMER auf den Anfang des Puffers; `bfp' zeigt auf das
     * nchste gltige Zeichen im Puffer; `bfe' zeigt (ausschlielich bei
     * Leseoperationen) auf das Ende der tatschlich im Puffer enthaltenen
     * Zeichen.
     */
    char *buf, *bfp, *bfe;
    /* `bfs' enthlt die Lnge des Puffers, nicht die Anzahl der noch
     * enthaltenen Zeichen
     */
    size_t bfs;
};

static size_t BUFSIZE = 4096;

/* xoffset(xfd, n)
 *     liefert die Dateiposition der n. Nachricht.
 *     Fr n = 0 mu dies immer 0 sein.
 */
static off_t xoffset (int xfd, int pos)
{
    unsigned char pbf[sizeof (off_t)];
    off_t res = 0, fptr, flen = lseek (xfd, 0, SEEK_END);
    int ix;
    memset (pbf, '\0', sizeof (off_t));
    if (pos < 0) return (off_t) -1;
    if (!pos) return (off_t) 0;
    fptr = (off_t) (pos - 1) * sizeof (off_t);
    if (fptr >= flen) { return (off_t) -1; }
    if (lseek (xfd, fptr, SEEK_SET) < 0) { return (off_t) -1; }
/*    if (read (xfd, pbf, sizeof (off_t)) != sizeof (off_t)) {*/
    if (read (xfd, pbf, sizeof (off_t)) < 0) {
	return (off_t) -1;
    }
    for (ix = 0; ix < sizeof (off_t); ++ix) {
	res <<= 8;
	res |= pbf[ix];
    }
    return res;
}

/* xlength(fd, n)
 *     liefert die Lnge der n. Nachricht
 */
static ssize_t xlength (int xfd, int pos)
{
    off_t ep, sp;
    sp = xoffset (xfd, pos);
    ep = xoffset (xfd, pos + 1);
    if (sp < 0 || ep < 0) {
	return (ssize_t) -1;
    }
    return (ssize_t) (ep - sp);
}

/* Set the buffer size for the read/write buffer */
void xSetBufSize (size_t siz)
{
    BUFSIZE = siz;
}

/* xOpen
** ffne die Folderdatei(en) zusammen mit der zugehrigen Indexdatei
** Es gibt folgende Arten diese Datei zu ffnen:
** a) im Schreibmodus (*amode == 'w'), in dem lediglich Daten an die vorhandene
**    Datei angehngt werden knnen (die Datei kann dabei auch neu erzeugt
**    worden sein),
** b) im Lesemodus (*amode == 'r'), in dem wahlfreier Zugriff, jedoch nur lesend
**    mglich ist.
** c) im Update-Modus (*amode == 'u'), in dem die Folderdatei(en) neu erzeugt
**    bzw. entleert werden; danach funktioniert dieser Modus genau wie der
**    Schreibmodus.
** d) im Index-Modus (*amode == 'i'), in dem zu einer vorhandenen Folderdatei
**    eine Indexdatei angelegt wird.
** Das zweite Zeichen von amode (amode[1]) gibt an, ob und wie eine Dateisperre
** fr die Datei angelegt werden soll:
** a) amode[1] == 's': es wird ein `shared lock' auf die gesamte Folderdatei
**    (und die zugehrige Indexdatei) gelegt; dies bedeutet, da mehrere
**    unterschiedliche Prozesse gleichzeitig auf diese Datei(en) zugreifen
**    knnen - jedoch keiner exklusiv.
** b) amode[1] == 'x': es wird ein `exclusive lock' auf die gesamte Folderdatei
**    (und die zugehrige Indexdatei) gelegt; in diesem Modus kann genau ein
**    Proze auf diese Datei(en) zugreifen kann.
*/
xFile xOpen (char *path, char *amode, ...)
{
    va_list ap;
    int dfd, xfd;
    int mode = 0, oFlags, iFlags, nFlags = 0, lock;
    char *ixpath;
    xFile res;
    /* Ich erlaube zwar nur lesenden Zugriff, ffne die Datei aber trotzdem
    ** lesend und schreibend. Der Grund hierfr ist einfach:
    ** Es gibt einige (vor allem ltere) Unix-Systeme, die ein Sperren von
    ** Dateien nur auf (auch) zum Schreiben geffneten Dateien zulassen.
    ** Da meine Routinen sich sowieso nach dem Mode-Flag (`i'...`u') richten,
    ** bleibt die Integritt der Mail-Folder trotzdem erhalten.
    */
    if (*amode == 'r') {
	iFlags = oFlags = O_RDWR;
    } else if (*amode == 'w' || *amode == 'u') {
	if (*amode == 'w') {
	    iFlags = oFlags = O_CREAT|O_APPEND|O_RDWR;
	} else {
	    iFlags = oFlags = O_CREAT|O_RDWR;
	    nFlags = O_CREAT|O_TRUNC|O_RDWR;
	}
	va_start (ap, amode);
	mode = va_arg (ap, int);
	va_end (ap);
    } else if (*amode == 'i') {
	oFlags = O_RDWR;
	iFlags = O_CREAT|O_TRUNC|O_RDWR;
	va_start (ap, amode);
	mode = va_arg (ap, int);
	va_end (ap);
    } else {
	return (xFile) 0;
    }
    /* Ich ziehe das File locking aus dem Zugriffsmodus heraus und lege es
     * mittels eines weiteren Zeichens in der Zeichenkette `amode' fest;
     * dies hat den Vorteil, da dieses Modul universeller einsetzbar ist.
     * Natrlich mssen `vxdeliver', ... entsprechend angepat werden.
     */
#ifdef __USE_BSD
    lock = (amode[1] == 's' ? LOCK_SH : (amode[1] == 'x' ? LOCK_EX : -1));
#else
    lock = (amode[1] == 'x' ? F_LOCK : -1);
#endif
    if (!(ixpath = (char *) malloc (strlen (path) + strlen (IXSFX) + 1))) {
	return (xFile) 0;
    }
    strcpy (ixpath, path);
    strcat (ixpath, IXSFX);
    if (!(res = (xFile) malloc (sizeof (struct __xFile)))) {
	free (ixpath);
	return res;
    }
    if ((dfd = open (path, oFlags, mode)) < 0) {
	free (res); free (ixpath);
	return (xFile) 0;
    }
    if (lock >= 0) {
#ifdef __USE_BSD
	if (flock (dfd, lock) < 0) {
#else
	if (lockf (dfd, lock, 0) < 0) {
#endif
	    close (dfd);
	    free (res); free (ixpath);
	    return (xFile) 0;
	}
    }
    if ((xfd = open (ixpath, iFlags, mode)) < 0) {
	close (dfd);
	free (res); free (ixpath);
	return (xFile) 0;
    }
    if (lock >= 0) {
#ifdef __USE_BSD
	if (flock (xfd, lock) < 0) {
#else
	if (lockf (xfd, lock, 0) < 0) {
#endif
	    close (xfd); close (dfd);
	    free (res); free (ixpath);
	    return (xFile) 0;
	}
    }
    if (*amode == 'u') {
	/* I didn't open the file(s) above with O_TRUNC, because O_TRUNC
	** even works with an active lock on a file; the trick is it, to
	** open a new file after `locking' and closing the old one immediately
	** thereafter.
	** The linux documentation says, the O_EXCL flag should not be used
	** because this doesn't work with NFS. So this seems the only way
	** to suspend erasing file contents if another process needs these
	** contents - and so has a lock set on this folder.
	*/
	int ndfd = open (path, nFlags, mode),
	    nxfd = open (ixpath, nFlags, mode);
	if (ndfd < 0 || nxfd < 0) {
	    if (nxfd >= 0) { close (nxfd); }
	    if (ndfd >= 0) { close (ndfd); }
	    close (xfd); close (dfd);
	    free (res); free (ixpath);
	    return (xFile) 0;
	}
#if 0
#ifdef __USE_BSD
	if (flock (ndfd, lock) < 0) {
/* } */
#else
	if (lockf (ndfd, lock, 0) < 0) {
#endif
	    close (nxfd); close (ndfd);
	    close (xfd); close (dfd);
	    free (res); free (ixpath);
	    return (xFile) 0;
	}
#ifdef __USE_BSD
	if (flock (nxfd, lock) < 0) {
/* } */
#else
	if (lockf (nxfd, lock, 0) < 0) {
#endif
	    close (nxfd); close (ndfd);
	    close (xfd); close (dfd);
	    free (res); free (ixpath);
	    return (xFile) 0;
	}
#endif
	/* closing the old file descriptors */
	close (dfd); dfd = ndfd;
	close (xfd); xfd = nxfd;
    }
    /* Now, even files opened with `u' are not simply overwritten if they
    ** are locked by another process
    */
    res->mode[0] = *amode;
    res->mode[1] = amode[1];
    res->dfd = dfd; res->xfd = xfd;
    if (*amode == 'r') {
	/* Bei einem Lesezugriff mu nach `open' die Ausgangsposition
	** zunchst die Position der ersten Nachricht sein.
	*/
	res->pos = xoffset (xfd, 0); res->len = xlength (xfd, 0);
    } else if (*amode == 'i') {
	/* Wenn der Folder im Index-Modus geffnet wurde, dann mu das Flag,
	** welches angibt, da eine Nachricht zu ende ist, stets ungesetzt
	** sein (und bleiben).
	*/
	res->mset = 0;
	res->pos = (off_t) 0;
	res->len = (size_t) lseek (dfd, 0, SEEK_END);
	lseek (dfd, 0, SEEK_SET);
    } else {
	/* Bei einem Schreib- oder Updatezugriff mu das Flag, welches angibt,
	** da eine Nachricht zu ende ist, zunchst gesetzt sein; wird nmlich
	** nichts geschrieben, dann mu die Datei unverndert bleiben; wird
	** hingegen etwas geschrieben, dann mu zum Schlu ein neuer Marker
	** fr die Dateiposition am Ende der Nachricht (`xSet' oder `xClose')
	** eingetragen werden. Dies darf aber nur dann geschehen, wenn
	** tatschlich etwas geschrieben wurde.
	*/
	res->mset = 1;
    }
    res->buf = (char *) 0;
    res->bfp = (char *) 0;
    res->bfe = (char *) 0;
    free (ixpath);
    return res;
}

/* Setze den internen Dateizeiger auf den Anfang der Nachricht `index' (von
 * Null ab gezhlt!) und die interne Lnge auf die Lnge der Nachricht.
 * Diese Operation funktioniert nur bei lesendem Zugriff auf die Datei.
 * (Warum die letzte Einschrnkung?
 *  Nun, ich gehe davon aus, da beim Eintreffen einer neuen Mail, diese immer
 *  hinten an den Folder angehngt wird; Schreibzugriff bedeutet hier also
 *  einen rein sequentiellen Zugriff, whrend man bei einem Lesezugriff einen
 *  Indexsequentiellen Zugriff mittels der Nummer der Nachricht als Index hat.
 *  Das sind dann genau die Operationen, die fr einen normalen Zugriff ben-
 *  tigt werden (SMTP - Schreibzugriff, POP3 - Lesezugriff).
 *  Der fr das Entfernen einzelner Messages notwendigen Zugriff mu sowieso
 *  auf eine ganz andere Art geschehen, so da ich hierbei verbleiben kann.)
 */
int xSeek (xFile fx, off_t index)
{
    off_t offset;
    ssize_t size;
    /* Vergewissere dich, da `fx' mit Lesezugriff geffnet wurde. */
    if (fx->mode[0] != 'r') { errno = EACCES; return -1; }
    /* Wenn der angegebene Index auerhalb des zulssigen Bereiches liegt,
     * dann liefert `xoffset' einen negativen Wert; kehre dann ebenfalls mit
     * einem negative Wert zurck.
     */
    if ((offset = xoffset (fx->xfd, index)) < 0) {
	return -2;
    }
    /* Berechne die Lnge der Nachricht aus zwei aufeinanderfolgenden Indices
     * Wenn diese < 0 ist, dann */
    if ((size = xlength (fx->xfd, index)) < 0) {
	return -3;
    }
    /* Trage die Werte fr die Dateiposition und die Message-Lnge in die
     * internen Variablen der Struktur ein.
     */
    fx->pos = offset;
    fx->len = size;
    /* Setze den Dateizeiger der Datendatei auf den Anfang der zugehrigen
     * E-Mail innerhalb der Datendatei.
     */
    offset = lseek (fx->dfd, fx->pos, SEEK_SET);
    if (fx->buf) {
	fx->bfp = fx->bfe = fx->buf;
    }
    return index;
}

__inline__ int bfalloc (xFile fx)
{
    if (!fx->buf) {
	if (!(fx->buf = (char *) malloc (BUFSIZE))) {
	    return -1;
	}
	fx->bfs = BUFSIZE;
	/* Aktuelle Schreibposition */
	fx->bfp = fx->buf;
	/* Ende der eingelesenen Daten */
	fx->bfe = fx->buf;
    }
    return 0;
}

ssize_t xFlush (xFile fx)
{
    ssize_t res;
    if (fx->bfp == fx->buf) {
	return 0;
    }
    res = write (fx->dfd, fx->buf, (size_t) (fx->bfp - fx->buf));
    fx->bfp = fx->buf;
    return res;
}

ssize_t xWrite (xFile fx, void *buf, size_t bfs)
{
    size_t rsz;
    ssize_t res = 0, r1;
    /* Wenn dei Datei weder fr Schreib- noch fr Update-Modus geffnet wurde,
    ** dann darf keine Schreiboperation ausgefhrt werden.
    */
    if (strchr ("ri", fx->mode[0])) { errno = EACCES; return -1; }
    if (bfalloc (fx) < 0) { return -1; }
    /* Wenn der Puffer angelegt wurde, dann mu(!) das Marker-Flag
     * zurckgesetzt werden, da dies schon als abgeschlossene Schreiboperation
     * gilt (bzw. betrachtet werden mu); es wird ja nicht sofort geschrieben,
     * sondern zeitversetzt - nmlich dann, wenn der Puffer gefllt ist.
     */
    fx->mset = 0;
    /* Restlnge des Puffers berechnen */
    rsz = fx->bfs - (fx->bfp - fx->buf);
    /* Solange noch Daten vorhanden sind */
    while (bfs > rsz) {
	/* bertrage Daten in den Schreibpuffer bis dieser gefllt ist. */
	memcpy (fx->bfp, buf, rsz);
	fx->bfp += rsz;
	/* Wenn der Puffer gefllt ist, dann bertrage seinen Inhalt in
	 * die Datendatei
	 */
	if ((r1 = xFlush (fx)) < 0) { return res; }
	/* Die Anzahl der bisher bertragenen Bytes geht nach `res'. */
	res += r1;
	/* Setze den Eingabedatenzeiger auf die nchsten Eingabedaten */
	buf =  (char *) buf + rsz;
	/* Verringere die Eingabedatenlnge um die Anzahl der bereits in
	 * die Datendatei bertragenen Bytes
	 */
	bfs -= rsz;
	/* Setze die Restlnge des Puffers neu */
	rsz = fx->bfs;
    }
    /* Wenn noch Daten brig sind, dann bertragge sie in den Puffer */
    if (bfs > 0) {
	memcpy (fx->bfp, buf, bfs);
	fx->bfp += bfs;
	res += bfs;
    }
    return res;
}

/* Flle den Puffer neu; diese Operation wird NUR dann ausgefhrt, wenn
 * der Puffer vollstndig leer ist.
 */
static ssize_t xfill (xFile fx)
{
    ssize_t rlen;
    ssize_t rest;
    rest = fx->len - (size_t) (lseek (fx->dfd, 0, SEEK_CUR) - fx->pos);
    /* Wenn keine Bytes in der Nachricht mehr brig sind, dann gib 0 zurck */
    if (rest <= 0) return 0;
    /* Wenn noch mehr Daten in der Nachricht sind als in den Puffer passen,
     * dann lies nur soviele von diesen Daten, wie in den Puffer passen.
     */
    if (rest > fx->bfs) rest = (ssize_t) fx->bfs;
    /* Wenn die Leseoperation fehlschlgt, dann gib -1 zurck. */
    if ((rlen = read (fx->dfd, fx->buf, (size_t) rest)) < 0) { return -1; }
    /* Wenn weniger Bytes gelesen wurden als erwartet, dann setze die
     * Restlnge (das Funktionsergebnis) so, da sie der Anzahl der
     * tatschlich gelesenen Bytes entspricht.
     */
    if (rlen < rest) rest = rlen;
    /* Setze den aktuellen Lesezeiger und das aktuelle Pufferende */
    fx->bfp = fx->buf;
    fx->bfe = fx->buf + rest;
    /* Kehre mit der Anzahl der tatschlich gelesenen Bytes zurck */
    return rest;
}

/* Fr die Leseoperation bentige ich einen weiteren Zeiger;
 * auerdem sollen nur noch soviele Daten bertragen werden, wie noch in der
 * Nachricht enthalten sind
 */
ssize_t xRead (xFile fx, void *buf, size_t bfs)
{
    size_t rsz;
    ssize_t res = 0;
    ssize_t rblen;
    /* berprfe, ob lesen mglich ist. */
    if (strchr ("wu", fx->mode[0])) { errno = EACCES; return -1; }
    /* Wenn kein Speicherplatz mehr brig fr den Puffer, dann kann auch
     * nichts mehr gelesen werden.
     */
    if (bfalloc (fx) < 0) { return -1; }
    /* Wenn noch Daten im Puffer sind, dann bertrage zunchst diese */
    rsz = fx->bfe - fx->bfp;
    if (rsz > 0) {
	if (rsz > bfs) { rsz = bfs; }
	memcpy (buf, fx->bfp, rsz);
	fx->bfp += rsz;
	buf = (char *) buf + rsz;
	bfs -= rsz;
	res += (ssize_t) rsz;
    }
    /* Solange noch Daten gelesen werden mssen */
    while (bfs > 0) {
	/* Wenn keine Daten mehr brig sind oder ein Fehler aufgetreten ist,
	 * dann beende und kehre mit der Anzahl der bisher gelesenen Daten
	 * zurck.
	 */
	if ((rblen = xfill (fx)) <= 0) { return res; }
	/* Berechne die Menge der Daten im Puffer */
	rsz = fx->bfe - fx->bfp;
	/* Wenn mehr Daten im Puffer sind, als bentigt werden, dann
	 * bertrage nur soviele Daten, wie bentigt werden.
	 */
	if (rsz > bfs) { rsz = bfs; }
	memcpy (buf, fx->bfp, (size_t) rsz);
	/* Setze den Pufferzeiger entsprechend weiter */
	fx->bfp += rsz;
	/* Setze den Ergebniszeiger entsprechend weiter */
	buf = (char *) buf + rsz;
	/* Berechne die Menge der noch zu lesenden Daten */
	bfs -= rsz;
	res += rsz;
    }
    return res;
}

ssize_t xMsgLen (xFile fx, int pos)
{
    if (fx->mode[0] != 'r') { errno = EACCES; return -1; }
    if (pos < 0) {
	return fx->len;
    }
    return xlength (fx->xfd, pos);
}

/* Trage eine Zeiger auf die aktuelle Dateiposition in der Datendatei in die
** Indexdatei ein.
** Wenn der Folder im Index-Modus (s.o.) geffnet wurde, dann mu stattdessen
** die gewnschte Dateiposition als zweiter Parameter angegeben werden - da
** dann von einer schon existierenden Folderdatei ohne Index ausgegangen
** werden mu.
*/
int xSet (xFile fx, ...)
{
    unsigned char pbf[sizeof (off_t)];
    int ix;
    off_t fofs;
    va_list av;
    /* Diese Operation darf in allen Modi auer dem Lese-Modus erfolgen. */
    if (fx->mode[0] == 'r') { errno = EACCES; return -1; }
    /* Setze nur dann einen Marker, wenn noch keiner vorhanden ist! */
    if (fx->mset) { return 0; }
    /* Wenn die Datei im Index-Modus geffnet wurde, dann mu die
    ** einzustellende Indexposition als zweites Argument bergeben werden;
    ** wurde sie hingegen im Schreib- oder Update-Modus geffnet, dann ist
    ** die Indexposition der aktuelle Stand des Dateizeigers nach einer
    ** erfolgten `flush'-Operation.
    */
    if (fx->mode[0] == 'i') {
	va_start (av, fx);
	fofs = va_arg (av, off_t);
	va_end (ap);
    } else {
	xFlush (fx);
	fofs = lseek (fx->dfd, 0, SEEK_CUR);
    }
    /* Die Dateipositionen speichere ich grundstzlich in `big endian order'
    ** ab. So gibt es keinerlei Probleme, das ganze auf Architekturen mit
    ** unterschiedlicher Byteorder zu verwenden.
    */
    /* Wandle die aktuelle Dateiposition in eine MSB-geordnete Bytefolge um */
    for (ix = sizeof (off_t) - 1; ix >= 0; --ix) {
	pbf[ix] = (unsigned char) (fofs & 0xff);
	fofs >>= 8;
    }
    /* Schreibe die eben erzeugte Bytefolge in die Indexdatei */
    if (write (fx->xfd, pbf, sizeof (off_t)) != sizeof (off_t)) {
	return (off_t) -1;
    }
    /* Im Index-Modus darf (selbstverstndlich das Setzen einer Indexposition
    ** niemals unterbunden werden; dieser Modus dient ja gerade zum Anlegen
    ** der Indexdatei. In den beiden anderen Fllen (Write- und Update-Modus)
    ** hingegen wird durch dieses Flag das Setzen einer ungltigen Indexposi-
    ** tion (keine weiteren Daten geschrieben) unterbunden. Das Setzen eines
    ** Index wird durch die erste `write'-Operation wieder ermglicht.
    */
    if (fx->mode[0] != 'i') fx->mset = 1;
    return 0;
}

/* Setze den Lesezeiger auf den Anfang der aktuellen Meldung zurck oder
** beende den Indexierungsmodus (`i').
** Wie gut, da ich diese Routine schon eingeplant hatte; ich bentige sie
** fr das Rcksetzen nach einer automatischen Erkennung einer komprimierten
** E-Mail, sowie fr die Umschaltung vom Schreib-, Update- und Indexierungs-
** in den Lesemodus.
*/
int xReSet (xFile fx)
{
    /* Also:
    **   Wenn der Folder im Schreib-, Update- oder Indexmodus geffnet wurde,
    **   dann soll er nun im Lesemodus geffnet sein, wobei dann die internen
    **   Zeiger auf die erste im Folder enthaltene E-Mail zurckgesetzt werden.
    */
    if (strchr ("iuw", fx->mode[0])) {
	if (fx->mode[0] != 'i') { xSet (fx); }
	fx->mode[0] = 'r';
	if (xSeek (fx, 0) < 0) { return -1; }
	return 0;
    } else {
	/* Ist der Folder hingegen im Lesemodus geffnet, dann werden die
	** internen Zeiger auf den Anfang der aktuellen E-Mail zurckgestellt.
	*/
	if (lseek (fx->dfd, fx->pos, SEEK_SET) < 0) { return -1; }
    }
    /* Diese Stelle hatte ich vergessen; natrlich mu (!!!) der Eingabepuffer
    ** nach einer Re-Positionierung auf den Anfang der E-Mail geleert werden,
    ** da sonst Zeichen beim nchsten Einlesevorgang doppelt gelesen werden
    ** knnten!
    */
    fx->bfp = fx->bfe = fx->buf;
    return 0;
}

/* Bei einer Schlieoperation mu folgendes getan werden:
** a) Wenn die Datei zum schreiben geffnet war, dann mu der Puffer zurck-
**    geschrieben werden.
** b) Wenn die Datei zum Schreiben geffnet war, dann mu die
**    Dateiendeposition in der Indexdatei abgelegt werden.
** c) Wenn ein Schreib- oder Lesepuffer angelegt wurde, dann mu dieser
**    freigegeben werden.
** d) Die Daten- und die Indexdatei mssen geschlossen werden.
** e) Das xFile-Handle mu freigegeben werden.
*/
int xClose (xFile fx)
{
    int e0 = 0, e1, e2;
    /* Wenn der Folder im Schreib- oder Update-Modus geffnet wurde,
    ** dann mu vor dem Schlieen eine xSet-Operation erfolgen.
    ** Da ich im Index-Modus sowieso die einzelnen Positionen von Hand setzen
    ** mu, wird dies von `xClose' nicht mehr bernommen.
    */
    if (strchr ("wu", fx->mode[0])) {
	/* a) und b) werden durch den Aufruf von xSet automatisch erledigt. */
	e0 = xSet (fx);
    }
    /* Schritt c) */
    if (fx->buf) { free (fx->buf); }
    /* Schritt d) */
    e1 = close (fx->dfd);
    e2 = close (fx->xfd);
    /* Schritt e) */
    free (fx);
    return (e0 ? e0 : (e1 ? e1 : e2));
}

/* Zustzliche Funktionen */

/* xPuts(fx, string)
**     Schreibe eine '\0'-terminierte Zeichenkette in die Datei `fx'
*/
ssize_t xPuts (char *buf, xFile fx)
{
    return xWrite (fx, buf, strlen (buf));
}

/* xGetc(fx)
**     Lies ein einzelnes Zeichen aus der Datei `fx'
** Analog zu `getc' liefert `xGetc' eine ganze Zahl. Das ermglicht es,
** im Fehlerfall (oder auch bei Dateiende) einen passenden Wert zurckzugeben;
** Im Fehlerfall wird `-2', bei Dateiende `-1' zurckgegeben.
*/
int xGetc (xFile fx)
{
    ssize_t rsz, rblen;
    /* Weder im Schreib- noch im Update-Modus knnen Daten gelesen werden. */
    if (strchr ("wu", fx->mode[0])) { errno = EACCES; return -2; }
    /* Wenn kein Speicherplatz mehr brig fr den Puffer, dann kann auch
    ** nichts mehr gelesen werden.
    */
    if (bfalloc (fx) < 0) { return -2; }
    /* Wenn noch Daten im Puffer sind, dann bertrage zunchst diese */
    rsz = fx->bfe - fx->bfp;
    if (rsz <= 0) {
	rblen = xfill (fx);
	if (rblen < 0) return -2;
	if (rblen == 0) return -1;
    }
    return (*fx->bfp++ & 0xff);
}

/* xGets (buf, siz, fx)
**     Lies eine Zeile - hchstens aber `siz' Zeichen aus der Datei`fx'
**     in den Puffer `buf'.
** Im Unterschied zu `fgets' wird hier nicht ein Zeiger auf eine Zeichenkette
** zurckgegeben, sondern die Anzahl der gelesenen Zeichen.
*/
ssize_t xGets (char *buf, size_t siz, xFile fx)
{
    ssize_t rsz, bsz, psz = 0;
    char *p = (char *) 0;
    if (bfalloc (fx) < 0) { return -2; }
    /* Weder im Schreib- noch im Update-Modus knnen Daten gelesen werden. */
    if (strchr ("wu", fx->mode[0])) { errno = EACCES; return (ssize_t) -1; }
    rsz = (ssize_t) siz - 1;
    while (!p && rsz > 0) {
	/* Teste zuerst mal den Pufferinhalt */
	bsz = fx->bfe - fx->bfp;
	if (bsz <= 0 && xfill (fx) <= 0) { break; }
	/* Suche nach dem ersten Zeilenende im Puffer */
	p = memchr (fx->bfp, '\n', fx->bfe - fx->bfp);
	/* Wenn ein Zeilenende gefunden wurde, dann bertrage alles bis zu
	** diesem (einschlielich); ansonsten bertrage einfach den gesamten
	** Pufferinhalt.
	*/
	bsz = (p ? (p + 1) : fx->bfe) - fx->bfp;
	/* bertrage natrlich nur soviele Bytes, wie in den Zielbereich
	** hineinpassen.
	*/
	if (bsz > rsz) { bsz = rsz; }
	/* - und zwar von der aktuellen Position im Puffer an. */
	memcpy (buf, fx->bfp, bsz);
	/* Setze die Zeiger entsprechend weiter.*/
	fx->bfp += bsz; buf += bsz; rsz -= bsz;
	/* Die gesamte Anzahl der bisher gelesenen Bytes steht hiernach in
	** `psz'.
	*/
	psz += bsz;
    }
    /* Setze in jedem Fall noch ein NUL-Zeichen hinterher */
    *buf = '\0';
    return psz;
}

/* xPutc (cc, fx)
**     Schreibe ein einzelnes Zeichen in die Datei `fx'.
** Im Fehlerfall wird `-2' zurckgegeben, sonst `0'. Die Ausgabe erfolgt -
** genau wie bai allen anderen hier definierten Schreib- und Leseoperationen -
** gepuffert.
*/
int xPutc (int cc, xFile fx)
{
    ssize_t rsz;
    /* Weder im Lese- noch im Index-Modus drfen Daten geschrieben werden. */
    if (strchr ("ri", fx->mode[0])) { errno = EACCES; return -2; }
    /* Wenn kein Speicherplatz mehr fr den Puffer brig ist, dann kann auch
    ** nichts mehr gelesen werden.
    */
    if (bfalloc (fx) < 0) { return -2; }
    rsz = fx->bfs - (fx->bfp - fx->buf);
    if (rsz <= 0) { xFlush (fx); }
    *fx->bfp++ = (char) cc;
    return 0;
}

/* xUnlink (filename)
**     lscht den Folder `filename' (mit samt seinem Index).
** Wie die Indexdatei zu einem Folder heit soll ausschlielich durch dieses
** Modul bestimmt werden. Um das zu gewhrleisten, mute ich eine erweiterte
** Dateilschfunktion einfhren, die mit dieser Situation umgehen kann.
*/
int xUnlink (char *file)
{
    char *ixfile, e1, e2;
    if (!(ixfile = (char *) malloc (strlen (file) + strlen (IXSFX) + 1))) {
	return -1;
    }
    strcpy (ixfile, file);
    strcat (ixfile, IXSFX);
    e1 = unlink (file);
    e2 = unlink (ixfile);
    free (ixfile);
    return (e1 ? e1 : e2);
}

/* xNoIndex (filename)
**     lscht die Indexdatei zu einem Folder.
** Diese Operation bentige ich fr die E-Mail-Folder eines Standardsystems
** (`smail' oder `sendmail').
** Damit mein POP-Server nmlich mit diesen umgehen kann, mu temporr eine
** Indexdatei erzeugt werden. Damit diese nach vollendeter POP3-Session aber
** wieder gelscht werden kann, bentige ich eine eigene Lschroutine.
*/
int xNoIndex (char *file)
{
    char *ixfile, e;
    if (!(ixfile = (char *) malloc (strlen (file) + strlen (IXSFX) + 1))) {
	return -1;
    }
    strcpy (ixfile, file);
    strcat (ixfile, IXSFX);
    e = unlink (ixfile);
    free (ixfile);
    return e;
}

/* xEmpty (filehandle)
**     gibt 1 (true) zurck, falls der Folder, auf den gerade mittels
**     `filehandle' zugegriffen wird, leer ist und 0 (false), wenn nicht.
*/
int xEmpty (xFile fx)
{
    off_t oldoff = lseek (fx->dfd, 0, SEEK_CUR);
    off_t newoff = lseek (fx->dfd, 0, SEEK_END);
    if (oldoff < 0 || newoff < 0) { return -1; }
    lseek (fx->dfd, oldoff, SEEK_SET);
    if (newoff) { return 0; }
    return 1;
}

/* xChown (filehandle, uid, gid)
 *     setzt den Besitzer bzw. die Gruppe der Folder- und Indexdatei auf
 *     die bergebenen Werte `uid' `gid'
 */
int xChown (xFile fx, uid_t uid, gid_t gid)
{
    int ec = fchown (fx->dfd, uid, gid);
    if (ec < 0) { return ec; }
    ec = fchown (fx->xfd, uid, gid);
    return ec;
}
